/*
	File: afmm.icl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Alegbraic FMM algorithms
	See also: CamlFloat
	Date: 7.23.2003
	
	$Id: afmm.icl,v 1.1 2003/07/25 17:30:25 shiv Exp shiv $
*/

implementation module afmm

import StdEnv, complex, MersenneTwister, blop

:: Tree a b = Leaf !a | Node !b !{Tree a b}

:: AfmmNode e :== (!{#Int}, !{Matrix e})
				
:: AfmmLeaf e :== Matrix e	// u or vH (not v!)

:: AfmmPartition e :== Tree (AfmmLeaf e) (AfmmNode e)

:: Afmm e = Afmm !(AfmmPartition e) !(AfmmPartition e) !(Partition e)

:: Partition e = Unstruct !(Matrix e)
			   | ExpCoeff !(Matrix e)
			   | Partitio !{{Partition e}}

instance noOfRows Afmm e | noOfRows Matrix e
where
	noOfRows (Afmm _ (Node (ms, _) _) _) = sumA ms
	noOfRows (Afmm _ (Leaf a)         _) = noOfRows a

instance noOfCols Afmm e | noOfCols Matrix e
where
	noOfCols (Afmm (Node (ns, _) _) _ _) = sumA ns
	noOfCols (Afmm (Leaf a)         _ _) = noOfCols a

sumA ms = sumR (size ms - 1) 0
where
	sumR i acc
		| i < 0     = acc
		| otherwise = sumR (i-1) (acc + ms.[i])


fMul :: !(Afmm Real) !(Matrix Real) !*(Matrix Real) -> *(Matrix Real)
fMul (Afmm   (Leaf _)
	(Leaf _) (Unstruct d)) b z = (d :*: b) z

fMul (Afmm   (Leaf vH)
	(Leaf u) (ExpCoeff e)) b z = (u :*: e *$ vH *$ b) z

fMul (Afmm           (cols=:(Leaf _))
	(rows=:(Leaf _)) (Partitio a)) b z
	| size a == 1 && size a.[0] == 1 = fMul (Afmm cols rows a.[0,0]) b z
	| otherwise = abort "fMul: Incompatible partitioning at leaf"

fMul (afmm=:(Afmm     cols
			rows (Partitio a))) b z =
	let g = fmmUpSweep cols b in fmmDownSweep afmm b g (Leaf (zeros 0 (noOfCols b))) z

fMul _ _ _ = abort "fMul: Bad partitioning"

fmmUpSweep (Leaf vH)          b = Leaf (vH *$ b)
fmmUpSweep (Node (ns, ws) cs) b = Node g         gs
where
	n = size ns
	bs = partitionNx1 ns b
	gs = { createArray n (Leaf (zeros 0 0)) &
		   [i] = fmmUpSweep cs.[i] bs.[i] \\ i <- [0 .. n-1]
		 }
	g = foldl (+$) (zeros (noOfRows ws.[0]) (noOfCols b))
				 [wi *$ getNodeData gi \\ wi <-: ws & gi <-: gs]

	getNodeData (Leaf g  ) = g
	getNodeData (Node g _) = g


fmmDownSweep (Afmm                cols
		(Node (ms, ws) rs) (Partitio a)) b g f z = (zf, Node 0 ff)
where
	(zf, ff) = matNx1 ms { createArray m (\a b z f g = undef) &
						   [i] = fmmDownSweep (Afmm cols rs.[i] a.[i]) b g
						   \\ i <- [0 .. m-1]
						 } z fs
	m = size ms
	fs = case f of
			Leaf f = {createArray m (zeros 0 0) & [i] = Leaf (ws.[i] *$ ff)
						\\ i <- [0 .. m-1]}
			Node _ fs = fs

fmmDownSweep (Afmm   (Node (ns, ws) cs)
			(Leaf u) (Partitio a)      ) b g f z = (zf, Node 0 ff)
where
	(zf, ff) = matNx1 ms { createArray m (\a b z f g = undef) &
						   [i] = fmmDownSweep (Afmm cols rs.[i] a.[i]) b g
						   \\ i <- [0 .. m-1]
						 } z fs
	n = size ns
	fs = case f of
			Leaf f = {createArray m (zeros 0 0) & [i] = Leaf (ws.[i] *$ ff)
						\\ i <- [0 .. m-1]}
			Node _ fs = fs

/*
iFmm (Afmm   (Node (ns, ws) cs)
	(Leaf u) (Partitio a)      ) b z f g =
	let (fs, z1) = fmmDownSweep a (createArray 1 (Leaf Bot))
								(fmmUpSweep a (partitionNx1 ns b)
											(createArray (size ns) (Leaf Bot)))
								z
	in
		case fs of
			Node {[0]=f} _ = (u :*++: f) z1
			Leaf Bot       =             z1


*/