/*
	File: blop.dcl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Clean interface to Lapack and Blas
	See also: CamlFloat
	Date: 6.17.2003
	
	$Id: blop.dcl,v 1.8 2003/12/16 17:25:14 shiv Exp shiv $
*/

definition module blop

import StdEnv, complex


						/////////////////////////
						// Matrix Construction //
						/////////////////////////

// Basic matrix type. Not abstract to prevent uniqueness problems. Please
// treat it as an abstract type. In particular the "Dense" and "Alias" constructors
// might disappear in future versions.
:: Matrix e = Dense !.{#e} !Int !Int !Int !Int !Int !Int
			| Alias !Int !Int !Int !Int !Int !Int
:: MatrixReal :== Matrix Real						// Real matrices (technical)
:: MatrixComplex :== Matrix Complex					// Complex matrices (technical)
:: Transformer e :== (*(Matrix e) -> *(Matrix e))	// Destructive updating
:: Transform e = QofQL !{#e} !(Factor e) // private
			   | QofLQ !{#e} !(Factor e) // private
			   | Permutation !{#Int}	 // public
			   | Product ![! Transform e !]			// Unitary, etc.
:: Factor e											// L, U, etc.

machineEpsilon :: Real

// a @ (i,j) -> a_{i,j}
class (@) infix 9 m e :: !u:(m e) !(!Int, !Int) -> e
instance @ Matrix Real
instance @ Matrix Complex

// a @! (i,j) -> (a_{i,j}, a)
class (@!) infix 9 m e :: !u:(m e) !(!Int,!Int) -> (e, u:(m e))
instance @! Matrix Real
instance @! Matrix Complex

// noOfRows a -> m when a is an mxn matrix
class noOfRows m e :: !u:(m e) -> Int
instance noOfRows Matrix e

// noOfRows` a -> (m, a) when a is an mxn matrix
class noOfRows` m e :: !u:(m e) -> (Int, u:(m e))
instance noOfRows` Matrix e

// noOfCols a -> n when a is an mxn matrix
class noOfCols m e :: !u:(m e) -> Int
instance noOfCols Matrix e

// noOfCols` a -> (n, a) when a is an mxn matrix
class noOfCols` m e :: !u:(m e) -> (Int, u:(m e))
instance noOfCols` Matrix e

// For printing matrix
instance toString MatrixReal
instance toString MatrixComplex

// set a i j x -> a such that a_ij = x
class set m e :: !*(m e) !Int !Int !e -> *(m e)
instance set Matrix Real
instance set Matrix Complex

// ((i,j) <=. x) a == set a i j x
class (<=.) infix 0 m e :: !(!Int,!Int) !e !*(m e) -> *(m e)
instance <=. Matrix Real
instance <=. Matrix Complex

// ((i1 ... i2, j1 ... j2) <=$ x) a == copy x into a $ (i1 ... i2, j1 ... j2)
class (<=$) infix 0 e :: !(!Range,!Range) !(Matrix e) !*(Matrix e) -> *(Matrix e)
instance <=$ Real
instance <=$ Complex

// Constructing matrices
/*
class Consts m e | zero e & one e
where
	// consts x m n -> a such that a is mxn and a_ij = x
	consts :: !e !Int !Int -> *(m e)
	// rands xs m n -> (a, xs) such that a is mxn and a_ij = xs_ij
	rands :: [e] !Int !Int -> (*(m e), [e])
	// zeros m n -> a such that a is mxn and a_ij = zero
	zeros m n :== consts zero m n
	// ones m n -> a such that a is mxn and a_ij = one
	ones m n :== consts one m n
instance Consts Matrix Real
instance Consts Matrix Complex
*/

// consts x m n -> a such that a is mxn and a_ij = x
class consts e :: !e !Int !Int -> *(Matrix e)
instance consts Real
instance consts Complex

// rands xs m n -> (a, xs) such that a is mxn and a_ij = xs_ij
class rands e :: [e] !Int !Int -> (*(Matrix e), [e])
instance rands Real
instance rands Complex

// zeros m n -> a such that a is mxn and a_ij = zero
class zeros e :: !Int !Int -> *(Matrix e)
instance zeros Real
instance zeros Complex

// ones m n -> a such that a is mxn and a_ij = one
class ones e :: !Int !Int -> *(Matrix e)
instance ones Real
instance ones Complex


// generates list of random complex numbers. See rands above.
genRandComplex :: !Int !Int -> [Complex]
// genRandReal is in MersenneTwister which this calls

// fromFun fn m n -> a such that a is mxn and a_ij = fn i j
class fromFun m e :: !(Int Int -> e) !Int !Int -> *(m e)
instance fromFun Matrix Real
instance fromFun Matrix Complex

// fun2mat fn a -> b such a_ij <- b_{ij} = fn i j
class fun2mat e :: !(Int Int -> e) !*(Matrix e) -> *(Matrix e)
instance fun2mat Real
instance fun2mat Complex

// eye m n = Leading principal mxn submatrix of identity matrix
class eye e :: !Int !Int -> *(Matrix e)
instance eye Real
instance eye Complex

// toMatrix a -> b, such that b is a matrix representing data in a
class toMatrix a e :: !u:(a e) !Int !Int -> v:(Matrix e), [u <= v]
// toMatrix {a00, a10, a20, .., a01, a11, a21, ..} -> a in column major order
instance toMatrix {#} e | Array {#} e

// to2D {{a00, a01, ..}, {a10, a11, ..}, ..} -> {a00, a10, a20, .., a01, a11, a21, ..}
// converts row major 2D arrays to column major 1D array
to2D :: !{#{#e}} -> *{#e} | Array {#} e

// Types to denote index ranges for selecting submatrices. See ($) below.
:: Range = Pair !Int !Int | (<.>)
(...) infix 0 :: !Int !Int -> Range

// a $ (r, c) -> b such that b is a submatrix of a, sharing storage area
// Form 1:
// 		a $ (r1 ... r2, c1 ... c2) -> b such that b contains the r1-th to r2-th
//		rows of a, and the c1-th to c2-th columns of a
// Form 2:
//		a $ (r1 ... r2, <.>) -> b such that b contains the r1-th to r2-th rows of
//		a, and all the columns of a.
// Form 3:
//		a $ (<.>, c1 ... c2) -> b such that b contains all the rows of a, and the
//		c1-th to c2-th columns of a.
// Form 4:
//		a $ (<.>, <.>) -> b such that b contains all the rows and columns of a (= b)
($) infix 9 :: !u:(Matrix e) !(!Range, !Range) -> u:(Matrix e)

// a $! (r, c) -> (copy of a $ (r, c), a). No sharing.
class ($!) infix 9 e :: !u:(Matrix e) !(!Range, !Range) -> (*(Matrix e), u:(Matrix e))
instance $! Real
instance $! Complex

// a $# (r, c) -> (b, a) such that b is a pointer/alias to a $ (r, c). This call
// is DANGEROUS in that it enables you to by pass the normal uniqueness mechanism
// inside Transformer calls. Not only that, b has no remembrance of a. So it can
// be used inadvertently with some other matrix c. See sparse.icl for how it is
// used in Transformer calls.
($#) infix 9 :: !u:(Matrix e) !(!Range, !Range) -> (Matrix e, u:(Matrix e))

// partition1x2 n0 n1 a -> (a0, a1) being 1x2 block partitions of a
partition1x2 :: !Int !Int !u:(Matrix e) -> (Matrix e, Matrix e)

// partition2x1 m0 m1 a -> (a0, a1) being 2x1 block partitions of a
partition2x1 :: !Int !Int !u:(Matrix e) -> (Matrix e, Matrix e)

// partition2x2 n0 n1
//				m0	  -> (a00, a01, a10, a11) being 2x2 block partition of a
//				m1  a
partition2x2 :: !Int !Int !Int !Int !u:(Matrix e) ->
							(Matrix e, Matrix e, Matrix e, Matrix e)

// partition1xInf [n0, n1, n2, ..] a -> [a0, a1, a2, ..] being column partitions of a
partition1xInf :: ![Int] .(Matrix e) -> [Matrix e]

// partitionInfx1 [m0, m1, m2, ..] a -> [a0, a1, a2, ..] being row partitions of a
partitionInfx1 :: ![Int] .(Matrix e) -> [Matrix e]

// partitionNx1 {m0, m1, .., mr} a -> {a0, a1, .., ar} being row partitions of a
partitionNx1 :: !{#Int} !.(Matrix e) -> {Matrix e}

// partition1xN {n0, n1, .., nr} a -> {a0, a1, .., ar} being column partitions of a
partition1xN :: !{#Int} !.(Matrix e) -> {Matrix e}

//												n					n
// matrix2x1 m0 f0						   m0 [ a0 ]		  m0 [ f0 a0 ]
//			 m1 f1 a -> b such that if a = m1 [ a1 ] then b = m1 [ f1 a1 ]
// b and a share storage, and a has n columns
matrix2x1 :: !Int !(Transformer e)
			 !Int !(Transformer e) !*(Matrix e) -> *(Matrix e)

// matrix1x2 n0 n1							   n0 n1				  n0	  n1
//			 f0 f1 a -> b such that if a = m [ a0 a1 ] then b = m [ (f0 a0) (f1 a1) ]
// a and b share storage, and a has m rows
matrix1x2 :: !Int             !Int
			 !(Transformer e) !(Transformer e) !*(Matrix e) -> *(Matrix e)

// matrix2x2 n0   n1							  n0   n1
//		  m0 f00 f01						 m0 [ a00 a01 ]
//		  m1 f10 f11 a -> b such that if a = m1 [ a10 a11 ] then
// a and b share storage and b =  		n0 			n1
//								m0 [ (f00 a00) (f01 a01) ]
//								m1 [ (f10 a10) (f11 a11) ]
matrix2x2 :: !Int             !Int
		!Int !(Transformer e) !(Transformer e)
		!Int !(Transformer e) !(Transformer e) !*(Matrix e) -> *(Matrix e)

// matrix3x3 n0   n1  n2						      n0   n1  n2
//		  m0 f00 f01 f02						 m0 [ a00 a01 a02]
//		  m1 f10 f11 f12						 m1 [ a10 a11 a12]
//		  m2 f20 f21 f22 a -> b such that if a = m2 [ a20 a21 a22] then
// a and b share storage and b =  		n0 			n1        n2
//								m0 [ (f00 a00) (f01 a01) (f02 a02) ]
//								m1 [ (f10 a10) (f11 a11) (f12 a12) ]
//								m2 [ (f20 a20) (f21 a21) (f22 a22) ]
matrix3x3 :: !Int                    !Int           !Int
		!Int !(Transformer e) !(Transformer e) !(Transformer e)
		!Int !(Transformer e) !(Transformer e) !(Transformer e)
		!Int !(Transformer e) !(Transformer e) !(Transformer e) !*(Matrix e) ->
																		*(Matrix e)

// matrix1xInf [n0, n1, ..]							 	n0		 n1	   ..
//			   [f0, f1, ..] a -> b such that b = m [ (f0 a0), (f1 a1), ..]
// a and b share storage, and a =    n0  n1  ..
//								  m [a0, a1, ..]
matrix1xInf :: ![Int] ![Transformer e] !*(Matrix e) -> *(Matrix e)

// matrixInfx1 [m0, m1, ..] [f0, f1, ..] a -> b such that a and b share storage and
//		    n				 n
// a = m0 [a0] and b = m0 [(f0 a0)]
//	   m1 [a1]		   m1 [(f1 a1)]
//	   .  [ .]		   .  [   .   ]
//	   .  [ .]		   .  [   .   ]
matrixInfx1 :: ![Int] ![Transformer e] !*(Matrix e) -> *(Matrix e)

// matrix1xN {n0, n1, ..}							  n0	  n1	 ..
//			 {f0, f1, ..} a -> b such that b = m { (f0 a0), (f1 a1), ..}
// a and b share storage, and a =    n0  n1  ..
//								  m {a0, a1, ..}
matrix1xN :: !{#Int} !{Transformer e} !*(Matrix e) -> *(Matrix e)

// matrixNx1 {m0, m1, ..} {f0, f1, ..} a -> b such that a and b share storage and
//		    n				 n
// a = m0 {a0} and b = m0 {(f0 a0)}
//	   m1 {a1}		   m1 {(f1 a1)}
//	   .  { .}		   .  {   .   }
//	   .  { .}		   .  {   .   }
matrixNx1 :: !{#Int} !{Transformer e} !*(Matrix e) -> *(Matrix e)


// matInfx1 [m0, m1, ..] [f0, f1, ..] a [s0, s1, ..] -> (b, [s0', s1', ..])
// such that a and b share storage and
//		    n				 n
// a = m0 [a0] and [(b0, s0')] = m0 [(f0 a0 s0)]
//	   m1 [a1]	   [(b1, s1')]   m1 [(f1 a1 s1)]
//	   .  [ .]	   [    .    ]   .  [   .      ]
//	   .  [ .]	   [    .    ]   .  [   .      ]
matInfx1 :: ![Int] ![(*(Matrix e) b -> (*(Matrix e), b))] !*(Matrix e) ![b] ->
	(*(Matrix e), [b])

// matNx1 {m0, m1, ..} {f0, f1, ..} a {s0, s1, ..} -> (b, {s0', s1', ..})
// such that a and b share storage and
//		    n				 n
// a = m0 {a0} and {(b0, s0')} = m0 {(f0 a0 s0)}
//	   m1 {a1}	   {(b1, s1')}   m1 {(f1 a1 s1)}
//	   .  { .}	   {    .    }   .  {   .      }
//	   .  { .}	   {    .    }   .  {   .      }
matNx1 :: !{#Int} !{(*(Matrix e) b -> (*(Matrix e), b))} !*(Matrix e) !{b} ->
	(*(Matrix e), {b})

// ooo a -> b such that a = b. Nop.
ooo :: !u:(Matrix e) -> u:(Matrix e)



							////////////////////
							// Matrix Algebra //
							////////////////////

// (a :+: b) c -> c <- a + b
class (:+:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :+: Real
instance :+: Complex
instance + MatrixReal
instance + MatrixComplex

// a +$ b -> c where, unlike +, c can be unique
class (+$)  infixl 6 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)
instance +$ Real
instance +$ Complex

// (a :-: b) c -> c <- a - b
class (:-:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :-: Real
instance :-: Complex
instance - MatrixReal
instance - MatrixComplex

// a -$ b -> c where, unlike -, c can be unique
class (-$)  infixl 6 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)
instance -$ Real
instance -$ Complex

// (a :*: b) c -> c <- a * b
class (:*:) infix 0 e :: !(Matrix e) !(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :*: Real
instance :*: Complex
instance * MatrixReal
instance * MatrixComplex

// a *$ b -> c, where, unlike *, c can be unique
class (*$) infixl 7 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)
instance *$ Real
instance *$ Complex

// (a :~*: b) c -> c <- a^H * b
class (:~*:) infix 0 e :: !(Matrix e) !(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :~*: Real
instance :~*: Complex

// a ~* b -> c such that c = a^H * b
class (~*) infixl 7 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)
instance ~* Real
instance ~* Complex

// (a :*~: b) c -> c <- a * b^H
class (:*~:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :*~: Real
instance :*~: Complex

// a *~ b -> c such that c = a * b^H
class (*~) infixl 7 e :: !u:(Matrix e) !v:(Matrix e) -> *(Matrix e)
instance *~ Real
instance *~ Complex

// (a :*++: b) c -> c <- c + a * b
class (:*++:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :*++: Real
instance :*++: Complex

// (a :*--: b) c -> c <- c - a * b
class (:*--:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :*--: Real
instance :*--: Complex

// (a :~*++: b) c -> c <- c + a^H * b
class (:~*++:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :~*++: Real
instance :~*++: Complex

// (a :~*--: b) c -> c <- c - a^H * b
class (:~*--:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :~*--: Real
instance :~*--: Complex

// (a :*~++: b) c -> c <- c + a * b^H
class (:*~++:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :*~++: Real
instance :*~++: Complex

// (a :*~--: b) c -> c <- c - a * b^H
class (:*~--:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :*~--: Real
instance :*~--: Complex

// d :.*: a -> a such that a_ij = d_i * a_ij
class (:.*:) e :: !u:{#Real} !*(Matrix e) -> *(Matrix e)
instance :.*: Real
instance :.*: Complex

// a :*.: d -> a such that a_ij = a_ij * d_j
class (:*.:) e :: !*(Matrix e) !u:{#Real} -> *(Matrix e)
instance :*.: Real
instance :*.: Complex

// scale alpha a -> a such that a_ij = a_ij * alpha
class scale e :: !e !*(Matrix e) -> *(Matrix e)
instance scale Real
instance scale Complex


						//////////////////////
						// Matrix Inversion //
						//////////////////////

// a :\: b -> b such b = pseudoinverse a * b
class (:\:) e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :\: Real
instance :\: Complex

// a $\ b -> c such c = pseudoinverse a * b
class ($\) e :: !u:(Matrix e) !v:(Matrix e) -> *(Matrix e)
instance $\ Real
instance $\ Complex

// (a :~\: b) c -> c such c = pseudoinverse a^H * b
class (:~\:) e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)
instance :~\: Real
instance :~\: Complex

// a ~\ b -> c such c = pseudoinverse a^H * b
class (~\) e :: !u:(Matrix e) !v:(Matrix e) -> *(Matrix e)
instance ~\ Real
instance ~\ Complex

// Solve using PLU factorization. Not a pseudoInverse.
// pluSolve a b -> b = a^{-1} b. Both a & b destroyed
class pluSolve e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)
instance pluSolve Real
instance pluSolve Complex

// pluSolveT a b -> b = a^{-T} b. a is detroyed
class pluSolveT e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)
instance pluSolveT Real
instance pluSolveT Complex

// See also \* and ~\* in "Matrix Factorization" section


							//////////////////
							// Matrix Norms //
							//////////////////

// norm1` a -> (one norm of a, a)
class norm1` m e :: !u:(m e) -> (Real, u:(m e))
instance norm1` Matrix Real
instance norm1` Matrix Complex

// norm1 a -> one norm of a
class norm1 m e :: !u:(m e) -> Real
instance norm1 Matrix Real
instance norm1 Matrix Complex

// normF` a -> (Frobenius norm of a, a)
class normF` m e :: !u:(m e) -> (Real, u:(m e))
instance normF` Matrix Real
instance normF` Matrix Complex

// normF a -> (Frobenius norm of a, a)
class normF m e :: !u:(m e) -> Real
instance normF Matrix Real
instance normF Matrix Complex

// normInf` a -> (infinity norm of a, a)
class normInf` m e :: !u:(m e) -> (Real, u:(m e))
instance normInf` Matrix Real
instance normInf` Matrix Complex

// normInf a -> infinity norm of a
class normInf m e :: !u:(m e) -> Real
instance normInf Matrix Real
instance normInf Matrix Complex


						///////////////////////////
						// Matrix Factorizations //
						///////////////////////////

// ql a -> (q, l), a is destroyed
class ql e :: !*(Matrix e) -> (Transform e, Factor e)
instance ql Real
instance ql Complex

// lq a -> (l, q), a is destroyed
class lq e :: !*(Matrix e) -> (Factor e, Transform e)
instance lq Real
instance lq Complex

// plu a -> (p, l, u), a is destroyed
class plu e :: !*(Matrix e) -> (Transform e, Factor e, Factor e)
instance plu Real
instance plu Complex

// q @* a -> a such that a = q * a
class (@*) infix 0 e :: !(Transform e) !*(Matrix e) -> *(Matrix e)
instance @* Real
instance @* Complex

// a *@ q -> a such that a = a * q
class (*@) infix 0 e :: !*(Matrix e) !(Transform e) -> *(Matrix e)
instance *@ Real
instance *@ Complex

// q ~@* a -> a such that a = q^H * a
class (~@*) infix 0 e :: !(Transform e) !*(Matrix e) -> *(Matrix e)
instance ~@* Real
instance ~@* Complex

// a *@~ q -> a such that a = a * q^H
class (*@~) infix 0 e :: !*(Matrix e) !(Transform e) -> *(Matrix e)
instance *@~ Real
instance *@~ Complex

// l \* b -> b such that b = l^{-1} b
class (\*) infixr 7 e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)
instance \* Real
instance \* Complex

// l ~\* b -> b such that b = l^{-1} b
class (~\*) infixr 7 e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)
instance ~\* Real
instance ~\* Complex

// See also $\ and ~\ in "Matrix Inversion" section


							////////////////////
							// Matrix Copying //
							////////////////////

// copy a -> b, such b_ij = a_ij, and b is unique
class copy e :: !u:(Matrix e) -> *(Matrix e)
instance copy Real
instance copy Complex

// copyL l -> L, such l_ij = L_ij if j <= i, else L_ij = 0
// L is same shape as original matrix. Desirable?
class copyL e :: !u:(Factor e) -> *(Matrix e)
instance copyL Real
instance copyL Complex

// copyU u -> U, such u_ij = u_ij if j >= i, else U_ij = 0
// U is not same shape as original matrix. Desirable?
class copyU e :: !u:(Factor e) -> *(Matrix e)
instance copyU Real
instance copyU Complex

// iD a b -> b such that a_ij = b_ij
class iD e :: !u:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance iD Real
instance iD Complex

// idL l L -> L such that L_ij = l_ij if j <= i
class idL e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)
instance idL Real
instance idL Complex

// idU u U -> U such that U_ij = u_ij if j >= i
class idU e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)
instance idU Real
instance idU Complex

// transp` a b -> b such that b_ij = a_ji
class transp` e :: !u:(Matrix e) !*(Matrix e) -> *(Matrix e)
instance transp` Real
instance transp` Complex

// transp a -> b such that b_ij = a_ij
class transp e :: !u:(Matrix e) -> *(Matrix e)
instance transp Real
instance transp Complex

// Converts a dense matrix to type Factor (lower triangular). Does no checking.
toLowerTriangular :: !(Matrix e) -> Factor e


					////////////////////////////////////////
					// Singular Value Decomposition (SVD) //
					////////////////////////////////////////

// svd a -> (u, s, vt) such that a = u * s * vt and a is destroyed
class svd e :: !*(Matrix e) -> (*(Matrix e), *{#Real}, *(Matrix e))
instance svd Real
instance svd Complex

// svdL a -> (u, s) such that a = u * s * vt and a is destroyed
class svdL e :: !*(Matrix e) -> (*(Matrix e), *{#Real})
instance svdL Real
instance svdL Complex

// svdR a -> (s, vt) such that a = u * s * vt and a is destroyed
class svdR e :: !*(Matrix e) -> (*{#Real}, *(Matrix e))
instance svdR Real
instance svdR Complex

// singValues a -> s such that a = u * s * vt and a is destroyed
class singValues e :: !*(Matrix e) -> *{#Real}
instance singValues Real
instance singValues Complex


							/////////////////////////
							// Matrix Meta Classes //
							/////////////////////////

// Clean 1.* and 2.* allows a max of 32 entries in a class.
// Hence we need meta classes.
class MatrixAccess e | noOfRows, noOfCols, noOfRows`, noOfCols`, <=., @! Matrix e
class MatrixArithU e | :-:, :*:, :*++:, :*--:, :\:, zeros, iD, copy
					 , ql, lq, ~@*, *@~, @*, \*, copyL, rands, <=$, $! e
class MatrixArith  e | +$, -$, *$, ~*, *~ e

							
							////////////////////////////
							// Fast Matrix Algorithms //
							//  Overloaded Interface  //
							////////////////////////////

class fastMul s e :: !(s e) !(Matrix e) !*(Matrix e) -> *(Matrix e)
class fastSolve s e :: !(s e) !(Matrix e) !*(Matrix e) -> *(Matrix e)
//class superfastMul s e :: !(s e) !(s e) -> (s e)
//class superfastSolve s e :: !(s e) !(s e) -> (s e)
//class fastULV s t f e :: !(s e) -> (t e, f e, t e)
//class fastTransformL  t e :: !(t e) !*(Matrix e) -> *(Matrix e)
//class fastTransformLT t e :: !(t e) !*(Matrix e) -> *(Matrix e)
//class fastTransformR  t e :: !*(Matrix e) !(t e) -> *(Matrix e)
//class fastTransformRT t e :: !*(Matrix e) !(t e) -> *(Matrix e)
//class superfastTransformL  s t e :: !(t e) !(s e) -> (s e)
//class superfastTransformLT s t e :: !(t e) !(s e) -> (s e)
//class superfastTransformR  s t e :: !(s e) !(t e) -> (s e)
//class superfastTransformRT s t e :: !(s e) !(t e) -> (s e)
//class fastSubs  f e :: !(f e) !*(Matrix e) -> *(Matrix e)
//class fastSubsT f e :: !(f e) !*(Matrix e) -> *(Matrix e)
//class superfastSubs  s f e :: !(f e) !(s e) -> (s e)
//class superfastSubsT s f e :: !(f e) !(s e) -> (s e)
//class sMatrix2x1 s e :: !(s e) !(s e) -> (Transform e, s e)
//class sMatrix2x2 s e :: !(s e) !(s e) !(s e) !(s e)
//								-> (Transform e, s e, Transform e)
//class sTransp s e :: !(s e) -> (s e)

