/*
	File: hss.icl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Fast HSS algorithms
	See also: CamlFloat
	Date: 7.2.2003
	
	$Id: hss.icl,v 1.5 2003/07/19 03:40:35 shiv Exp shiv $
*/

implementation module hss

import StdEnv, complex, MersenneTwister, blop

// General purpose binary tree
:: Btree a b = Leaf !a | Node !b !.(Btree a b) !.(Btree a b)

:: HssLeafData e = { u :: !.(Matrix e), v :: !.(Matrix e), d :: !.(Matrix e) }
:: HssNodeData e = { nl :: !Int, 		nr :: !Int,
					 ml :: !Int,        mr :: !Int,
					 bu :: !(Matrix e), bl :: !(Matrix e),
					 rl :: !(Matrix e), rr :: !(Matrix e),
					 wl :: !(Matrix e), wr :: !(Matrix e)
				   }
:: HssS e :== Btree .(HssLeafData e) (HssNodeData e)
:: Hss e = Hss .(Btree .(HssLeafData e) (HssNodeData e))


copyHssS :: !u:(HssS e) -> *(HssS e) | copy e
copyHssS (Leaf a)     = Leaf {u = copy a.u, v = copy a.v, d = copy a.d}
copyHssS (Node a l r) = Node a (copyHssS l) (copyHssS r)

copyHss :: !u:(Hss e) -> *(Hss e) | copy e
copyHss (Hss hss) = Hss (copyHssS hss)

instance noOfRows Hss e | noOfRows Matrix e
where
	noOfRows (Hss (Leaf a))     = noOfRows a.d
	noOfRows (Hss (Node a l r)) = a.ml + a.mr

instance noOfCols Hss e | noOfCols Matrix e
where
	noOfCols (Hss (Leaf a))     = noOfCols a.d
	noOfCols (Hss (Node a l r)) = a.nl + a.nr

instance noOfRows` Hss e | noOfRows` Matrix e
where
	noOfRows` (Hss (Leaf a))
		#! (m, d) = noOfRows` a.d
		= (m, Hss (Leaf {a & d = d}))
	noOfRows` (Hss (Node a l r)) = (a.ml + a.mr, Hss (Node a l r))

instance noOfCols` Hss e | noOfCols` Matrix e
where
	noOfCols` (Hss (Leaf a))
		#! (n, d) = noOfCols` a.d
		= (n, Hss (Leaf {a & d = d}))
	noOfCols` (Hss (Node a l r)) = (a.nl + a.nr, Hss (Node a l r))


:: HssReal    :== HssS Real
:: HssComplex :== HssS Complex

instance fastMul Hss Real
where
	fastMul (Hss hss) b z = fastMulM hss b z

instance fastMul Hss Complex
where
	fastMul (Hss hss) b z = fastMulM hss b z

// fastMulM :: !(Hss e) !(Matrix e) !*(Matrix e) -> *(Matrix e) | MatrixArithU
//															    , MatrixArith
//															    , zeros e
fastMulM hss b z :== fastMulM hss b z
where			
	fastMulM (Leaf a) b z = (a.d :*: b) z
	fastMulM hss      b z = downSweep hss (upSweep hss b) (zeros 0 (noOfCols b)) z
	where
		//upSweep :: !(Hss e) !(Matrix e) ->
		//			Btree (Matrix e, Matrix e) (Matrix e) | MatrixArith e
		upSweep (Leaf a) b = Leaf (a.v ~* b, a.d *$ b)
		upSweep (Node a l r) b =
			let (bl, br) = partition2x1 a.nl a.nr b
				glTree = upSweep l bl
				grTree = upSweep r br
				gl = getNodeData glTree
				gr = getNodeData grTree
			in
				Node (a.wl ~* gl +$ a.wr ~* gr) glTree grTree
	
		//downSweep :: !(Hss e) !(Btree (Matrix e, Matrix e) (Matrix e))
		//			 !(Matrix e) !*(Matrix e) -> *(Matrix e) | MatrixArithU
		//			 										 , MatrixArith e
		downSweep (Leaf a) (Leaf (_, db)) f z = ((a.u :*++: f) o (iD db)) z
		downSweep (Node a lHss rHss) (Node _ lGtree rGtree) f z =
			matrix2x1 a.ml (downSweep lHss lGtree (a.bu *$ gr +$ a.rl *$ f))
					  a.mr (downSweep rHss rGtree (a.bl *$ gl +$ a.rr *$ f)) z
		where
			gl = getNodeData lGtree
			gr = getNodeData rGtree
	
		getNodeData (Leaf (a, _)) = a
		getNodeData (Node a _ _)  = a


:: LeafSplit = { m :: !Int, n :: !Int, k :: !Int, l :: !Int }
:: NodeSplit = { ku :: !Int, lv :: !Int }
:: HssSpine :== Btree LeafSplit NodeSplit

randHss :: [e] !HssSpine -> (Hss e, [e]) | rands e
randHss rnds spine =
	let (rnds1, hssS, _, _, _, _) = rec1 spine rnds
	in
		(Hss hssS, rnds1)
where
	rec1 (Node a lSpine rSpine) rnds
		#! (rnds, lHss, k0, l0, m0, n0) = rec1 lSpine rnds
		#! (rnds, rHss, k1, l1, m1, n1) = rec1 rSpine rnds
		#! (bu, rnds) = rands rnds k0 l1
		#! (bl, rnds) = rands rnds k1 l0
		#! (rl, rnds) = rands rnds k0 a.ku
		#! (rr, rnds) = rands rnds k1 a.ku
		#! (wl, rnds) = rands rnds l0 a.lv
		#! (wr, rnds) = rands rnds l1 a.lv
		= ( rnds
		  , Node { ml = m0, mr = m1, nl = n0, nr = n1
				 , bu = bu, bl = bl
				 , wl = wl, wr = wr
				 , rl = rl, rr = rr
				 }
				 lHss rHss
		  , a.ku
		  , a.lv
		  , m0 + m1
		  , n0 + n1
		  )
	rec1 (Leaf a) rnds
		#! (u, rnds) = rands rnds a.m a.k
		#! (v, rnds) = rands rnds a.n a.l
		#! (d, rnds) = rands rnds a.m a.n
		= ( rnds
		  , Leaf { u = u, v = v, d = d }
		  , a.k
		  , a.l
		  , a.m
		  , a.n
		  )


instance fastSolve Hss Real
where
	fastSolve (Hss hss) b x = fastSolveS hss b x
	
instance fastSolve Hss Complex
where
	fastSolve (Hss hss) b x = fastSolveS hss b x


:: WchiLeaf e = { w :: !(Transform e), chi :: !(Matrix e) }
:: WchiNode e = { w :: !(Transform e), chi :: !(Matrix e), nm0 :: !Int, nm1 :: !Int }

fastSolveS :: !(HssS e) !(Matrix e) !*(Matrix e) -> *(Matrix e) | MatrixArith
															    , MatrixArithU
															    , MatrixAccess e
fastSolveS (Leaf a) b x = (copy a.d) :\: (iD b x)
fastSolveS hss b x =
	let (wChiTree, _, _, _, _, _) = ulSweep hss b (zeros 0 nrhs)
	in
		vSweep hss wChiTree (zeros 0 nrhs) x
where
	nrhs = noOfCols b
	
	eliminateLeaf :: !(HssLeafData e) !(Matrix e) -> ( Btree (WchiLeaf e) (WchiNode e)
													 , Matrix e, Matrix e, Matrix e
													 , Matrix e, Matrix e
													 )
										| MatrixArith, MatrixArithU, MatrixAccess e
	eliminateLeaf a b
	| noOfCols a.u >= noOfRows a.u  = // Nothing to compress
		( Leaf {WchiLeaf | w = Product [!!], chi = zeros 0 nrhs }
		, zeros (noOfCols a.v) nrhs // g
		, b							// remaining b
		, a.u						// remaining u
		, a.v						// remaining v
		, a.d						// remaining d
		)
	| otherwise =					  // Compress
		let (q, l_u) = ql` (copy a.u)
			beta = q ~@* (copy b)
			(b0, b1) = partition2x1 (m - ku) ku beta
			ku = noOfCols a.u
			m = noOfRows a.u
			n = noOfCols a.d
			qd = q ~@* (copy a.d)
			(qd0, qd1) = partition2x1 (m - ku) ku qd
			(d00, w) = lq` (copy qd0)
			(d10, d11) = partition1x2 (m - ku) (n - m + ku) ((copy qd1) *@~ w)
			chi = d00 \* (copy b0)
			(v00, v01) = partition2x1 (m - ku) (n - m + ku) (w @* (copy a.v))
		in
			( Leaf {WchiLeaf | w = w, chi = chi }
			, v00 ~* chi
			, b1 -$ d10 *$ chi
			, (copyL l_u) $ (m - ku ... m-1,  <.>)
			, v01
			, d11
			)

	ulSweep :: !(HssS e) !(Matrix e) !(Matrix e) -> ( Btree (WchiLeaf e) (WchiNode e)
													, Matrix e, Matrix e, Matrix e
													, Matrix e, Matrix e
													)
										| MatrixArith, MatrixArithU, MatrixAccess e
	ulSweep (Leaf a) b f = eliminateLeaf a (b -$ a.u *$ f)
	ulSweep (Node a lHss rHss) b f =	// Merge, depth first
		let (b0, b1) = partition2x1 a.ml a.mr b
			(lWchiTree, gl, bl, ul, vl, dl) = ulSweep lHss b0 (              a.rl *$ f)
			(rWchiTree, gr, br, ur, vr, dr) = ulSweep rHss b1 (a.bl *$ gl +$ a.rr *$ f)
			m0 = noOfRows dl
			n0 = noOfCols dl
			m1 = noOfRows dr
			n1 = noOfCols dr
			d = matrix2x2   n0                       n1
						m0 (iD dl) (ul :*: (a.bu *~ vr))
						m1 (ur :*: (a.bl *~ vl)) (iD dr) (zeros (m0 + m1) (n0 + n1))
			u = matrix2x1 m0 (ul :*: a.rl)
						  m1 (ur :*: a.rr) (zeros (m0 + m1) (noOfCols a.rl))
			v = matrix2x1 n0 (vl :*: a.wl)
						  n1 (vr :*: a.wr) (zeros (n0 + n1) (noOfCols a.wl))
			bm = matrix2x1 m0 (bl :-: ul *$ a.bu *$ gr)
						   m1 (iD br                  ) (zeros (m0 + m1) nrhs)
			(Leaf wCh, g, bmr, umr, vmr, dmr) = eliminateLeaf {u = u, v = v, d = d} bm
		in
			( Node {w = wCh.WchiLeaf.w, chi = wCh.WchiLeaf.chi, nm0 = n0, nm1 = n1}
				   lWchiTree rWchiTree
			, g +$ a.wl ~* gl +$ a.wr ~* gr
			, bmr
			, umr
			, vmr
			, dmr
			)

	vSweep (Leaf _) (Leaf {WchiLeaf | w = w, chi = chi}) xp x
		=  w ~@* (matrix2x1 (noOfRows chi) (iD chi)
							(noOfRows xp ) (iD xp ) x)
	vSweep (Node a lHss rHss) (Node {WchiNode | w = w, chi = chi, nm0 = n0, nm1 = n1}
									lWchi rWchi) xp x
		= matrix2x1 a.nl (vSweep lHss lWchi xp0)
					a.nr (vSweep rHss rWchi xp1) x
	where
		mc = noOfRows chi
		mx = noOfRows xp
		chip = w ~@* (matrix2x1 mc (iD chi)
								mx (iD xp ) (zeros (mc + mx) nrhs))
		(xp0, xp1) = partition2x1 n0 n1 chip
	vSweep _ _ _ _ = abort "fastSolve:downSweep: Mismatched trees."


// Tests

hss1 :: Hss Real
hss1 = Hss (Node { nl = 3, nr = 2, ml = 4, mr = 5
				 , bu = consts 1.0 1 1, bl = consts 1.0 1 1
				 , rl = consts 0.0 1 0, rr = consts 1.0 1 0
				 , wl = consts 1.0 1 0, wr = consts 1.0 1 0
				 }
				 ( Leaf { u = consts 1.0 4 1, v = consts 1.0 3 1, d = consts 1.0 4 3 } )
				 ( Leaf { u = consts 1.0 5 1, v = consts 1.0 2 1, d = consts 1.0 5 2 } )
		   )


spine1 :: HssSpine
spine1 = Node { ku = 0, lv = 0 }
			  ( Leaf { k = 1, l = 1, m = 4, n = 3 } )
			  ( Leaf { k = 1, l = 1, m = 5, n = 2 } )

spine2 :: HssSpine
spine2 = Node { ku = 0, lv = 0 }
			  ( Leaf { k = 1, l = 2, m = 4, n = 3 } )
			  ( Leaf { k = 3, l = 4, m = 5, n = 6 } )

spine3 :: HssSpine
spine3 = Node { ku = 0, lv = 0 }
			  ( Node { ku = 2, lv = 3 }
			  		 ( Leaf { k = 1, l = 2, m = 3, n = 6 } )
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 5 } )
			  )
			  ( Node { ku = 3, lv = 2 }
			  		 ( Leaf { k = 2, l = 1, m = 5, n = 4 } )
			  		 ( Leaf { k = 1, l = 2, m = 6, n = 3 } )
			  )

spine4 :: HssSpine
spine4 = Node { ku = 0, lv = 0 }
			  ( Node { ku = 2, lv = 3 }
			  		 ( Node { ku = 6, lv = 5 }
			  				( Leaf { k = 1, l = 2, m = 4, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 6 } )
			  		 )
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 5 } )
			  )
			  ( Node { ku = 3, lv = 2 }
			  		 ( Leaf { k = 2, l = 1, m = 5, n = 4 } )
			  		 ( Node { ku = 5, lv = 6 }
			  				( Leaf { k = 1, l = 2, m = 4, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 6 } )
			  		 )
			  )

spine5 :: HssSpine
spine5 = Node { ku = 0, lv = 0 }
			  ( Leaf { k = 1, l = 1, m = 4, n = 4 } )
			  ( Leaf { k = 1, l = 1, m = 5, n = 5 } )

spine6 :: HssSpine
spine6 = Node { ku = 0, lv = 0 }
			  ( Node { ku = 2, lv = 3 }
			  		 ( Leaf { k = 1, l = 2, m = 3, n = 3 } )
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  )
			  ( Node { ku = 3, lv = 2 }
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  		 ( Leaf { k = 1, l = 2, m = 3, n = 3 } )
			  )

spine7 :: HssSpine
spine7 = Node { ku = 0, lv = 0 }
			  ( Node { ku = 2, lv = 3 }
			  		 ( Node { ku = 6, lv = 5 }
			  				( Leaf { k = 1, l = 2, m = 4, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 6 } )
			  		 )
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 5 } )
			  )
			  ( Node { ku = 3, lv = 2 }
			  		 ( Leaf { k = 2, l = 1, m = 5, n = 4 } )
			  		 ( Node { ku = 5, lv = 6 }
			  				( Leaf { k = 1, l = 2, m = 4, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 6 } )
			  		 )
			  )

spine8 :: HssSpine
spine8 = Node { ku = 0, lv = 0 }
			  ( Node { ku = 2, lv = 3 }
			  		 ( Node { ku = 6, lv = 5 }
			  				( Leaf { k = 1, l = 2, m = 3, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 5 } )
			  		 )
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  )
			  ( Node { ku = 3, lv = 2 }
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  		 ( Node { ku = 5, lv = 6 }
			  				( Leaf { k = 1, l = 2, m = 3, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 5 } )
			  		 )
			  )

spine9 :: HssSpine
spine9 = Node { ku = 0, lv = 0 }
			  ( Node { ku = 2, lv = 3 }
			  		 ( Node { ku = 6, lv = 5 }
			  				( Leaf { k = 1, l = 2, m = 3, n = 3 } )
			  				( Leaf { k = 3, l = 4, m = 5, n = 5 } )
			  		 )
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  )
			  ( Node { ku = 3, lv = 2 }
			  		 ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  		 ( Node { ku = 5, lv = 6 }
			  				( Node { ku = 2, lv = 3 }
			  		 			   ( Node { ku = 6, lv = 5 }
			  							  ( Leaf { k = 1, l = 2, m = 3, n = 3 } )
			  							  ( Leaf { k = 3, l = 4, m = 5, n = 5 } )
			  		 			   )
			  		 			   ( Leaf { k = 2, l = 1, m = 4, n = 4 } )
			  				)
			  				( Leaf { k = 3, l = 4, m = 5, n = 5 } )
			  		 )
			  )

tst1 =
	let m = noOfRows hss1
		n = noOfCols hss1
		(b, _) = eye n n
	in
		toString (fastMul hss1 (eye n n) (consts 0.0 m n))

tst2 =
	let (hss, _) = randHss (genRandReal 1) spine1
		m = noOfRows hss
		n = noOfCols hss
	in
		toString (fastMul hss (eye n n) (consts 0.0 m n))
		
tst3 =
	let (hss, _) = randHss (genRandReal 1) spine2
		m = noOfRows hss
		n = noOfCols hss
	in
		toString (fastMul hss (eye n n) (consts 0.0 m n))
		
tst4 =
	let (hss, _) = randHss (genRandReal 1) spine3
		m = noOfRows hss
		n = noOfCols hss
	in
		toString (fastMul hss (eye n n) (consts 0.0 m n))
		
tst5 =
	let (hss, _) = randHss (genRandReal 1) spine4
		m = noOfRows hss
		n = noOfCols hss
	in
		toString (fastMul hss (eye n n) (consts 0.0 m n))

test1 =
	let (hss=:(Hss hssS), rnds) = randHss (genRandReal 1) spine5
		(b, _) = rands rnds m nrhs
		m = noOfRows hss
		n = noOfCols hss
		nrhs = 1
		x = fastSolveS hssS b (zeros n nrhs)
	in
		norm1 (fastMul hss x (zeros m nrhs) -$ b)
		
test2 =
	let (hss=:(Hss hssS), rnds) = randHss (genRandReal 1) spine6
		(b, _) = rands rnds m nrhs
		m = noOfRows hss
		n = noOfCols hss
		nrhs = 1
		x = fastSolveS hssS b (zeros n nrhs)
	in
		norm1 (fastMul hss x (zeros m nrhs) -$ b)
		
test3 =
	let (hss=:(Hss hssS), rnds) = randHss (genRandReal 1) spine8
		(b, _) = rands rnds m nrhs
		m = noOfRows hss
		n = noOfCols hss
		nrhs = 1
		x = fastSolveS hssS b (zeros n nrhs)
	in
		norm1 (fastMul hss x (zeros m nrhs) -$ b)
		
test4 =
	let (hss=:(Hss hssS), rnds) = randHss (genRandReal 1) spine9
		(b, _) = rands rnds m nrhs
		m = noOfRows hss
		n = noOfCols hss
		nrhs = 1
		x = fastSolveS hssS b (zeros n nrhs)
	in
		norm1 (fastMul hss x (zeros m nrhs) -$ b)
		
Start = test4





