/*
	File: sparse.icl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Block sparse matrices
	See also: CamlFloat
	Date: 10.04.2003
	
	$Id$
*/

implementation module sparse

import StdEnv, StdOverloadedList, complex, blop, context, MersenneTwister, Debug

:: Sparse e = Sparse !.{Context (Matrix e)} !.{Context (Matrix e)}

skeleton :: !Int !Int -> *(Sparse e)
skeleton m n = Sparse (createArray m cempty) (createArray n cempty)

instance noOfRows Sparse e | noOfRows Matrix e
where
	noOfRows (Sparse rows _) = noOfRowsM rows

noOfRowsM rows :== sum [ case (choose rowi) of
	 						None = 0
	 						Some (_, a) = noOfRows a
	 					 \\ rowi <-: rows
	 				   ]

instance noOfCols Sparse e | noOfCols Matrix e
where
	noOfCols (Sparse _ cols) = noOfColsM cols

noOfColsM cols = sum [ case (choose colj) of
						None = 0
						Some (_, a) = noOfCols a
					   \\ colj <-: cols
				     ]

instance noOfRows` Sparse e | noOfRows Matrix e
where
	noOfRows` (Sparse rows cols)
		#! (n, rows) = usize rows
		#! (m, rows) = doSum n rows 0 0
		= (m, Sparse rows cols)
	where
		doSum n rows i acc
			| i >= n = (acc, rows)
			#! (rowi, rows) = uselect rows i
			= case (choose rowi) of
				None = doSum n rows (i+1) acc
				Some (_, a) = doSum n rows (i+1) (acc + noOfRows a)

instance noOfCols` Sparse e | noOfCols Matrix e
where
	noOfCols` (Sparse rows cols)
		#! (n, cols) = usize cols
		#! (m, cols) = doSum n cols 0 0
		= (m, Sparse rows cols)
	where
		doSum n cols i acc
			| i >= n = (acc, cols)
			#! (coli, cols) = uselect cols i
			= case (choose coli) of
				None = doSum n cols (i+1) acc
				Some (_, a) = doSum n cols (i+1) (acc + noOfCols a)


getBlock :: !(Sparse e) !Int !Int -> Option (Matrix e)
getBlock (Sparse rows _) i j = fst (cget rows.[i] j)

getBlock` :: !*(Sparse e) !Int !Int -> (Option (Matrix e), *(Sparse e))
getBlock` (Sparse rows cols) i j
	#! (rowi, rows) = uselect rows i
	#! (v, rowi) = cget rowi j
	= (v, Sparse {rows & [i] = rowi} cols)

setBlock :: !*(Sparse e) !Int !Int !(Option (Matrix e)) -> *(Sparse e)
setBlock (Sparse rows cols) i j v
	#! (rowi, rows) = uselect rows i
	#! (colj, cols) = uselect cols j
	= Sparse {rows & [i] = cset rowi j v} {cols & [j] = cset colj i v}

rowPartitions :: !(Sparse e) -> {#Int} | noOfRows Matrix e
rowPartitions (Sparse rows _) = {createArray m 0 & [i] = case (choose rows.[i]) of
															None = 0
															Some (_, a) = noOfRows a
								 \\ i <- [0, 1 .. m-1]
								}
where
	m = size rows

rowPartitions` :: !*(Sparse e) -> ({#Int}, *(Sparse e)) | noOfRows Matrix e
rowPartitions` (Sparse rows cols)
	#! (m, rows) = usize rows
	#! rparts = createArray m 0
	#! (rparts, rows) = getRowSizes rows m 0 rparts
	= (rparts, Sparse rows cols)
where
	getRowSizes rows m i rparts
		| i >= m = (rparts, rows)
		#! (rowi, rows) = uselect rows i
		= case (choose rowi) of
			None = getRowSizes rows m (i+1) {rparts & [i] = 0}
			Some (_, a) = getRowSizes rows m (i+1) {rparts & [i] = noOfRows a}
								
colPartitions :: !(Sparse e) -> {#Int} | noOfCols Matrix e
colPartitions (Sparse _ cols) = {createArray n 0 & [i] = case (choose cols.[i]) of
															None = 0
															Some (_, a) = noOfCols a
								 \\ i <- [0 .. n-1]
								}
where
	n = size cols

colPartitions` :: !*(Sparse e) -> ({#Int}, *(Sparse e)) | noOfCols Matrix e
colPartitions` (Sparse rows cols)
	#! (n, cols) = usize cols
	#! cparts = createArray n 0
	#! (cparts, cols) = getColSizes cols n 0 cparts
	= (cparts, Sparse rows cols)
where
	getColSizes cols n i cparts
		| i >= n = (cparts, cols)
		#! (coli, cols) = uselect cols i
		= case (choose coli) of
			None = getColSizes cols n (i+1) {cparts & [i] = 0}
			Some (_, a) = getColSizes cols n (i+1) {cparts & [i] = noOfCols a}
								

instance fastMul Sparse e | zero, fun2mat, MatrixAccess, MatrixArithU e
where
	fastMul sp=:(Sparse rows cols) x b =
		matrixNx1 ms {createArray m ooo & [i] = mulBlockRow rows.[i]
						\\ i <- [0 .. m-1] } b
	where
		ms = rowPartitions sp
		xs = partitionNx1 (colPartitions sp) x
		m = size ms
		mulBlockRow row bi = seqS [! aij :*++: xs.[j] \\ (j, aij) <|- toList row !]
								  (fun2mat (\i j -> zero) bi)
		seqS [!!] b = b
		seqS [!f0 : fs!] b = seqS fs (f0 b)


// Extract i-th row as a single matrix, along with its column indices
mergeRow :: !Int !*{Context (Matrix e)} ->
	(Matrix e, [!Int!], [Int], *{Context (Matrix e)}) | MatrixAccess
													  , MatrixArithU e
mergeRow i rows
	#! (ai, rows) = rows![i]
	= let (as, is, ns) = mergeR (toList ai)
	  in
	  	(as, is, ns, rows)
where
	mergeR as = let ns = [ noOfCols ai \\ (_, ai) <|- as ]
					m = let [! (_, a0) : _ !] = as in noOfRows a0
					n = sum ns
					ais = [ iD ai \\ (_, ai) <|- as ]
				in
					( matrix1xInf ns ais (zeros m n)
					, [! i \\ (i, _) <|- as !]
					, ns
					)

// Extract i-th and j-th row as a single matrix, along with their column indices
mergeTwoRows :: !Int !Int !*{Context (Matrix e)} ->
		(Matrix e, [!Int!], [Int], *{Context (Matrix e)}) | MatrixAccess
														  , MatrixArithU e
mergeTwoRows i j rows
	#! (ai, rows) = rows![i]
	#! (aj, rows) = rows![j]
	= let (as, is, ns) = merge2R (toList ai) (toList aj)
	  in
	  	(as, is, ns, rows)
where
	merge2R as bs = let (ns, is, a0s, a1s) = m2 as bs [] [!!] [] []
						n = sum ns
						m0 = let [! (_, a0) : _ !] = as in noOfRows a0
						m1 = let [! (_, b0) : _ !] = bs in noOfRows b0
					in
						( matrix2x1 m0 (matrix1xInf ns a0s)
									m1 (matrix1xInf ns a1s) (zeros (m0+m1) n)
						, is
						, ns
						)
	where
		// m2 :: ![!(Int,Matrix e)!] ![!(Int,Matrix e)!] ![Int] ![!Int!]
		//		![Transformer e] ![Transformer e] ->
		//		([Int], [!Int!], [Transformer e], [Transformer e]) 
		m2 [!!] [!!] ns is a0s a1s = ( reverse ns, Reverse is, reverse a0s
									 , reverse a1s
									 )
		m2 as=:[!!] [! (i,b) : bs !] ns is a0s a1s =
			m2 as bs [noOfCols b : ns] [! i:is !] [ooo : a0s] [iD b : a1s]
		m2 [! (i,a) : as !] bs=:[!!] ns is a0s a1s =
			m2 as bs [noOfCols a : ns] [! i:is !] [iD a : a0s] [ooo : a1s]
		m2 as=:[! (i,a) : ass !] bs=:[! (j,b) : bss !] ns is a0s a1s =
			if (i < j)
			   (m2 ass bs [noOfCols a : ns] [! i : is !] [iD a : a0s] [ooo : a1s])
			   (if (i > j)
			   	   (m2 as bss [noOfCols b : ns] [! j : is !]
			   			[ooo : a0s] [iD b : a1s])
			   	   (m2 ass bss [noOfCols a : ns] [! j : is !]
			   			[iD a : a0s] [iD b : a1s])
			   )

// insertRow i js as, inserts (as !! n) at column (js !! n) of the i-th row.
insertRow :: !Int ![!Int!] ![Matrix e] !*(Sparse e) -> *(Sparse e)
insertRow i [!!] [] sp = sp
insertRow i [! j : js !] [aj : as] sp = insertRow i js as
											(setBlock sp i j (Some aj))
insertRow i _ _ sp = abort "insertRow: mismatch"


// Merge the given row "as" below (and to) the i-th row.
mergeInsertRow :: !Int ![!Int!] ![Matrix e] !*(Sparse e) -> *(Sparse e) |
												MatrixAccess, MatrixArithU e
mergeInsertRow i js as sp=:(Sparse rows cols)
	| isEmpty as || i < 0 = sp
	= let (bsc, rows1) = rows![i]
		  bs = toList bsc
		  m1 = let [a0:_] = as in noOfRows a0
		  m0 = if (IsEmpty bs) 0 (let (_, b0) = Hd bs in noOfRows b0)
		  (bsn, jsn) = mir m0 bs m1 js as [] [!!]
	  in
	  	if (m1 > 0)
	  	   (insertRow i jsn bsn (Sparse rows1 cols))
	  	   (Sparse rows1 cols)
where
	mir _ [!!] _ [!!] [] bsn jsn = (bsn, jsn)
	mir m0 [!(jb,b):bs!] m1 [!!] [] bsn jsn =
		mir m0 bs m1 [!!] []
			[ matrix2x1 m0 (iD b)
						m1 ooo (zeros (m0+m1) (noOfCols b)) : bsn ]
			[!jb:jsn!]
	mir m0 [!!] m1 [!j:js!] [a:as] bsn jsn =
		mir m0 [!!] m1 js as
			[ matrix2x1 m0 ooo
						m1 (iD a) (zeros (m0+m1) (noOfCols a)) : bsn ]
			[!j:jsn!]
	mir m0 bss=:[!(jb,b):bs!] m1 jss=:[!j:js!] ass=:[a:as] bsn jsn =
		let n0 = noOfCols b
			n1 = noOfCols a
		in
			if (jb < j)
			   (mir m0 bs m1 jss ass [ matrix2x1 m0 (iD b)
			   									 m1 ooo (zeros (m0+m1) n0) : bsn ]
			   		[!jb:jsn!])
			   (if (jb > j)
			   	   (mir m0 bss m1 js as
			   	   		[ matrix2x1 m0 ooo
			   	   					m1 (iD a) (zeros (m0+m1) n1) : bsn ]
			   	   		[!j:jsn!])
			   	   (mir m0 bs m1 js as
			   	   		[ matrix2x1 m0 (iD b)
			   	   					m1 (iD a) (zeros (m0+m1) n0) : bsn ]
			   	   		[!j:jsn!])
			   )


// Copying ql routine.
qlC a = let (q, l) = ql (copy a) in (q, copyL l)

:: SparseTransform e = SpTransf !{#Int} ![!(!(Transform e), !Int, !Int)!]

// The q factor is correct only for square matrices. Otherwise we require a
// complicated block-circular-shift-down operation.
qlFactor :: !*(Sparse e) -> (SparseTransform e, Sparse e) | MatrixAccess
														  , MatrixArithU e
qlFactor (Sparse rows cols)
	#! (n, cols) = usize cols
	#! (m, rows) = usize rows
	#! (rparts, Sparse rows cols) = rowPartitions` (Sparse rows cols)
	#! (qs, rows, cols) = qlF (m-1) m rows n cols [!!]
	= (SpTransf rparts (Reverse qs), Sparse rows cols)
where
	qlF ipiv mr rows nc cols qs
		| ipiv < 0 || mr <= 0 || nc <= 0 = (qs, rows, cols)
		#! (an, cols) = cols![nc-1]
		= let ans = Filter (\(l,_) -> l <= ipiv) (toList an)
			  k = Length ans
		  in
		  	case k of
		  	0 = qlF ipiv mr rows (nc-1) cols qs
		  	1 = let (j,apiv) = let (j,apiv) = Hd ans
		  					   in
		  						if (j == ipiv) (j,apiv) (abort "qlF:wrong pivot")
		  			(mrows, is, ns, rows1) = mergeRow j rows
		  			(q, l) = qlC mrows
		  			nn = noOfCols apiv
		  			m = noOfRows mrows
		  			(m0, m1) = if (m >= nn)
		  						  (m-nn, nn)
		  						  (abort "qlF: fat diagonal block")
		  			(mr0, mr1) = partition2x1 m0 m1 l
		  			mr0s = partition1xInf ns mr0
		  			mr1s = partition1xInf ns mr1
		  			sp = insertRow j is mr1s (Sparse rows1 cols)
		  			(Sparse rows2 cols1) =
		  				if (m0 > 0)
		  				   (let sp1 = mergeInsertRow (j-1) is mr0s sp
		  				   	in
		  				   		if (j > 0)
		  						   (setBlock sp1 (j-1) (nc-1) None)
		  						   sp1
		  				   )
		  				   sp
		  		in
		  			qlF (ipiv-1) (mr-1) rows2 (nc-1) cols1 [!(q,j,j):qs!]
		  	_ = let (isa, _) = SplitAt (k-1) ans
		  			j = let (j, _) = Last ans
		  				in
		  					if (j == ipiv) j (abort "qlF:wrong pivot")
		  			(is, _) = Unzip isa
		  			(qsn, rowsn, colsn) = seqS [elim i j \\ i <|- Reverse is]
		  									  (qs, rows, cols)
		  		in
		  			qlF (ipiv-1) (mr-1) rowsn (nc-1) colsn qsn
	
	seqS [] s = s
	seqS [f0:fs] s = seqS fs (f0 s)

	elim i j (qs, rows, cols) =
		let (mrows, is, ns, rowsn) = mergeTwoRows i j rows
			(q, l) = qlC mrows
			m = noOfRows mrows
			(bs, rowsn2) = rowsn![j]
			(Some (jpiv, piv), _) = kadesi bs
			nn = noOfCols piv
			(m0, m1) = if (m >= nn)
						  (m-nn, nn)
						  (abort "elim: fat diagonal block")
			(mr0, mr1) = partition2x1 m0 m1 l
			mr0s = partition1xInf ns mr0
			mr1s = partition1xInf ns mr1
			sp = Sparse rowsn2 cols
			sp2 = insertRow i is mr0s sp
			sp3 = insertRow j is mr1s sp2
			(Sparse rowsn3 colsn) = setBlock sp3 i jpiv None
		in
			([! (q,i,j) : qs !] , rowsn3, colsn)


transfL :: !(SparseTransform e) !*(Matrix e) -> *(Matrix e) | MatrixAccess
															, MatrixArithU e
transfL (SpTransf rparts qs) a = trL (Reverse qs) a
where
	m = size rparts
	rc :: {#Int}
	rc = loop (createArray (m+1) 0) 1
	loop rc i
		| i > m = rc
		#! (rci, rc) = rc![i-1]
		= loop {rc & [i] = rci + rparts.[i-1]} (i+1)
	
	//trL :: ![!(!(Transform e), !Int, !Int)!] !*(Matrix e) -> *(Matrix e)
	//										| MatrixAccess, MatrixArithU e
	trL [!!] a = a
	trL [! (q,i,j) : qs !] a
		#! (ai, a) = a $! (rc.[i] ... rc.[i+1]-1, <.>)
		| i == j
			#! ai = q @* ai
			= trL qs (((rc.[i] ... rc.[i+1]-1, <.>) <=$ ai) a)
		#! m0 = rparts.[i]
		#! m1 = rparts.[j]
		#! (n, a) = noOfCols` a
		#! (aj, a) = a $! (rc.[j] ... rc.[j+1]-1, <.>)
		#! aij = q @* (matrix2x1 m0 (iD ai)
								 m1 (iD aj) (zeros (m0+m1) n))
		#! (ain, ajn) = partition2x1 m0 m1 aij
		#! a = ((rc.[i] ... rc.[i+1]-1, <.>) <=$ ain) a
		#! a = ((rc.[j] ... rc.[j+1]-1, <.>) <=$ ajn) a
		= trL qs a
	
transfLT :: !(SparseTransform e) !*(Matrix e) -> *(Matrix e) | MatrixAccess
															 , MatrixArithU e
transfLT (SpTransf rparts qs) a = trL qs a
where
	m = size rparts
	rc :: {#Int}
	rc = loop (createArray (m+1) 0) 1
	loop rc i
		| i > m = rc
		#! (rci, rc) = rc![i-1]
		= loop {rc & [i] = rci + rparts.[i-1]} (i+1)
	
	//trL :: ![!(!(Transform e), !Int, !Int)!] !*(Matrix e) -> *(Matrix e)
	//										| MatrixAccess, MatrixArithU e
	trL [!!] a = a
	trL [! (q,i,j) : qs !] a
		#! (ai, a) = a $! (rc.[i] ... rc.[i+1]-1, <.>)
		| i == j
			#! ai = q ~@* ai
			= trL qs (((rc.[i] ... rc.[i+1]-1, <.>) <=$ ai) a)
		#! m0 = rparts.[i]
		#! m1 = rparts.[j]
		#! (n, a) = noOfCols` a
		#! (aj, a) = a $! (rc.[j] ... rc.[j+1]-1, <.>)
		#! aij = q ~@* (matrix2x1 m0 (iD ai)
								  m1 (iD aj) (zeros (m0+m1) n))
		#! (ain, ajn) = partition2x1 m0 m1 aij
		#! a = ((rc.[i] ... rc.[i+1]-1, <.>) <=$ ain) a
		#! a = ((rc.[j] ... rc.[j+1]-1, <.>) <=$ ajn) a
		= trL qs a


fSubs :: !(Sparse e) *(Matrix e) -> *(Matrix e) | MatrixAccess, MatrixArithU e
fSubs sp=:(Sparse rows cols) b
	#! (m, b) = noOfRows` b
	#! (n, b) = noOfCols` b
	| m == noOfRowsM rows && n == noOfColsM cols
		= loop b 0
	= abort "fSubs: mismatch"
where
	mr = size rows
	nc = size cols
	cparts = colPartitions sp
	rc :: {#Int}
	rc = iloop (createArray (nc+1) 0) 1
	where
		iloop rc i
			| i > nc = rc
			#! (rci, rc) = rc![i-1]
			= iloop {rc & [i] = rci + cparts.[i-1]} (i+1)
	
	loop b i
		| i >= mr = b
		#! (bi, b) = b $! (rc.[i] ... rc.[i+1]-1, <.>)
		#! (bi, b) = subRow b (Filter (\(x,_) -> x < i) (toList rows.[i])) bi
		= case (getBlock sp i i) of
			None = abort "fSubs: Zero diagonal block"
			Some spii =
				loop (((rc.[i] ... rc.[i+1]-1, <.>) <=$
						(toLowerTriangular spii) \* bi) b)
					 (i+1)
	
	subRow b [!!] bi = (bi, b)
	subRow b [! (j,lij) : lijs !] bi
		#! (bj, b) = b $! (rc.[j] ... rc.[j+1]-1, <.>)
		= subRow b lijs ((lij :*--: bj) bi)


randBlkBanded :: ![e] !{#Int} !{#Int} !Int !Int -> (*(Sparse e), [e]) |
							MatrixAccess, MatrixArithU e
randBlkBanded es ms ns bl bu
	#! sp = skeleton m n
	= loop es sp 0 0
where
	m = size ms
	n = size ns
	loop es sp i j
		| i >= m = (sp, es)
		| j > i + bu || j >= n = loop es sp (i+1) (max 0 (i+1-bl))
		= let (a, esn) = rands es ms.[i] ns.[j]
		  in
		  	loop esn (setBlock sp i j (Some a)) i (j+1)


shallowCopy :: !(Sparse e) -> *(Sparse e)
shallowCopy (Sparse rows cols) =
	Sparse {createArray m rows.[0] & [i] = rows.[i] \\ i <- [1 .. m-1]}
		   {createArray n cols.[0] & [i] = cols.[i] \\ i <- [1 .. n-1]}
where
	m = size rows
	n = size cols


test1 = let m = 4
			n = 3
			k = 3
			sp = skeleton m n
			sp1 = setBlock sp 0 0 (Some (consts 1.0 k k))
			sp2 = setBlock sp1 1 1 (Some (consts (fromInt 2) k k))
			sp3 = setBlock sp2 2 2 (Some (consts 3.0 k k))
			sp4 = setBlock sp3 3 2 (Some (consts 4.0 k k))
			sp5 = setBlock sp4 3 0 (Some (consts 5.0 k k))
			x = eye (n * k) (n * k)
		in
			toString (fastMul sp5 x (zeros (m * k) (n * k)))

test2 = let k = 8
			m0 = 3
			n0 = 3
			ms = createArray k m0
			ns = createArray k n0
			m = sum [mi \\ mi <-: ms]
			n = sum [ni \\ ni <-: ns]
			bl = 2
			bu = 1
			(sp, _) = randBlkBanded (genRandReal 1) ms ns bl bu
			(qsp, lsp) = qlFactor (shallowCopy sp)
			lspc = fastMul lsp (eye n n) (zeros n n)
			qlspc = transfL qsp (copy lspc)
			spc = fastMul sp (eye n n) (zeros m n)
		in
			norm1 (qlspc - spc) <= fromInt n * norm1 spc * machineEpsilon

test3 = let sp = skeleton 2 2
			(a00, xs) = rands (genRandReal 1) 4 2
			(a10, xs1) = rands xs 4 2
			(a11, _) = rands xs1 4 2
			sp1 = setBlock sp 0 0 (Some a00)
			sp2 = setBlock sp1 1 0 (Some a10)
			sp3 = setBlock sp2 1 1 (Some a11)
			(_, lsp) = qlFactor sp3
		in
			toString (fastMul lsp (eye 4 4) (zeros 4 4))

test4 = let sp = skeleton 2 2
			(a00, xs) = rands (genRandReal 1) 2 2
			(a10, xs1) = rands xs 2 2
			(a11, _) = rands xs1 2 2
			sp2 = (setBlock (setBlock (setBlock sp 0 0 (Some a00)) 1 1 (Some a11))
							1 0 (Some a10))
			(qsp, lsp) = qlFactor (shallowCopy sp2)
			spc = fastMul sp2 (eye 4 4) (zeros 4 4)
		in
			norm1 (transfL qsp (fastMul lsp (eye 4 4) (zeros 4 4)) - spc)

test5 = let k = 8
			m0 = 3
			n0 = 3
			ms = createArray k m0
			ns = createArray k n0
			m = sum [mi \\ mi <-: ms]
			n = sum [ni \\ ni <-: ns]
			bl = 1
			bu = 1
			(sp, _) = randBlkBanded (genRandReal 1) ms ns bl bu
			(qsp, lsp) = qlFactor (shallowCopy sp)
			lspc = fastMul lsp (eye n n) (zeros n n)
			spc = fastMul sp (eye n n) (zeros m n)
			qtspc = transfLT qsp (copy spc)
		in
			norm1 (qtspc - lspc) <= fromInt n * norm1 lspc * machineEpsilon

test6 = let k = 8
			m0 = 3
			n0 = 3
			ms = createArray k m0
			ns = createArray k n0
			m = sum [mi \\ mi <-: ms]
			n = sum [ni \\ ni <-: ns]
			bl = 2
			bu = 1
			(sp, _) = randBlkBanded (genRandReal 1) ms ns bl bu
			(qsp, lsp) = qlFactor (shallowCopy sp)
			eyec = fSubs lsp (transfLT qsp (copy spc))
			spc = fastMul sp (eye n n) (zeros m n)
		in
			norm1 (eyec - (eye n n)) <= fromInt (2 * n) * norm1 spc * machineEpsilon


Start = test6
