/*
	File: blop.icl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Clean interface to Lapack and Blas
	See also: CamlFloat
	Date: 6.15.2003
	
	$Id: blop.icl,v 1.12 2004/04/19 22:25:58 shiv Exp shiv $
*/

// WARNING: All my_* calls are dangerous as they violate uniqueness attributes.

implementation module blop

import StdEnv, complex, MersenneTwister, Debug

// Present in "Clean System Files". Manual maintenance.
import code from "blop_c.o"

// {#{#}} is not an unboxed 2D array. Column-major order. Zero-based.
:: Matrix e = Dense !.{#e} !Int !Int !Int !Int !Int !Int
			| Alias !Int !Int !Int !Int !Int !Int
:: MatrixReal :== Matrix Real
:: MatrixComplex :== Matrix Complex

machineEpsilon :: Real
machineEpsilon =:
	let guess eps = if (eps + 1.0 == 1.0) eps (guess (eps / 2.0)) in
	guess (1.0 / 2.0)

// We need to heavily use macros to prevent conditional instances in the .dcl file.

arefM aa i j :== aref aa i j
where
	aref (Dense a m n sr lr sc lc) i j = a.[m * (sc + j) + (sr + i)]
	
class (@) infix 9 m e :: !u:(m e) !(!Int, !Int) -> e

instance @ Matrix Real
where
	(@) (Dense a m n sr lr sc lc) (i,j) = a.[m * (sc + j) + (sr + i)]

instance @ Matrix Complex
where
	(@) (Dense a m n sr lr sc lc) (i,j) = a.[m * (sc + j) + (sr + i)]


class (@!) infix 9 m e :: !u:(m e) !(!Int,!Int) -> (e, u:(m e))

instance @! Matrix Real
where
	(@!) (Dense a m n sr lr sc lc) (i,j)
		#! (e, a) = a![m * (sc + j) + (sr + i)]
		= (e, Dense a m n sr lr sc lc)

instance @! Matrix Complex
where
	(@!) (Dense a m n sr lr sc lc) (i,j)
		#! (e, a) = a![m * (sc + j) + (sr + i)]
		= (e, Dense a m n sr lr sc lc)


noOfRowsM a :== noOfRows a
where
	noOfRows (Dense _ _ _ sr lr _ _) = lr - sr + 1

noOfColsM a :== noOfCols a
where
	noOfCols (Dense _ _ _ _ _ sc lc) = lc - sc + 1

class noOfRows` m e :: !u:(m e) -> (Int, u:(m e))

instance noOfRows` Matrix e
where
	noOfRows` a=:(Dense _ _ _ sr lr _ _) = (lr - sr + 1, a)

class noOfRows m e :: !u:(m e) -> Int

instance noOfRows Matrix e
where
	noOfRows a = noOfRowsM a


class noOfCols` m e :: !u:(m e) -> (Int, u:(m e))

instance noOfCols` Matrix e
where
	noOfCols` a=:(Dense _ _ _ _ _ sc lc) = (lc - sc + 1, a)

class noOfCols m e :: !u:(m e) -> Int

instance noOfCols Matrix e
where
	noOfCols a = noOfColsM a


instance toString MatrixReal
where
	toString a = toStringM a

instance toString MatrixComplex
where
	toString a = toStringM a

toStringM a :== foldl (+++) "" (flatten [row i \\ i <- [0 .. m-1]])
where
	m = noOfRowsM a
	n = noOfColsM a
	cols = [0 .. n-1]
	row i = [toString (arefM a i j) +++ "  " \\ j <- cols] ++ ["\n"]


class set m e :: !*(m e) !Int !Int !e -> *(m e)

instance set Matrix Real
where
	set (Dense a m n sr lr sc lc) i j e
		#! a = {a & [m * (sc + j) + (sr + i)] = e}
		= Dense a m n sr lr sc lc

instance set Matrix Complex
where
	set (Dense a m n sr lr sc lc) i j e
		#! a = {a & [m * (sc + j) + (sr + i)] = e}
		= Dense a m n sr lr sc lc

/*
class Consts m e | zero e & one e
where
	consts :: !e !Int !Int -> *(m e)
	rands :: [e] !Int !Int -> (*(m e), [e])
	zeros m n :== consts zero m n
	ones m n :== consts one m n

zerosM m n :== Dense (createArray (m * n) zero) m n 0 (m-1) 0 (n-1)

instance Consts Matrix Real
where
	consts s m n = Dense (createArray (m * n) s) m n 0 (m-1) 0 (n-1)
	
	rands xs m n = ( Dense {createArray (m * n) 0.0 & [i] = xi \\
							i <- [0 .. m*n-1] & xi <- x1s
						   } m n 0 (m-1) 0 (n-1)
				   , x2s
				   )
	where
		(x1s, x2s) = splitAt (m * n) xs

instance Consts Matrix Complex
where
	consts s m n = Dense (createArray (m * n) s) m n 0 (m-1) 0 (n-1)
	
	rands xs m n = ( Dense {createArray (m * n) (0.0+:0.0) & [i] = xi \\
							i <- [0 .. m*n-1] & xi <- x1s
						   } m n 0 (m-1) 0 (n-1)
				   , x2s
				   )
	where
		(x1s, x2s) = splitAt (m * n) xs
*/

class consts e :: !e !Int !Int -> *(Matrix e)
instance consts Real
where
	consts s m n = Dense (createArray (m * n) s) m n 0 (m-1) 0 (n-1)
instance consts Complex
where
	consts s m n = Dense (createArray (m * n) s) m n 0 (m-1) 0 (n-1)

class zeros e :: !Int !Int -> *(Matrix e)
instance zeros Real
where
	zeros m n = consts 0.0 m n
instance zeros Complex
where
	zeros m n = consts (0.0+:0.0) m n

zerosM m n :== Dense (createArray (m * n) zero) m n 0 (m-1) 0 (n-1)


class ones e :: !Int !Int -> *(Matrix e)
instance ones Real
where
	ones m n = consts 1.0 m n
instance ones Complex
where
	ones m n = consts (1.0+:0.0) m n


class rands e :: [e] !Int !Int -> (*(Matrix e), [e])

instance rands Real
where
	rands xs m n = ( Dense {createArray (m * n) 0.0 & [i] = xi \\
							i <- [0 .. m*n-1] & xi <- x1s
						   } m n 0 (m-1) 0 (n-1)
				   , x2s
				   )
	where
		(x1s, x2s) = splitAt (m * n) xs
		
instance rands Complex
where
	rands xs m n = ( Dense {createArray (m * n) (0.0+:0.0) & [i] = xi \\
							i <- [0 .. m*n-1] & xi <- x1s
						   } m n 0 (m-1) 0 (n-1)
				   , x2s
				   )
	where
		(x1s, x2s) = splitAt (m * n) xs


normRM` s a :== normRM` s a
where
	normRM` s (Dense a m n sr lr sc lc)
		#! nrm = my_dlange s (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
			(createArray (lr - sr + 1) 0.0)
		= (nrm, Dense a m n sr lr sc lc)

normCM` s a :== normCM` s a
where
	normCM` s (Dense a m n sr lr sc lc)
		#! nrm = my_zlange s (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
			(createArray (lr - sr + 1) 0.0)
		= (nrm, Dense a m n sr lr sc lc)

normRM s a :== fst (normRM` s a)
normCM s a :== fst (normCM` s a)


class norm1` m e :: !u:(m e) -> (Real, u:(m e))

instance norm1` Matrix Real
where
	norm1` a = normRM` 0 a

instance norm1` Matrix Complex
where
	norm1` a = normCM` 0 a

class norm1 m e :: !u:(m e) -> Real

instance norm1 Matrix Real
where
	norm1 a = normRM 0 a

instance norm1 Matrix Complex
where
	norm1 a = normCM 0 a


class normF` m e :: !u:(m e) -> (Real, u:(m e))

instance normF` Matrix Real
where
	normF` a = normRM` 1 a

instance normF` Matrix Complex
where
	normF` a = normCM` 1 a

class normF m e :: !u:(m e) -> Real

instance normF Matrix Real
where
	normF a = normRM 1 a

instance normF Matrix Complex
where
	normF a = normCM 1 a


class normInf` m e :: !u:(m e) -> (Real, u:(m e))

instance normInf` Matrix Real
where
	normInf` a = normRM` 2 a

instance normInf` Matrix Complex
where
	normInf` a = normCM` 2 a

class normInf m e :: !u:(m e) -> Real

instance normInf Matrix Real
where
	normInf a = normRM 2 a

instance normInf Matrix Complex
where
	normInf a = normCM 2 a


class fromFun m e :: !(Int Int -> e) !Int !Int -> *(m e)

instance fromFun Matrix Real
where
	fromFun fn m n = Dense {createArray (m * n) 0.0 & [m * j + i] = fn i j \\
								j <- [0 .. n-1], i <- [0 .. m-1]}
							m n 0 (m-1) 0 (n-1)

instance fromFun Matrix Complex
where
	fromFun fn m n = Dense {createArray (m * n) zero & [m * j + i] = fn i j \\
								j <- [0 .. n-1], i <- [0 .. m-1]}
							m n 0 (m-1) 0 (n-1)


:: Transformer e :== (*(Matrix e) -> *(Matrix e))


class (:+:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :+: Real
where
	(:+:) a b c = addMacro a b c

instance :+: Complex
where
	(:+:) a b c = addMacro a b c

instance + MatrixReal
where
	(+) a b = addMacro a b (zerosM (noOfRowsM a) (noOfColsM a))

instance + MatrixComplex
where
	(+) a b = addMacro a b (zerosM (noOfRowsM a) (noOfColsM a))

addMacro a b c :== addInto a b c
where
	addInto aa=:(Dense a ma na sra lra sca lca) bb=:(Dense b mb nb srb lrb scb lcb)
				(Dense c mc nc src lrc scc lcc) =
		Dense {c & [mc * (scc + j) + (src + i)] =
				a.[ma * (sca + j) + (sra + i)] +
				b.[mb * (scb + j) + (srb + i)] \\ j <- [0 .. n-1], i <- [0 .. m-1]}
			mc nc src lrc scc lcc
		where
			m = let m = noOfRowsM aa in if (m == noOfRowsM bb) m (abort "+: Mismatch")
			n = let n = noOfColsM aa in if (n == noOfColsM bb) n (abort "+: Mismatch")


class (+$) infixl 6 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)

instance +$ Real
where
	(+$) a b = addMacro a b (zerosM (noOfRowsM a) (noOfColsM a))

instance +$ Complex
where
	(+$) a b = addMacro a b (zerosM (noOfRowsM a) (noOfColsM a))


class (:-:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :-: Real
where
	(:-:) a b c = subMacro a b c

instance :-: Complex
where
	(:-:) a b c = subMacro a b c


instance - MatrixReal
where
	(-) a b = subMacro a b (zerosM (noOfRowsM a) (noOfColsM b))

instance - MatrixComplex
where
	(-) a b = subMacro a b (zerosM (noOfRowsM a) (noOfColsM b))


class (-$) infixl 6 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)

instance -$ Real
where
	(-$) a b = subMacro a b (zerosM (noOfRowsM a) (noOfColsM a))

instance -$ Complex
where
	(-$) a b = subMacro a b (zerosM (noOfRowsM a) (noOfColsM a))


my_ddup :: !{#Real} -> !*{#Real}
my_ddup a =
	code {
		ccall ddup "A:A"
	}

my_zdup :: !{#Complex} -> !*{#Complex}
my_zdup a =
	code {
		ccall zdup "A:A"
	}

dmake2 :: !u:{#Real} -> (u:{#Real}, u:{#Real})
dmake2 a = (my_ddup a, my_ddup a)

zmake2 :: !u:{#Complex} -> (u:{#Complex}, u:{#Complex})
zmake2 a = (my_zdup a, my_zdup a)

my_dsub :: !Int !Int !{#Real} !Int !Int !{#Real} !Int !Int !{#Real} !Int !Int -> Int
my_dsub m n a dispa lda b dispb ldb c dispc ldc =
	code {
		ccall my_dsub "IIAIIAIIAII:I"
	}

my_zsub :: !Int !Int !{#Complex} !Int !Int !{#Complex} !Int !Int
			!{#Complex} !Int !Int -> Int
my_zsub m n a dispa lda b dispb ldb c dispc ldc =
	code {
		ccall my_zsub "IIAIIAIIAII:I"
	}

/*
instance - (Matrix e) | Array {#} e & zero e & - e
where
	(-) a b = subMacro a b (zerosM (noOfRowsM a) (noOfColsM b))
*/

subMacro a b c :== subInto a b c
where
	subInto aa=:(Dense a ma na sra lra sca lca) bb=:(Dense b mb nb srb lrb scb lcb)
				(Dense c mc nc src lrc scc lcc) =
		Dense {c & [mc * (scc + j) + (src + i)] =
				a.[ma * (sca + j) + (sra + i)] -
				b.[mb * (scb + j) + (srb + i)] \\ j <- [0 .. n-1], i <- [0 .. m-1]}
			mc nc src lrc scc lcc
		where
			m = let m = noOfRowsM aa in if (m == noOfRowsM bb) m (abort "-: Mismatch")
			n = let n = noOfColsM aa in if (n == noOfColsM bb) n (abort "-: Mismatch")

subMacroR a b c :== subInto a b c
where
	subInto aa=:(Dense a ma na sra lra sca lca) bb=:(Dense b mb nb srb lrb scb lcb)
				(Dense c mc nc src lrc scc lcc)
		#! s = my_dsub (lra - sra + 1) (lca - sca + 1) a (ma * sca + sra) ma
					b (mb * scb + srb) mb c (mc * scc + src) mc
		| s == s = Dense c mc nc src lrc scc lcc

subMacroC a b c :== subInto a b c
where
	subInto aa=:(Dense a ma na sra lra sca lca) bb=:(Dense b mb nb srb lrb scb lcb)
				(Dense c mc nc src lrc scc lcc)
		#! s = my_zsub (lra - sra + 1) (lca - sca + 1) a (ma * sca + sra) ma
					b (mb * scb + srb) mb c (mc * scc + src) mc
		| s == s = Dense c mc nc src lrc scc lcc


:: Range = Pair !Int !Int | (<.>)

(...) infix 0 :: !Int !Int -> Range
(...) i j = Pair i j

getRange :: !Range !Int !Int -> (Int,Int)
getRange (Pair i j) f l = (i, j)
getRange (<.>) f l = (f, l)

($) infix 9 :: !u:(Matrix e) !(!Range, !Range) -> u:(Matrix e)
($) (Dense a ma na sr lr sc lc) (rs, cs)
| 0 <= r1 && r2 >= r1-1 && 0 <= c1 && c2 >= c1 - 1 && r1 <= m && r2 < m
	&& c1 <= n && c2 < n = Dense a ma na (sr+r1) (sr+r2) (sc+c1) (sc+c2)
| otherwise = abort ("$:Invalid sub-matrix." +++ toString m)
where
	(r1, r2) = getRange rs 0 (lr - sr)
	(c1, c2) = getRange cs 0 (lc - sc)
	m = lr - sr + 1
	n = lc - sc + 1


($#) infix 9 :: !u:(Matrix e) !(!Range, !Range) -> (Matrix e, u:(Matrix e))
($#) (Dense a ma na sr lr sc lc) (rs, cs)
| 0 <= r1 && r2 >= r1-1 && 0 <= c1 && c2 >= c1 - 1 && r1 <= m && r2 < m
	&& c1 <= n && c2 < n
		= (Alias ma na (sr+r1) (sr+r2) (sc+c1) (sc+c2), Dense a ma na sr lr sc lc)
| otherwise = abort ("$#:Invalid sub-matrix." +++ toString m)
where
	(r1, r2) = getRange rs 0 (lr - sr)
	(c1, c2) = getRange cs 0 (lc - sc)
	m = lr - sr + 1
	n = lc - sc + 1


class ($!) infix 9 e :: !u:(Matrix e) !(!Range, !Range) -> (*(Matrix e), u:(Matrix e))

instance $! Real
where
	($!) (Dense a ma na sr lr sc lc) (rs, cs)
	| 0 <= r1 && r2 >= r1-1 && 0 <= c1 && c2 >= c1 - 1 && r1 <= m && r2 < m
													&& c1 <= n && c2 < n
		#! (b, a) = loop (createArray (mb * nb) 0.0) a 0 0
		= ( Dense b mb nb 0 (mb-1) 0 (nb-1)
		  , Dense a ma na sr lr sc lc
		  )
	| otherwise = abort "$!:Invalid sub-matrix."
	where
		(r1, r2) = getRange rs 0 (lr - sr)
		(c1, c2) = getRange cs 0 (lc - sc)
		m = lr - sr + 1
		n = lc - sc + 1
		mb = r2 - r1 + 1
		nb = c2 - c1 + 1
		loop b a i j
			| j >= nb = (b, a)
			| i >= mb = loop b a 0 (j+1)
			#! (aij, a) = a![ma * (sc + c1 + j) + sr + r1 + i]
			= loop {b & [mb * j + i] = aij} a (i+1) j

instance $! Complex
where
	($!) (Dense a ma na sr lr sc lc) (rs, cs)
	| 0 <= r1 && r2 >= r1-1 && 0 <= c1 && c2 >= c1 - 1 && r1 <= m && r2 < m
															&& c1 <= n && c2 < n
		#! (b, a) = loop (createArray (mb * nb) (0.0+:0.0)) a 0 0
		= ( Dense b mb nb 0 (mb-1) 0 (nb-1)
		  , Dense a ma na sr lr sc lc
		  )
	| otherwise = abort "$!:Invalid sub-matrix."
	where
		(r1, r2) = getRange rs 0 (lr - sr)
		(c1, c2) = getRange cs 0 (lc - sc)
		m = lr - sr + 1
		n = lc - sc + 1
		mb = r2 - r1 + 1
		nb = c2 - c1 + 1
		loop b a i j
			| j >= nb = (b, a)
			| i >= mb = loop b a 0 (j+1)
			#! (aij, a) = a![ma * (sc + c1 + j) + sr + r1 + i]
			= loop {b & [mb * j + i] = aij} a (i+1) j

partition1x2 :: !Int !Int !u:(Matrix e) -> (Matrix e, Matrix e)
partition1x2 n1 n2 a
| 0 <= n1 && 0 <= n2 && n1 + n2 == noOfColsM a = ( a $ (<.>,  0 ... n1-1   )
												 , a $ (<.>, n1 ... n1+n2-1)
												 )
| otherwise = abort "partition1x2:Invalid sub-matrix"

partition2x1 :: !Int !Int !u:(Matrix e) -> (Matrix e, Matrix e)
partition2x1 m1 m2 a
| 0 <= m1 && 0 <= m2 && m1 + m2 == noOfRowsM a = ( a $ ( 0 ... m1-1   , <.>)
												 , a $ (m1 ... m1+m2-1, <.>)
												 )
| otherwise = abort "partition2x1:Invalid sub-matrix"

class part2x1 e :: !Int !Int !u:(Matrix e) -> (u:(Matrix e), u:(Matrix e))

instance part2x1 Real
where
	part2x1 m1 m2 a=:(Dense aa ma na sr lr sc lc)
	| 0 <= m1 && 0 <= m2 && m1 + m2 == noOfRowsM a =
		( Dense (my_ddup aa) ma na sr1 sr2 sc lc
		, Dense (my_ddup aa) ma na sr3 sr4 sc lc
		)
	| otherwise = abort "part2x1:Invalid sub-matrix"


partition2x2 :: !Int !Int !Int !Int !u:(Matrix e) ->
							(Matrix e, Matrix e, Matrix e, Matrix e)
partition2x2 n0 n1 m0 m1 a
| 0 <= n0 && 0 <= n1 && 0 <= m0 && 0 <= m1 && n0 + n1 == noOfColsM a &&
	m0 + m1 == noOfRowsM a = ( a $ (r0, c0), a $ (r0, c1)
							 , a $ (r1, c0), a $ (r1, c1)
							 )
| otherwise = abort "partition2x2:Invalid sub-matrix."
where
	r0 =  0 ... m0-1
	r1 = m1 ... m0+m1-1
	c0 =  0 ... n0-1
	c1 = n0 ... n1+n0-1

partitionInfx1 :: ![Int] .(Matrix e) -> [Matrix e]
partitionInfx1 [] a
| noOfRowsM a == 0 = []
| otherwise = abort "partitionInfx1: Invalid partitioning."
partitionInfx1 [m0 : ms] a = [ a0 : partitionInfx1 ms a1 ]
where
	(a0, a1) = partition2x1 m0 (noOfRowsM a - m0) a

partition1xInf :: ![Int] .(Matrix e) -> [Matrix e]
partition1xInf [] a
| noOfColsM a == 0 = []
| otherwise = abort "partition1xInf: Invalid partitioning."
partition1xInf [n0 : ns] a = [ a0 : partition1xInf ns a1 ]
where
	(a0, a1) = partition1x2 n0 (noOfColsM a - n0) a

partitionNx1 :: !{#Int} !.(Matrix e) -> {Matrix e}
partitionNx1 ms a = {createArray (size ms) a & [i] = ai \\
						i  <- [0 .. ] &
						ai <- partitionInfx1 [mi \\ mi <-: ms] a
					}

partition1xN :: !{#Int} !.(Matrix e) -> {Matrix e}
partition1xN ns a = {createArray (size ns) a & [i] = ai \\
						i  <- [0 .. ] &
						ai <- partition1xInf [ni \\ ni <-: ns] a
					}


matrix2x1 :: !Int !(Transformer e)
			 !Int !(Transformer e) !*(Matrix e) -> *(Matrix e)
matrix2x1 m0 a0
		  m1 a1 (Dense a m n sr lr sc lc)
	| m0 + m1 <> lr - sr + 1 = abort "Invalid partitioning"
	#! (Dense a _ _ _ _ _ _) = a0 (Dense a m n sr      (sr+m0-1) sc lc)
	#! (Dense a _ _ _ _ _ _) = a1 (Dense a m n (sr+m0) lr        sc lc)
	= Dense a m n sr lr sc lc

matrix1x2 :: !Int             !Int
			 !(Transformer e) !(Transformer e) !*(Matrix e) -> *(Matrix e)
matrix1x2 n0 n1
		  a0 a1 (Dense a m n sr lr sc lc)
	| n0 + n1 <> lc - sc + 1 = abort "Invalid partitioning"
	#! (Dense a _ _ _ _ _ _) = a0 (Dense a m n sr lr sc      (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = a1 (Dense a m n sr lr (sc+n0) lc)
	= Dense a m n sr lr sc lc

matrix2x2 :: !Int                    !Int
		!Int !(Transformer e) !(Transformer e)
		!Int !(Transformer e) !(Transformer e) !*(Matrix e) -> *(Matrix e)
matrix2x2  n0  n1
		m0 a00 a01
		m1 a10 a11 (Dense a m n sr lr sc lc)
	| n0 + n1 <> lc - sc + 1 || m0 + m1 <> lr - sr + 1 = abort "Invalid partitioning"
	#! (Dense a _ _ _ _ _ _) = a00 (Dense a m n sr      (sr+m0-1) sc      (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = a10 (Dense a m n (sr+m0) lr        sc      (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = a01 (Dense a m n sr      (sr+m0-1) (sc+n0) lc)
	#! (Dense a _ _ _ _ _ _) = a11 (Dense a m n (sr+m0) lr        (sc+n0) lc)
	= Dense a m n sr lr sc lc

matrix3x3 :: !Int                    !Int           !Int
		!Int !(Transformer e) !(Transformer e) !(Transformer e)
		!Int !(Transformer e) !(Transformer e) !(Transformer e)
		!Int !(Transformer e) !(Transformer e) !(Transformer e) !*(Matrix e) -> *(Matrix e)
matrix3x3  n0  n1   n2
		m0 a00 a01 a02
		m1 a10 a11 a12
		m2 a20 a21 a22 (Dense a m n sr lr sc lc)
	| n0 + n1 + n2 <> lc - sc + 1 || m0 + m1 + m2 <> lr - sr + 1 = abort "Invalid partitioning"
	#! (Dense a _ _ _ _ _ _) = a00 (Dense a m n sr        (sr+m0-1)    sc      (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = a10 (Dense a m n (sr+m0)   (sr+m0+m1-1) sc      (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = a20 (Dense a m n (sr+m0+m1) lr          sc      (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = a01 (Dense a m n sr        (sr+m0-1)    (sc+n0) (sc+n0+n1-1))
	#! (Dense a _ _ _ _ _ _) = a11 (Dense a m n (sr+m0)   (sr+m0+m1-1) (sc+n0) (sc+n0+n1-1))
	#! (Dense a _ _ _ _ _ _) = a21 (Dense a m n (sr+m0+m1) lr          (sc+n0) (sc+n0+n1-1))
	#! (Dense a _ _ _ _ _ _) = a02 (Dense a m n sr        (sr+m0-1)    (sc+n0+n1) lc)
	#! (Dense a _ _ _ _ _ _) = a12 (Dense a m n (sr+m0)   (sr+m0+m1-1) (sc+n0+n1) lc)
	#! (Dense a _ _ _ _ _ _) = a22 (Dense a m n (sr+m0+m1) lr          (sc+n0+n1) lc)
	= Dense a m n sr lr sc lc

matrixInfx1 :: ![Int] ![Transformer e] !*(Matrix e) -> *(Matrix e)
matrixInfx1 [] [] a
	| noOfRowsM a == 0 = a
	| otherwise = abort "Invalid partitioning"
matrixInfx1 [m0 : ms] [a0 : as] (Dense a m n sr lr sc lc)
	| m0 > lr - sr + 1 = abort "Invalid partitioning"
	#! (Dense a _ _ _ _ _ _) = a0 (Dense a m n sr (sr+m0-1) sc lc)
	#! (Dense a _ _ _ _ _ _) = matrixInfx1 ms as (Dense a m n (sr+m0) lr sc lc)
	= Dense a m n sr lr sc lc

matrix1xInf :: ![Int] ![Transformer e] !*(Matrix e) -> *(Matrix e)
matrix1xInf [] [] a
	| noOfColsM a == 0 = a
	| otherwise = abort "Invalid partitioning"
matrix1xInf [n0 : ns] [a0 : as] (Dense a m n sr lr sc lc)
	| n0 > lc - sc + 1 = abort "Invalid partitioning"
	#! (Dense a _ _ _ _ _ _) = a0 (Dense a m n sr lr sc (sc+n0-1))
	#! (Dense a _ _ _ _ _ _) = matrix1xInf ns as (Dense a m n sr lr (sc+n0) lc)
	= Dense a m n sr lr sc lc

matrixNx1 :: !{#Int} !{Transformer e} !*(Matrix e) -> *(Matrix e)
matrixNx1 ms as a = matrixInfx1 [mi \\ mi <-: ms] [ai \\ ai <-: as] a

matrix1xN :: !{#Int} !{Transformer e} !*(Matrix e) -> *(Matrix e)
matrix1xN ns as a = matrix1xInf [ni \\ ni <-: ns] [ai \\ ai <-: as] a

matNx1 :: !{#Int} !{(*(Matrix e) b -> (*(Matrix e), b))} !*(Matrix e) !{b} ->
	(*(Matrix e), {b})
matNx1 ms as a ss =
	let (af, sss) = matInfx1 [mi \\ mi <-: ms] [ai \\ ai <-: as] a [si \\ si <-: ss]
	in
		case sss of
			[] = (af, {})
			[s : _] = (af, {createArray (size ms) s & [i] = si
							\\ i <- [0 ..] & si <- sss})

matInfx1 :: ![Int] ![(*(Matrix e) b -> (*(Matrix e), b))] !*(Matrix e) ![b] ->
	(*(Matrix e), [b])
matInfx1 [] [] a []
	| noOfRowsM a == 0 = (a, [])
	| otherwise = abort "matInfx1: Invalid partitioning"
matInfx1 [m0 : ms] [a0 : as] (Dense a m n sr lr sc lc) [s0 : ss]
	| m0 > lr - sr + 1 = abort "matInfx1: Mismatched sizes"
	#! (Dense a _ _ _ _ _ _, s0) = a0 (Dense a m n sr (sr+m0-1) sc lc) s0
	#! (Dense a _ _ _ _ _ _, ss) = matInfx1 ms as (Dense a m n (sr+m0) lr sc lc) ss
	= (Dense a m n sr lr sc lc, [s0 : ss])

class fun2mat e :: !(Int Int -> e) !*(Matrix e) -> *(Matrix e)
instance fun2mat Real
where
	fun2mat fn a = fun2matM fn a
instance fun2mat Complex
where
	fun2mat fn a = fun2matM fn a

fun2matM fn a :== fun2mat` fn a
where
	fun2mat` fn (Dense a m n sr lr sc lc) =
		Dense {a & [m * (sc + j) + (sr + i)] = fn i j \\ j <- [0 .. lc-sc]
													   , i <- [0 .. lr-sr]
			  }
			  m n sr lr sc lc


class eye e :: !Int !Int -> *(Matrix e)

instance eye Real
where
	eye m n = eyeM m n

instance eye Complex
where
	eye m n = eyeM m n

eyeM m n :== Dense {createArray (m * n) zero & [i * m + i] = one \\
					i <- [0 .. (min m n) - 1]
				   } m n 0 (m-1) 0 (n-1)

class toMatrix a e :: !u:(a e) !Int !Int -> v:(Matrix e), [u <= v]

instance toMatrix {#} e | Array {#} e
where
	toMatrix a m n
		| size a == m * n = Dense a m n 0 (m-1) 0 (n-1)
		| otherwise = abort "toMatrix: wrong dimensions"

to2D :: !{#{#e}} -> *{#e} | Array {#} e
to2D a2D = { createArray (m * n) a2D.[0].[0] & [m * j + i] = a2D.[i].[j] \\
			 j <- [0 .. n-1], i <- [0 .. m-1]
		   }
where
	m = size a2D
	n = size a2D.[0]

ooo :: !u:(Matrix e) -> u:(Matrix e)
ooo a = a

class (<=.) infix 0 m e :: !(!Int,!Int) !e !*(m e) -> *(m e)
instance <=. Matrix Real
where
	(<=.) (i, j) x (Dense a m n sr lr sc lc) =
		Dense {a & [m * (sc + j) + (sr + i)] = x} m n sr lr sc lc
instance <=. Matrix Complex
where
	(<=.) (i, j) x (Dense a m n sr lr sc lc) =
		Dense {a & [m * (sc + j) + (sr + i)] = x} m n sr lr sc lc


class (<=$) infix 0 e :: !(!Range,!Range) !(Matrix e) !*(Matrix e) -> *(Matrix e)

instance <=$ Real
where
	(<=$) (rs, cs) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_dlacpy 0 m n a (ma * sca + sra) ma
							b (mb * (scb + j) + srb + i) mb
		| s == s = Dense b mb nb srb lrb scb lcb
		= undef
	where
		(i,il) = getRange rs 0 (lrb - srb)
		(j,jl) = getRange cs 0 (lcb - scb)
		m = let m = lra - sra + 1
			in if (m == il - i + 1) m (abort "<=$:mismatch")
		n = let n = lca - sca + 1
			in if (n == jl - j + 1) n (abort "<=$:mismatch")

instance <=$ Complex
where
	(<=$) (rs, cs) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_zlacpy 0 m n a (ma * sca + sra) ma
							b (mb * (scb + j) + srb + i) mb
		| s == s = Dense b mb nb srb lrb scb lcb
		= undef
	where
		(i,il) = getRange rs 0 (lrb - srb)
		(j,jl) = getRange cs 0 (lcb - scb)
		m = let m = lra - sra + 1
			in if (m == il - i + 1) m (abort "<=$:mismatch")
		n = let n = lca - sca + 1
			in if (n == jl - j + 1) n (abort "<=$:mismatch")


// Cannot interface directly to Fortran routines because of pointers.
my_dlange :: !Int !Int !Int !u:{#Real} !Int !Int !{#Real} -> Real
my_dlange norm m n a disp lda work =
	code
	{
		ccall my_dlange "IIIAIIA:R"
	}

my_zlange :: !Int !Int !Int !u:{#Complex} !Int !Int !{#Real} -> Real
my_zlange norm m n a disp lda work =
	code
	{
		ccall my_zlange "IIIAIIA:R"
	}

/*
	CBLAS parameters. Keep in syncs with system cblas.h file.
   enum CBLAS_ORDER {CblasRowMajor=101, CblasColMajor=102 };
   enum CBLAS_TRANSPOSE {CblasNoTrans=111, CblasTrans=112, CblasConjTrans=113,
                         AtlasConj=114};
   enum CBLAS_UPLO  {CblasUpper=121, CblasLower=122};
   enum CBLAS_DIAG  {CblasNonUnit=131, CblasUnit=132};
   enum CBLAS_SIDE  {CblasLeft=141, CblasRight=142};
*/
CblasRowMajor  :== 101
CblasColMajor  :== 102
CblasNoTrans   :== 111
CblasTrans     :== 112
CblasConjTrans :== 113
AtlasConj      :== 114
CblasUpper     :== 121
CblasLower     :== 122
CblasNonUnit   :== 131
CblasUnit      :== 132
CblasLeft      :== 141
CblasRight     :== 142


class (:*:) infix 0 e :: !(Matrix e) !(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :*: Real
where
	(:*:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lcb - scb <> lcc - scc || lca - sca <> lrb - srb
			= abort ":*: : Non-conforming matrices"
		| lra < sra || lcb < scb = cc
		| lca < sca = scale 0.0 cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasNoTrans
			(lra - sra + 1) (lcb - scb + 1) (lca - sca + 1) 1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 0.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
		= abort "Huh?"

instance :*: Complex
where
	(:*:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lcb - scb <> lcc - scc || lca - sca <> lrb - srb
			= abort ":*: : Non-conforming matrices"
		| lra < sra || lcb < scb = cc
		| lca < sca = scale (0.0+:0.0) cc
		#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasNoTrans
			(lra - sra + 1) (lcb - scb + 1) (lca - sca + 1) {1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {0.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
		= abort "Huh?"

instance * MatrixReal
where
	(*) a b = (a :*: b) (zeros (noOfRowsM a) (noOfColsM b))
		
instance * MatrixComplex
where
	(*) a b = (a :*: b) (zeros (noOfRowsM a) (noOfColsM b))
		

class (*$) infixl 7 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)

instance *$ Real
where
	(*$) a b = (a :*: b) (zeros (noOfRowsM a) (noOfColsM b))

instance *$ Complex
where
	(*$) a b = (a :*: b) (zeros (noOfRowsM a) (noOfColsM b))


class (:~*:) infix 0 e :: !(Matrix e) !(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :~*: Real
where
	(:~*:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lca - sca <> lrc - src || lcb - scb <> lcc - scc || lra - sra <> lrb - srb
			= abort ":~*: : Non-conforming matrices"
		| lca < sca || lcb < scb = cc
		| lra < sra = scale 0.0 cc
		#! s = my_cblas_dgemm CblasColMajor CblasTrans CblasNoTrans
			(lca - sca + 1) (lcb - scb + 1) (lra - sra + 1) 1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 0.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
		= abort "Huh?"

instance :~*: Complex
where
	(:~*:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lca - sca <> lrc - src || lcb - scb <> lcc - scc || lra - sra <> lrb - srb
			= abort ":~*: : Non-conforming matrices"
		| lca < sca || lcb < scb = cc
		| lra < sra = scale (0.0 +: 0.0) cc
		#! s = my_cblas_zgemm CblasColMajor CblasConjTrans CblasNoTrans
			(lca - sca + 1) (lcb - scb + 1) (lra - sra + 1) {1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {0.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
		= abort "Huh?"

class (~*) infixl 7 e :: !(Matrix e) !(Matrix e) -> *(Matrix e)

instance ~* Real
where
	(~*) a b = (a :~*: b) (zeros (noOfColsM a) (noOfColsM b))

instance ~* Complex
where
	(~*) a b = (a :~*: b) (zeros (noOfColsM a) (noOfColsM b))


class (:*~:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :*~: Real
where
	(:*~:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb
			= abort "Non-conforming matrices"
		| lra < sra || lrb < srb = cc
		| lca < sca = scale 0.0 cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) 1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 0.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
		= abort "Huh?"

instance :*~: Complex
where
	(:*~:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb
			= abort "Non-conforming matrices"
		| lra < sra || lrb < srb = cc
		| lca < sca = scale (0.0 +: 0.0) cc
		#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasConjTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) {1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {0.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
		= abort "Huh?"

class (*~) infixl 7 e :: !u:(Matrix e) !v:(Matrix e) -> *(Matrix e)

instance *~ Real
where
	(*~) a b = (a :*~: b) (zeros (noOfRowsM a) (noOfRowsM b))

instance *~ Complex
where
	(*~) a b = (a :*~: b) (zeros (noOfRowsM a) (noOfRowsM b))


class (:*++:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :*++: Real
where
	(:*++:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lcb - scb <> lcc - scc || lca - sca <> lrb - srb
			= abort "Non-conforming matrices"
		| lra < sra || lcb < scb || lca < sca = cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasNoTrans
			(lra - sra + 1) (lcb - scb + 1) (lca - sca + 1) 1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)

instance :*++: Complex
where
	(:*++:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lcb - scb <> lcc - scc || lca - sca <> lrb - srb
			= abort "Non-conforming matrices"
		| lra < sra || lcb < scb || lca < sca = cc
		#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasNoTrans
			(lra - sra + 1) (lcb - scb + 1) (lca - sca + 1) {1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {1.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)


class (:*--:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :*--: Real
where
	(:*--:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lcb - scb <> lcc - scc || lca - sca <> lrb - srb
			= abort "Non-conforming matrices"
		| lra < sra || lcb < scb || lca < sca = cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasNoTrans
			(lra - sra + 1) (lcb - scb + 1) (lca - sca + 1) -1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)

instance :*--: Complex
where
	(:*--:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lcb - scb <> lcc - scc || lca - sca <> lrb - srb
			= abort "Non-conforming matrices"
		| lra < sra || lcb < scb || lca < sca = cc
		#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasNoTrans
			(lra - sra + 1) (lcb - scb + 1) (lca - sca + 1) {-1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {1.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)


class (:~*++:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :~*++: Real
where
	(:~*++:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lca - sca <> lrc - src || lcb - scb <> lcc - scc || lra - sra <> lrb - srb
			= abort "Non-conforming matrices"
		| lca < sca || lcb < scb || lra < sra = cc
		#! s = my_cblas_dgemm CblasColMajor CblasTrans CblasNoTrans
			(lca - sca + 1) (lcb - scb + 1) (lra - sra + 1) 1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)

instance :~*++: Complex
where
	(:~*++:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lca - sca <> lrc - src || lcb - scb <> lcc - scc || lra - sra <> lrb - srb
			= abort "Non-conforming matrices"
		| lca < sca || lcb < scb || lra < sra = cc
		#! s = my_cblas_zgemm CblasColMajor CblasConjTrans CblasNoTrans
			(lca - sca + 1) (lcb - scb + 1) (lra - sra + 1) {1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {1.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)


class (:~*--:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :~*--: Real
where
	(:~*--:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lca - sca <> lrc - src || lcb - scb <> lcc - scc || lra - sra <> lrb - srb
			= abort "Non-conforming matrices"
		| lca < sca || lcb < scb || lra < sra = cc
		#! s = my_cblas_dgemm CblasColMajor CblasTrans CblasNoTrans
			(lca - sca + 1) (lcb - scb + 1) (lra - sra + 1) -1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)

instance :~*--: Complex
where
	(:~*--:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lca - sca <> lrc - src || lcb - scb <> lcc - scc || lra - sra <> lrb - srb
			= abort "Non-conforming matrices"
		| lca < sca || lcb < scb || lra < sra = cc
		#! s = my_cblas_zgemm CblasColMajor CblasConjTrans CblasNoTrans
			(lca - sca + 1) (lcb - scb + 1) (lra - sra + 1) {-1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {1.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)


class (:*~++:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :*~++: Real
where
	(:*~++:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb
			= abort "Non-conforming matrices"
		| lra < sra || lrb < srb || lca < sca = cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) 1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)

instance :*~++: Complex
where
	(:*~++:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb
			= abort "Non-conforming matrices"
		| lra < sra || lrb < srb || lca < sca = cc
		#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasConjTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) {1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {1.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)


class (:*~--:) infix 0 e :: !u:(Matrix e) !v:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :*~--: Real
where
	(:*~--:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb
			= abort ":*~--:: Non-conforming matrices"
		| lra < sra || lrb < srb || lca < sca = cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) -1.0
			a (ma * sca + sra) ma b (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)
	(:*~--:) (Alias ma na sra lra sca lca) (Alias mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb ||
		  ma <> mc || mb <> mc || na <> nc || nb <> nc
		  	= abort ":*~--:: Non-conforming matrices or alias problem"
		| lra < sra || lrb < srb || lca < sca = cc
		#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) -1.0
			c (ma * sca + sra) ma c (mb * scb + srb) mb 1.0 c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)

instance :*~--: Complex
where
	(:*~--:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
			cc=:(Dense c mc nc src lrc scc lcc)
		| lra - sra <> lrc - src || lrb - srb <> lcc - scc || lca - sca <> lcb - scb
			= abort "Non-conforming matrices"
		| lra < sra || lrb < srb || lca < sca = cc
		#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasConjTrans
			(lra - sra + 1) (lrb - srb + 1) (lca - sca + 1) {-1.0 +: 0.0}
			a (ma * sca + sra) ma b (mb * scb + srb) mb {1.0 +: 0.0}
			c (mc * scc + src) mc
		| s == s = (Dense c mc nc src lrc scc lcc)


my_cblas_dgemm :: !Int !Int !Int !Int !Int !Int !Real !{#Real} !Int !Int
					!{#Real} !Int !Int !Real !{#Real} !Int !Int -> Int
my_cblas_dgemm order transA transB m n k alpha a dispa lda b dispb ldb beta
						c dispc ldc =
	code {
		ccall my_cblas_dgemm "IIIIIIRAIIAIIRAII:I"
	}

// Complex numbers are passed as 1x1 arrays of complex numbers.
my_cblas_zgemm :: !Int !Int !Int !Int !Int !Int !{#Complex} !{#Complex} !Int !Int
					!{#Complex} !Int !Int !{#Complex} !{#Complex} !Int !Int -> Int
my_cblas_zgemm order transA transB m n k alpha a dispa lda b dispb ldb beta
						c dispc ldc =
	code {
		ccall my_cblas_zgemm "IIIIIIAAIIAIIAAII:I"
	}


// Factorization routines

:: Transform e = QofQL !{#e} !(Factor e)
			   | QofLQ !{#e} !(Factor e)
			   | Permutation !{#Int}
			   | Product ![! Transform e !]

:: Factor e = LofQL !{#e} !Int !Int !Int !Int !Int !Int
			| LofLQ !{#e} !Int !Int !Int !Int !Int !Int
			| LofPLU !{#e} !Int !Int !Int !Int !Int !Int
			| UofPLU !{#e} !Int !Int !Int !Int !Int !Int


toLowerTriangular :: !(Matrix e) -> Factor e
toLowerTriangular (Dense a ma na sr lr sc lc) = LofQL a ma na sr lr sc lc


class ql e :: !*(Matrix e) -> (Transform e, Factor e)

instance ql Real
where
	ql (Dense a m n sr lr sc lc)
		| lc < sc || lr < sr = (Product [!!], LofQL a m n sr lr sc lc)
		#! work = createArray 1 0.0
		#! tau = createArray (min (lr - sr + 1) (lc - sc + 1)) 0.0
		#! info = my_dgeqlf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work -1
		#! lwork = fromReal work.[0]
		| info == 0 && lwork >= max 1 (lc - sc + 1)
			#! work = createArray lwork 0.0
			#! info = my_dgeqlf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work lwork
			| info == 0 =
				let aa = LofQL a m n sr lr sc lc in (QofQL tau aa, aa)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)

instance ql Complex
where
	ql (Dense a m n sr lr sc lc)
		| lc < sc || lr < sr = (Product [!!], LofQL a m n sr lr sc lc)
		#! work = createArray 1 (0.0 +: 0.0)
		#! tau = createArray (min (lr - sr + 1) (lc - sc + 1)) (0.0 +: 0.0)
		#! info = my_zgeqlf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lc - sc + 1)
			#! work = createArray lwork (0.0 +: 0.0)
			#! info = my_zgeqlf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work lwork
			| info == 0 =
				let aa = LofQL a m n sr lr sc lc in (QofQL tau aa, aa)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)


class (@*) infix 0 e :: !(Transform e) !*(Matrix e) -> *(Matrix e)

instance @* Real
where
	(@*) (QofQL tau (LofQL a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lrb - srb = abort "@*: Mismatch"
		#! work = createArray 1 0.0
		| lr - sr >= lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_dormql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr)
								m tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal work.[0] >= max 1 (lcb - scb + 1)
				#! lwork = fromReal work.[0]
				#! info = my_dormql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork 0.0) lwork
				| info == 0 = (Dense b mb nb srb lrb scb lcb)
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_dormql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m
							tau b (mb * scb + srb) mb work -1
		| info == 0 && fromReal work.[0] >= max 1 (lcb - scb + 1)
			#! lwork = fromReal work.[0]
			#! info = my_dormql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m
						tau b (mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(@*) (QofLQ tau (LofLQ a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lrb - srb = abort "@*: Mismatch"
		#! work = createArray 1 0.0
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_dormlq 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal work.[0]
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_dormlq 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(@*) (Permutation ms) a = permuteLr ms a
	(@*) (Product ts) a = apply ts a
	where
		apply [!!] a = a
		apply [!t:ts!] a = t @* (apply ts a)

instance @* Complex
where
	(@*) (QofQL tau (LofQL a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lrb - srb = abort "@*: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		| lr - sr > lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_zunmql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
				#! lwork = fromReal (realPart work.[0])
				#! info = my_zunmql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
								lwork
				| info == 0 = (Dense b mb nb srb lrb scb lcb)
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_zunmql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m
							tau b (mb * scb + srb) mb work -1
		| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zunmql 0 0 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m
								tau b (mb * scb + srb) mb
								(createArray lwork (0.0 +: 0.0)) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(@*) (QofLQ tau (LofLQ a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lrb - srb = abort "@*: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_zunmlq 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_zunmlq 0 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
							lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(@*) (Permutation ms) a = permuteLc ms a
	(@*) (Product ts) a = apply ts a
	where
		apply [!!] a = a
		apply [!t:ts!] a = t @* (apply ts a)


class (*@) infix 0 e :: !*(Matrix e) !(Transform e) -> *(Matrix e)

instance *@ Real
where
	(*@) (Dense b mb nb srb lrb scb lcb) (QofQL tau (LofQL a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lcb - scb = abort "*@: Mismatch"
		#! work = createArray 1 0.0
		| lr - sr >= lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_dormql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal work.[0] >= max 1 (lrb - srb + 1)
				#! lwork = fromReal work.[0]
				#! info = my_dormql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a
									(m * sc + sr) m tau b (mb * scb + srb) mb
									(createArray lwork 0.0) lwork
				| info == 0 = Dense b mb nb srb lrb scb lcb
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_dormql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m
							tau b (mb * scb + srb) mb work -1
		| info == 0 && fromReal work.[0] >= max 1 (lcb - scb + 1)
			#! lwork = fromReal work.[0]
			#! info = my_dormql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m tau b
								(mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@) (Dense b mb nb srb lrb scb lcb) (QofLQ tau (LofLQ a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lcb - scb = abort "*@: Mismatch"
		#! work = createArray 1 0.0
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_dormlq 1 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal work.[0]
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_dormlq 1 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@) a (Permutation ns) = permuteRr a (invPermute ns)
	(*@) a (Product ts) = apply a ts
	where
		apply a [!!]     = a
		apply a [!t:ts!] = apply (a *@ t) ts

instance *@ Complex
where
	(*@) (Dense b mb nb srb lrb scb lcb) (QofQL tau (LofQL a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lcb - scb = abort "*@: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		| lr - sr >= lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_zunmql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
				#! lwork = fromReal (realPart work.[0])
				#! info = my_zunmql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a
									(m * sc + sr) m tau b (mb * scb + srb) mb
									(createArray lwork (0.0 +: 0.0)) lwork
				| info == 0 = (Dense b mb nb srb lrb scb lcb)
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_zunmql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m tau b
							(mb * scb + srb) mb work -1
		| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zunmql 1 0 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m tau b
								(mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
								lwork
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@) (Dense b mb nb srb lrb scb lcb) (QofLQ tau (LofLQ a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lcb - scb = abort "*@: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_zunmlq 1 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_zunmlq 1 0 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
							lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@) a (Permutation ns) = permuteRc a (invPermute ns)
	(*@) a (Product ts) = apply a ts
	where
		apply a [!!]     = a
		apply a [!t:ts!] = apply (a *@ t) ts


class (~@*) infix 0 e :: !(Transform e) !*(Matrix e) -> *(Matrix e)

instance ~@* Real
where
	(~@*) (QofQL tau (LofQL a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lrb - srb = abort "~@*: Mismatch"
		#! work = createArray 1 0.0
		| lr - sr + 1 >= lc - sc + 1
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_dormql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr)
								m tau b (mb * scb + srb) mb work -1
			#! lwork = fromReal work.[0]
			| info == 0 && lwork >= max 1 (lcb - scb + 1)
				#! info = my_dormql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a
									(m * sc + sr) m tau b (mb * scb + srb) mb
									(createArray lwork 0.0) lwork
				| info == 0 = (Dense b mb nb srb lrb scb lcb)
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_dormql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr)
							m tau b (mb * scb + srb) mb work -1
		| info == 0 && fromReal work.[0] >= max 1 (lcb - scb + 1)
			#! lwork = fromReal work.[0]
			#! info = my_dormql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m tau b
								(mb * scb + srb) mb
								(createArray lwork 0.0) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(~@*) (QofLQ tau (LofLQ a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lrb - srb = abort "~@*: Mismatch"
		#! work = createArray 1 0.0
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_dormlq 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal work.[0]
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_dormlq 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(~@*) (Permutation ms) a = permuteLr (invPermute ms) a
	(~@*) (Product ts) a = apply ts a
	where
		apply [!!] a = a
		apply [!t:ts!] a = apply ts (t ~@* a)

instance ~@* Complex
where
	(~@*) (QofQL tau (LofQL a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lrb - srb = abort "~@*: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		| lr - sr >= lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_zunmql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
				#! lwork = fromReal (realPart work.[0])
				#! info = my_zunmql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
								lwork
				| info == 0 = (Dense b mb nb srb lrb scb lcb)
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_zunmql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_zunmql 0 1 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m
								tau b (mb * scb + srb) mb
								(createArray lwork (0.0 +: 0.0)) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(~@*) (QofLQ tau (LofLQ a m n sr lr sc lc)) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lrb - srb = abort "~@*: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_zunmlq 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lcb - scb + 1)
			#! info = my_zunmlq 0 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
							lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(~@*) (Permutation ms) a = permuteLc (invPermute ms) a
	(~@*) (Product ts) a = apply ts a
	where
		apply [!!] a = a
		apply [!t:ts!] a = apply ts (t ~@* a)


class lq e :: !*(Matrix e) -> (Factor e, Transform e)

instance lq Real
where
	lq (Dense a m n sr lr sc lc)
		| lr < sr || lc < sc = (LofLQ a m n sr lr sc lc, Product [!!])
		#! work = createArray 1 0.0
		#! tau = createArray (min (lr - sr + 1) (lc - sc + 1)) 0.0
		#! info = my_dgelqf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work -1
		#! lwork = fromReal work.[0]
		| info == 0 && lwork >= max 1 (lr - sr + 1)
			#! work = createArray lwork 0.0
			#! info = my_dgelqf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work lwork
			| info == 0 =
				let aa = LofLQ a m n sr lr sc lc in (aa, QofLQ tau aa)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)

instance lq Complex
where
	lq (Dense a m n sr lr sc lc)
		| lr < sr || lc < sc = (LofLQ a m n sr lr sc lc, Product [!!])
		#! work = createArray 1 (0.0 +: 0.0)
		#! tau = createArray (min (lr - sr + 1) (lc - sc + 1)) (0.0 +: 0.0)
		#! info = my_zgelqf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lr - sr + 1)
			#! work = createArray lwork (0.0 +: 0.0)
			#! info = my_zgelqf (lr - sr + 1) (lc - sc + 1) a (m * sc + sr) m
							tau work lwork
			| info == 0 =
				let aa = LofLQ a m n sr lr sc lc in (aa, QofLQ tau aa)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)


class (*@~) infix 0 e :: !*(Matrix e) !(Transform e) -> *(Matrix e)

instance *@~ Real
where
	(*@~) (Dense b mb nb srb lrb scb lcb) (QofQL tau (LofQL a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lcb - scb = abort "*@~: Mismatch"
		#! work = createArray 1 0.0
		| lr - sr >= lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_dormql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal work.[0] >= max 1 (lcb - scb + 1)
				#! lwork = fromReal work.[0]
				#! info = my_dormql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a
									(m * sc + sr) m tau b (mb * scb + srb) mb
									(createArray lwork 0.0) lwork
				| info == 0 = Dense b mb nb srb lrb scb lcb
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_dormql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m
							tau b (mb * scb + srb) mb work -1
		| info == 0 && fromReal work.[0] >= max 1 (lrb - srb + 1)
			#! lwork = fromReal work.[0]
			#! info = my_dormql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m tau b
								(mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@~) (Dense b mb nb srb lrb scb lcb) (QofLQ tau (LofLQ a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lcb - scb = abort "*@~: Mismatch"
		#! work = createArray 1 0.0
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_dormlq 1 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal work.[0]
		| info == 0 && lwork >= max 1 (lrb - srb + 1)
			#! info = my_dormlq 1 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork 0.0) lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@~) a (Permutation ns) = permuteRr a ns
	(*@~) a (Product ts) = apply a ts
	where
		apply a [!!]     = a
		apply a [!t:ts!] = (apply a ts) *@~ t

instance *@~ Complex
where
	(*@~) (Dense b mb nb srb lrb scb lcb) (QofQL tau (LofQL a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lr - sr <> lcb - scb = abort "*@~: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		| lr - sr >= lc - sc
			#! k = min (lr - sr + 1) (lc - sc + 1)
			#! info = my_zunmql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
								tau b (mb * scb + srb) mb work -1
			| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
				#! lwork = fromReal (realPart work.[0])
				#! info = my_zunmql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a
									(m * sc + sr) m tau b (mb * scb + srb) mb
									(createArray lwork (0.0 +: 0.0)) lwork
				| info == 0 = Dense b mb nb srb lrb scb lcb
				| otherwise = abort ("info = " +++ toString info)
			| otherwise = abort ("info = " +++ toString info)
		#! k = lr - sr + 1
		#! info = my_zunmql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a
							(m * (sc + lc - sc - lr + sr) + sr) m
							tau b (mb * scb + srb) mb work -1
		| info == 0 && fromReal (realPart work.[0]) >= max 1 (lcb - scb + 1)
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zunmql 1 1 (lrb - srb + 1) (lcb - scb + 1) k a
								(m * (sc + lc - sc - lr + sr) + sr) m tau b
								(mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
								lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@~) (Dense b mb nb srb lrb scb lcb) (QofLQ tau (LofLQ a m n sr lr sc lc))
		| lrb < srb = Dense b mb nb srb lrb scb lcb
		| lc - sc <> lcb - scb = abort "*@~: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		#! k = min (lr - sr + 1) (lc - sc + 1)
		#! info = my_zunmlq 1 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb work -1
		#! lwork = fromReal (realPart work.[0])
		| info == 0 && lwork >= max 1 (lrb - srb + 1)
			#! info = my_zunmlq 1 1 (lrb - srb + 1) (lcb - scb + 1) k a (m * sc + sr) m
							tau b (mb * scb + srb) mb (createArray lwork (0.0 +: 0.0))
							lwork
			| info == 0 = (Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	(*@~) a (Permutation ns) = permuteRc a ns
	(*@~) a (Product ts) = apply a ts
	where
		apply a [!!]     = a
		apply a [!t:ts!] = (apply a ts) *@~ t


my_dgeqlf :: !Int !Int !{#Real} !Int !Int !{#Real} !{#Real} !Int -> Int
my_dgeqlf m n a dispa lda tau work lwork =
	code {
		ccall my_dgeqlf "IIAIIAAI:I"
	}

my_dormql :: !Int !Int !Int !Int !Int !{#Real} !Int !Int !{#Real} !{#Real} !Int !Int
				!{#Real} !Int -> Int
my_dormql side trans m n k a dispa lda tau c dispc ldc work lwork =
	code {
		ccall my_dormql "IIIIIAIIAAIIAI:I"
	}

my_zgeqlf :: !Int !Int !{#Complex} !Int !Int !{#Complex} !{#Complex} !Int -> Int
my_zgeqlf m n a dispa lda tau work lwork =
	code {
		ccall my_zgeqlf "IIAIIAAI:I"
	}

my_zgelqf :: !Int !Int !{#Complex} !Int !Int !{#Complex} !{#Complex} !Int -> Int
my_zgelqf m n a dispa lda tau work lwork =
	code {
		ccall my_zgelqf "IIAIIAAI:I"
	}

my_dgelqf :: !Int !Int !{#Real} !Int !Int !{#Real} !{#Real} !Int -> Int
my_dgelqf m n a dispa lda tau work lwork =
	code {
		ccall my_dgelqf "IIAIIAAI:I"
	}

my_dormlq :: !Int !Int !Int !Int !Int !{#Real} !Int !Int !{#Real} !{#Real} !Int !Int
				!{#Real} !Int -> Int
my_dormlq side trans m n k a dispa lda tau c dispc ldc work lwork =
	code {
		ccall my_dormlq "IIIIIAIIAAIIAI:I"
	}

my_zunmql :: !Int !Int !Int !Int !Int !{#Complex} !Int !Int !{#Complex} !{#Complex}
				!Int !Int !{#Complex} !Int -> Int
my_zunmql side trans m n k a dispa lda tau c dispc ldc work lwork =
	code {
		ccall my_zunmql "IIIIIAIIAAIIAI:I"
	}

my_zunmlq :: !Int !Int !Int !Int !Int !{#Complex} !Int !Int !{#Complex} !{#Complex}
				!Int !Int !{#Complex} !Int -> Int
my_zunmlq side trans m n k a dispa lda tau c dispc ldc work lwork =
	code {
		ccall my_zunmlq "IIIIIAIIAAIIAI:I"
	}


class copy e :: !u:(Matrix e) -> *(Matrix e)

instance copy Real
where
	copy (Dense a ma na sra lra sca lca)
		#! b = createArray (m * n) 0.0
		#! s = my_dlacpy 0 m n a (ma * sca + sra) ma b 0 m
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1

instance copy Complex
where
	copy (Dense a ma na sra lra sca lca)
		#! b = createArray (m * n) (0.0 +: 0.0)
		#! s = my_zlacpy 0 m n a (ma * sca + sra) ma b 0 m
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1


class iD e :: !u:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance iD Real
where
	iD (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_dlacpy 0 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1

instance iD Complex
where
	iD (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_zlacpy 0 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1


class copyL e :: !u:(Factor e) -> *(Matrix e)

instance copyL Real
where
	copyL (LofQL a ma na sra lra sca lca)
		| m == n
			#! b = createArray (m * n) 0.0
			#! s = my_dlacpy 1 m n a (ma * sca + sra) ma b 0 m
			| s == s = Dense b m n 0 (m-1) 0 (n-1)
			| otherwise = undef
		| m > n
			#! b = createArray (m * n) 0.0
			#! s = my_dlacpy 1 n n a (ma * sca + sra + m - n) ma b (m-n) m
			| s == s = Dense b m n 0 (m-1) 0 (n-1)
			| otherwise = undef
		| m < n
			#! b = createArray (m * n) 0.0
			#! s = my_dlacpy 1 m m a (ma * (sca + n - m) + sra) ma b (m * (n - m)) m
			| s == s
				#! s = my_dlacpy 0 m (n - m) a (ma * sca + sra) ma b 0 m
				| s == s = Dense b m n 0 (m-1) 0 (n-1)
				| otherwise = undef
			| otherwise = undef
	where
		m = lra - sra + 1
		n = lca - sca + 1

	copyL (LofLQ a ma na sra lra sca lca)
		#! b = createArray (m * n) 0.0
		#! s = my_dlacpy 1 m n a (ma * sca + sra) ma b 0 m
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1

	copyL (LofPLU a ma na sra lra sca lca)
		#! b = createArray (m * n) 0.0
		#! s = my_dlacpy 1 m n a (ma * sca + sra) ma b 0 m
		#! b = loop b 0
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		loop b i
			| i >= min m n = b
			= loop {b & [m * i + i] = 1.0} (i+1)

instance copyL Complex
where
	copyL (LofQL a ma na sra lra sca lca)
		| m == n
			#! b = createArray (m * n) (0.0 +: 0.0)
			#! s = my_zlacpy 1 m n a (ma * sca + sra) ma b 0 m
			| s == s = Dense b m n 0 (m-1) 0 (n-1)
			| otherwise = undef
		| m > n
			#! b = createArray (m * n) (0.0 +: 0.0)
			#! s = my_zlacpy 1 n n a (ma * sca + sra + m - n) ma b (m-n) m
			| s == s = Dense b m n 0 (m-1) 0 (n-1)
			| otherwise = undef
		| m < n
			#! b = createArray (m * n) (0.0 +: 0.0)
			#! s = my_zlacpy 1 m m a (ma * (sca + n - m) + sra) ma b (m * (n - m)) m
			| s == s
				#! s = my_zlacpy 0 m (n - m) a (ma * sca + sra) ma b 0 m
				| s == s = Dense b m n 0 (m-1) 0 (n-1)
				| otherwise = undef
			| otherwise = undef
	where
		m = lra - sra + 1
		n = lca - sca + 1

	copyL (LofLQ a ma na sra lra sca lca)
		#! b = createArray (m * n) (0.0 +: 0.0)
		#! s = my_zlacpy 1 m n a (ma * sca + sra) ma b 0 m
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1

	copyL (LofPLU a ma na sra lra sca lca)
		#! b = createArray (m * n) (0.0+:0.0)
		#! s = my_zlacpy 1 m n a (ma * sca + sra) ma b 0 m
		#! b = loop b 0
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		loop b i
			| i >= min m n = b
			= loop {b & [m * i + i] = 1.0+:0.0} (i+1)


class copyU e :: !u:(Factor e) -> *(Matrix e)

instance copyU Real
where
	copyU (UofPLU a ma na sra lra sca lca)
		#! b = createArray (k * n) 0.0
		#! s = my_dlacpy 2 k n a (ma * sca + sra) ma b 0 k
		| s == s = Dense b k n 0 (k - 1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n

instance copyU Complex
where
	copyU (UofPLU a ma na sra lra sca lca)
		#! b = createArray (k * n) (0.0+:0.0)
		#! s = my_zlacpy 2 k n a (ma * sca + sra) ma b 0 k
		| s == s = Dense b k n 0 (k - 1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n


class idL e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)

instance idL Real
where
	idL (LofQL a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| m == n
			#! s = my_dlacpy 1 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
			| s == s = Dense b mb nb srb lrb scb lcb
			| otherwise = undef
		| m > n
			#! s = my_dlacpy 1 n n a (ma * sca + sra + m - n) ma b
							(mb * scb + srb + m - n) mb
			| s == s = Dense b mb nb srb lrb scb lcb
			| otherwise = undef
		| m < n
			#! s = my_dlacpy 1 m m a (ma * (sca + n - m) + sra) ma b
							(mb * (scb + n - m) + srb) mb
			| s == s
				#! s = my_dlacpy 0 m (n - m) a (ma * sca + sra) ma b (mb * scb + srb)
								mb
				| s == s = Dense b mb nb srb lrb scb lcb
				| otherwise = undef
			| otherwise = undef
	where
		m = lra - sra + 1
		n = lca - sca + 1

	idL (LofLQ a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_dlacpy 1 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1

	idL (LofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_dlacpy 1 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
		#! b = loop b 0
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1
		loop b i
			| i >= min m n = b
			= loop {b & [mb * (scb + i) + srb + i] = 1.0} (i+1)

instance idL Complex
where
	idL (LofQL a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| m == n
			#! s = my_zlacpy 1 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
			| s == s = Dense b mb nb srb lrb scb lcb
			| otherwise = undef
		| m > n
			#! s = my_zlacpy 1 n n a (ma * sca + sra + m - n) ma b
							(mb * scb + srb + m - n) mb
			| s == s = Dense b mb nb srb lrb scb lcb
			| otherwise = undef
		| m < n
			#! s = my_zlacpy 1 m m a (ma * (sca + n - m) + sra) ma b
							(mb * (scb + n - m) + srb) mb
			| s == s
				#! s = my_zlacpy 0 m (n - m) a (ma * sca + sra) ma b (mb * scb + lrb)
								mb
				| s == s = Dense b mb nb srb lrb scb lcb
				| otherwise = undef
			| otherwise = undef
	where
		m = lra - sra + 1
		n = lca - sca + 1

	idL (LofLQ a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_zlacpy 1 m n a (ma * sca + sra) ma b (mb * scb + lrb) mb
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1

	idL (LofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_zlacpy 1 m n a (ma * sca + sra) ma b (mb * scb + srb) mb
		#! b = loop b 0
		| s == s = (Dense b mb nb srb lrb scb lcb)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		loop b i
			| i >= min m n = b
			= loop {b & [mb * (scb + i) + srb + i] = 1.0+:0.0} (i+1)


class idU e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)

instance idU Real
where
	idU (UofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_dlacpy 2 k n a (ma * sca + sra) ma b (mb * scb + srb) mb
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n

instance idU Complex
where
	idU (UofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		#! s = my_zlacpy 2 k n a (ma * sca + sra) ma b (mb * scb + srb) mb
		| s == s = Dense b mb nb srb lrb scb lcb
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n

/*
class copyU e :: (Factor e) -> *(Matrix e)

instance copyU Real
where
	copyU (_ a ma na sra lra sca lca)
		#! b = createArray (m * n) 0.0
		#! s = my_dlacpy 2 m n a (ma * sca + sra) ma b 0 m
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1

instance copyU Complex
where
	copyU (_ a ma na sra lra sca lca)
		#! b = createArray (m * n) (0.0 +: 0.0)
		#! s = my_zlacpy 2 m n a (ma * sca + sra) ma b 0 m
		| s == s = Dense b m n 0 (m-1) 0 (n-1)
	where
		m = lra - sra + 1
		n = lca - sca + 1
*/

my_dlacpy :: !Int !Int !Int !{#Real} !Int !Int !{#Real} !Int !Int -> Int
my_dlacpy side m n a dispa lda b dispb ldb =
	code {
		ccall my_dlacpy "IIIAIIAII:I"
	}
	
my_zlacpy :: !Int !Int !Int !{#Complex} !Int !Int !{#Complex} !Int !Int -> Int
my_zlacpy side m n a dispa lda b dispb ldb =
	code {
		ccall my_zlacpy "IIIAIIAII:I"
	}


class transp` e :: !u:(Matrix e) !*(Matrix e) -> *(Matrix e)

instance transp` Real
where
	transp` (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| lra - sra <> lcb - scb || lca - sca <> lrb - srb = abort "transp`: Mismatch"
		= Dense {b & [mb * (scb + j) + srb + i] = a.[ma * (sca + i) + sra + j] \\
					j <- [0 .. n-1], i <- [0 .. m-1]}
				mb nb srb lrb scb lcb
	where
		n = lcb - scb + 1
		m = lrb - srb + 1

instance transp` Complex
where
	transp` (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| lra - sra <> lcb - scb || lca - sca <> lrb - srb = abort "transp`: Mismatch"
		= Dense {b & [mb * (scb + j) + srb + i] = conjg a.[ma * (sca + i) + sra + j] \\
					j <- [0 .. n-1], i <- [0 .. m-1]}
				mb nb srb lrb scb lcb
	where
		n = lcb - scb + 1
		m = lrb - srb + 1


class transp e :: !u:(Matrix e) -> *(Matrix e)

instance transp Real
where
	transp a = transp` a (zerosM (noOfColsM a) (noOfRowsM a))

instance transp Complex
where
	transp a = transp` a (zerosM (noOfColsM a) (noOfRowsM a))


// Full rank pseudo-inverse

class (:\:) e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :\: Real
where
	(:\:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> max m n = abort ":\:: Mismatch"
		#! work = createArray 1 0.0
		#! info = my_dgels trans m n nrhs a dispa lda b dispb ldb work -1
		| info == 0 && fromReal work.[0] >= max 1 (min m n + max (min m n) nrhs)
			#! lwork = fromReal work.[0]
			#! info = my_dgels trans m n nrhs a dispa lda b dispb ldb
								(createArray lwork 0.0) lwork
			| info == 0 =
				if (m >= n)
					(Dense b mb nb srb (lrb - m + n) scb lcb)
					(Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		lda = ma
		ldb = mb
		trans = 0 // No transpose
		
instance :\: Complex
where
	(:\:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> max m n = abort ":\:: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		#! info = my_zgels trans m n nrhs a dispa lda b dispb ldb work -1
		| info == 0 && fromReal (realPart work.[0]) >=
											max 1 (min m n + max (min m n) nrhs)
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zgels trans m n nrhs a dispa lda b dispb ldb
								(createArray lwork (0.0 +: 0.0)) lwork
			| info == 0 =
				if (m >= n)
					(Dense b mb nb srb (lrb - m + n) scb lcb)
					(Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		lda = ma
		ldb = mb
		trans = 0 // No transpose

class ($\) e :: !u:(Matrix e) !v:(Matrix e) -> *(Matrix e)

instance $\ Real
where
	($\) a b = divM a b

instance $\ Complex
where
	($\) a b = divM a b

divM a b = divMM a b
where
	divMM a b
		| m >= n = copy a :\: copy b
		#! bc = matrix2x1 m  (iD b)
					   (n-m) ooo    (zeros n nrhs)
		#! x = copy a :\: bc
		= x $ (0 ... m-1, <.>)
	where
		m = noOfRows a
		n = noOfCols a
		nrhs = noOfCols b

class (:~\:) e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)

instance :~\: Real
where
	(:~\:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> max m n = abort ":~\:: Mismatch"
		#! work = createArray 1 0.0
		#! info = my_dgels trans m n nrhs a dispa lda b dispb ldb work -1
		| info == 0 && fromReal work.[0] >= max 1 (min m n + max (min m n) nrhs)
			#! lwork = fromReal work.[0]
			#! info = my_dgels trans m n nrhs a dispa lda b dispb ldb
								(createArray lwork 0.0) lwork
			| info == 0 =
				if (m <= n)
					(Dense b mb nb srb (lrb - n + m) scb lcb)
					(Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		lda = ma
		ldb = mb
		trans = 1 // Transpose
		
instance :~\: Complex
where
	(:~\:) (Dense a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| lcb < scb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> max m n = abort ":~\:: Mismatch"
		#! work = createArray 1 (0.0 +: 0.0)
		#! info = my_zgels trans m n nrhs a dispa lda b dispb ldb work -1
		| info == 0 && fromReal (realPart work.[0]) >=
											max 1 (min m n + max (min m n) nrhs)
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zgels trans m n nrhs a dispa lda b dispb ldb
								(createArray lwork (0.0 +: 0.0)) lwork
			| info == 0 =
				if (m <= n)
					(Dense b mb nb srb (lrb - n + m) scb lcb)
					(Dense b mb nb srb lrb scb lcb)
			| otherwise = abort ("info = " +++ toString info)
		| otherwise = abort ("info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		lda = ma
		ldb = mb
		trans = 1 // Conjugate ranspose

class (~\) e :: !u:(Matrix e) !v:(Matrix e) -> *(Matrix e)

instance ~\ Real
where
	(~\) a b = divM a b

instance ~\ Complex
where
	(~\) a b = divTM a b

divTM a b = divTMM a b
where
	divTMM a b
		| m <= n = copy a :~\: copy b
		#! bc = matrix2x1 n  (iD b)
					   (m-n) ooo    (zeros n nrhs)
		#! x = copy a :~\: bc
		= x $ (0 ... n-1, <.>)
	where
		m = noOfRows a
		n = noOfCols a
		nrhs = noOfCols b

my_dgels :: !Int !Int !Int !Int !{#Real} !Int !Int !{#Real} !Int !Int !{#Real} !Int
			-> Int
my_dgels trans m n nrhs a dispa lda b dispb ldb work lwork =
	code {
		ccall my_dgels "IIIIAIIAIIAI:I"
	}

my_zgels :: !Int !Int !Int !Int !{#Complex} !Int !Int !{#Complex} !Int !Int
			!{#Complex} !Int -> Int
my_zgels trans m n nrhs a dispa lda b dispb ldb work lwork =
	code {
		ccall my_zgels "IIIIAIIAIIAI:I"
	}

class (~\*) infixr 7 e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)

instance ~\* Real
where
	(~\*) (LofQL a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "triSolveT: mismatch"
		| m == n
			#! info = my_dtrtrs uplo trans diag n nrhs a dispa ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolveT: info = " +++ toString info)
		| m > n
			#! info = my_dtrtrs uplo trans diag n nrhs a (dispa + m - n) ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("~\*: info = " +++ toString info)
		| otherwise = abort ("~\*: Inverting L of fat QL")
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 1 // Transpose
		diag = 0  // Not unit diagonal
	(~\*) (LofLQ a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 ||  lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "~\*: mismatch"
		#! info = my_dtrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0 = Dense b mb nb srb lrb scb lcb
		| otherwise = abort ("~\*: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 1 // Transpose
		diag = 0  // Not unit diagonal
	(~\*) (LofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> m = abort "~\*: mismatch"
		| m > n
			#! s = my_cblas_dgemm CblasColMajor CblasTrans CblasNoTrans
				(m - n) nrhs n -1.0 a (ma * (sca + n) + sra) ma b (mb * scb + srb + n) mb
				1.0 b (mb * scb + srb) mb
			| s == s
				#! info = my_dtrtrs uplo trans diag n nrhs a dispa ma b dispb mb
				| info == 0 = Dense b mb nb srb lrb scb lcb
				= abort ("~\*:LofPLU: info = " +++ toString info)
			= undef
		#! info = my_dtrtrs uplo trans diag m nrhs a dispa ma b dispb mb
		| info == 0 = Dense b mb nb srb lrb scb lcb
		= abort ("~\*:LofPLU: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 1 // Transpose
		diag = 1  // Unit diagonal
	(~\*) (UofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> n = abort "~\*:UofPLU: mismatch"
		#! info = my_dtrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0
			| m >= n = Dense b mb nb srb lrb scb lcb
			= abort "~\*:UofPLU: fat U"
		= abort ("~\*: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 1  // Upper
		trans = 1 // Transpose
		diag = 0  // Unit diagonal

instance ~\* Complex
where
	(~\*) (LofQL a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "triSolveT: mismatch"
		| m == n
			#! info = my_ztrtrs uplo trans diag n nrhs a dispa ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolveT: info = " +++ toString info)
		| m > n
			#! info = my_ztrtrs uplo trans diag n nrhs a (dispa + m - n) ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolveT: info = " +++ toString info)
		| otherwise = abort ("triSolveT: Inverting L of fat QL")
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 1 // Transpose
		diag = 0  // Not unit diagonal
	(~\*) (LofLQ a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "~\*: mismatch"
		#! info = my_ztrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0 = Dense b mb nb srb lrb scb lcb
		| otherwise = abort ("~\*: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 1 // Transpose
		diag = 0  // Not unit diagonal
	(~\*) (LofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> m = abort "~\*: mismatch"
		| m > n
			#! s = my_cblas_zgemm CblasColMajor CblasConjTrans CblasNoTrans
				(m - n) nrhs n {-1.0+:0.0} a (ma * (sca + n) + sra) ma
				b (mb * scb + srb + n) mb {1.0+:0.0} b (mb * scb + srb) mb
			| s == s
				#! info = my_ztrtrs uplo trans diag n nrhs a dispa ma b dispb mb
				| info == 0 = Dense b mb nb srb lrb scb lcb
				= abort ("~\*:LofPLU: info = " +++ toString info)
			= undef
		#! info = my_ztrtrs uplo trans diag m nrhs a dispa ma b dispb mb
		| info == 0 = Dense b mb nb srb lrb scb lcb
		= abort ("~\*:LofPLU: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 1 // Transpose
		diag = 1  // Unit diagonal
	(~\*) (UofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> n = abort "~\*:UofPLU: mismatch"
		#! info = my_ztrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0
			| m >= n = Dense b mb nb srb lrb scb lcb
			= abort "~\*:UofPLU: fat U"
		= abort ("~\*: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 1  // Upper
		trans = 1 // Transpose
		diag = 0  // Unit diagonal

class (\*) infixr 7 e :: !u:(Factor e) !*(Matrix e) -> *(Matrix e)

instance \* Real
where
	(\*) (LofQL a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "triSolve: mismatch"
		| m == n
			#! info = my_dtrtrs uplo trans diag n nrhs a dispa ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolve: info = " +++ toString info)
		| m > n
			#! info = my_dtrtrs uplo trans diag n nrhs a (dispa + m - n) ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolve: info = " +++ toString info)
		| otherwise = abort "triSolve: inverting L of fat QL"
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 0 // No transpose
		diag = 0  // Not unit diagonal
	(\*) (LofLQ a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "triSolve: mismatch"
		#! info = my_dtrtrs uplo trans diag m nrhs a dispa ma b dispb mb
		| info == 0 = Dense b mb nb srb lrb scb lcb
		| otherwise = abort ("triSolve: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 0 // No transpose
		diag = 0  // Not unit diagonal
	(\*) (LofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> m = abort "triSolve: mismatch"
		#! info = my_dtrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0
			| m <= n = Dense b mb nb srb lrb scb lcb
			#! s = my_cblas_dgemm CblasColMajor CblasNoTrans CblasNoTrans
				(m - n) nrhs n -1.0 a (ma * sca + sra + n) ma b (mb * scb + srb) mb
				1.0 b (mb * scb + srb + n) mb
			| s == s = Dense b mb nb srb lrb scb lcb
		| otherwise = abort ("triSolve: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 0 // No transpose
		diag = 1  // Unit diagonal
	(\*) (UofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> n = abort "\*:UofPLU: mismatch"
		#! info = my_dtrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0
			| m >= n = Dense b mb nb srb lrb scb lcb
			= matrix2x1 (min m n) ooo
					(n - min m n) (scale 0.0) (Dense b mb nb srb lrb scb lcb)
		| otherwise = abort ("\*:UofPLU: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 1  // Upper
		trans = 0 // No transpose
		diag = 0  // Unit diagonal

instance \* Complex
where
	(\*) (LofQL a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "triSolve: mismatch"
		| m == n
			#! info = my_ztrtrs uplo trans diag n nrhs a dispa ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolve: info = " +++ toString info)
		| m > n
			#! info = my_ztrtrs uplo trans diag n nrhs a (dispa + m - n) ma b dispb mb
			| info == 0 = Dense b mb nb srb lrb scb lcb
			| otherwise = abort ("triSolve: info = " +++ toString info)
		| otherwise = abort "triSolve: Inverting L of fat QL"
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 0 // No transpose
		diag = 0  // Not unit diagonal
	(\*) (LofLQ a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> min n m = abort "triSolve: mismatch"
		#! info = my_ztrtrs uplo trans diag m nrhs a dispa ma b dispb mb
		| info == 0 = Dense b mb nb srb lrb scb lcb
		| otherwise = abort ("triSolve: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 0 // No transpose
		diag = 0  // Not unit diagonal
	(\*) (LofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> m = abort "triSolve: mismatch"
		#! info = my_ztrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0
			| m <= n = Dense b mb nb srb lrb scb lcb
			#! s = my_cblas_zgemm CblasColMajor CblasNoTrans CblasNoTrans
				(m - n) nrhs n {-1.0+:0.0} a (ma * sca + sra + n) ma b (mb * scb + srb) mb
				{1.0+:0.0} b (mb * scb + srb + n) mb
			| s == s = Dense b mb nb srb lrb scb lcb
		| otherwise = abort ("triSolve: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 0  // Lower
		trans = 0 // No transpose
		diag = 1  // Unit diagonal
	(\*) (UofPLU a ma na sra lra sca lca) (Dense b mb nb srb lrb scb lcb)
		| nrhs < 1 || lrb < srb = Dense b mb nb srb lrb scb lcb
		| lrb - srb + 1 <> n = abort "triSolve: mismatch"
		#! info = my_ztrtrs uplo trans diag (min m n) nrhs a dispa ma b dispb mb
		| info == 0
			| m >= n = Dense b mb nb srb lrb scb lcb
			= matrix2x1 (min m n) ooo
					(n - min m n) (scale (0.0+:0.0)) (Dense b mb nb srb lrb scb lcb)
		| otherwise = abort ("triSolve: info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		nrhs = lcb - scb + 1
		dispa = ma * sca + sra
		dispb = mb * scb + srb
		uplo = 1  // Upper
		trans = 0 // No transpose
		diag = 0  // Unit diagonal


my_dtrtrs :: !Int !Int !Int !Int !Int !{#Real} !Int !Int !{#Real} !Int !Int -> Int
my_dtrtrs uplo trans diag n nrhs a dispa lda b dispb ldb =
	code {
		ccall my_dtrtrs "IIIIIAIIAII:I"
	}

my_ztrtrs :: !Int !Int !Int !Int !Int !{#Complex} !Int !Int !{#Complex} !Int
				!Int -> Int
my_ztrtrs uplo trans diag n nrhs a dispa lda b dispb ldb =
	code {
		ccall my_ztrtrs "IIIIIAIIAII:I"
	}


genRandComplex :: !Int !Int -> [Complex]
genRandComplex s1 s2 = map (\(x,y) -> {re=x, im=y})
						   (zip2 (genRandReal s1) (genRandReal s2))


// SVD routines

class svd e :: !*(Matrix e) -> (*(Matrix e), *{#Real}, *(Matrix e))

instance svd Real
where
	svd (Dense a ma na sra lra sca lca)
		#! u = createArray (m * k) 0.0
		#! vt = createArray (k * n) 0.0
		#! s = createArray k 0.0
		#! work = createArray 1 0.0
		#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1
		| info == 0
			#! lwork = fromReal work.[0]
			#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork 0.0) lwork
			| info == 0 = (	Dense u m k 0 (m-1) 0 (k-1)
						  , s
						  , Dense vt k n 0 (k-1) 0 (n-1)
						  )
			| otherwise = abort ("svd : info = " +++ toString info)
		| otherwise = abort ("svd : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 1 // 'S'
		jobvt = 1 // 'S'

instance svd Complex
where
	svd (Dense a ma na sra lra sca lca)
		#! u = createArray (m * k) (0.0+:0.0)
		#! vt = createArray (k * n) (0.0+:0.0)
		#! s = createArray k 0.0
		#! work = createArray 1 (0.0+:0.0)
		#! rwork = createArray (5*k*(k+1)) 0.0
		#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1 rwork
		| info == 0 && fromReal (realPart work.[0]) >= 2 * k + max m n
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork (0.0+:0.0)) lwork rwork
			| info == 0 = (	Dense u m k 0 (m-1) 0 (k-1)
						  , s
						  , Dense vt k n 0 (k-1) 0 (n-1)
						  )
			| otherwise = abort ("svd : info = " +++ toString info)
		| otherwise = abort ("svd : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 1 // 'S'
		jobvt = 1 // 'S'

class svdL e :: !*(Matrix e) -> (*(Matrix e), *{#Real})

instance svdL Real
where
	svdL (Dense a ma na sra lra sca lca)
		#! u = createArray (m * k) 0.0
		#! vt = createArray 1 0.0
		#! s = createArray k 0.0
		#! work = createArray 1 0.0
		#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1
		| info == 0
			#! lwork = fromReal work.[0]
			#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork 0.0) lwork
			| info == 0 = (	Dense u m k 0 (m-1) 0 (k-1)
						  , s
						  )
			| otherwise = abort ("svdL : info = " +++ toString info)
		| otherwise = abort ("svdL : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 1 // 'S'
		jobvt = 3 // 'N'

instance svdL Complex
where
	svdL (Dense a ma na sra lra sca lca)
		#! u = createArray (m * k) (0.0+:0.0)
		#! vt = createArray 1 (0.0+:0.0)
		#! s = createArray k 0.0
		#! work = createArray 1 (0.0+:0.0)
		#! rwork = createArray (5*k*(k+1)) 0.0
		#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1 rwork
		| info == 0 && fromReal (realPart work.[0]) >= 2 * k + max m n
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork (0.0+:0.0)) lwork rwork
			| info == 0 = (	Dense u m k 0 (m-1) 0 (k-1)
						  , s
						  )
			| otherwise = abort ("svdL : info = " +++ toString info)
		| otherwise = abort ("svdL : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 1 // 'S'
		jobvt = 3 // 'N'

class svdR e :: !*(Matrix e) -> (*{#Real}, *(Matrix e))

instance svdR Real
where
	svdR (Dense a ma na sra lra sca lca)
		#! u = createArray 1 0.0
		#! vt = createArray (k * n) 0.0
		#! s = createArray k 0.0
		#! work = createArray 1 0.0
		#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1
		| info == 0
			#! lwork = fromReal work.[0]
			#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork 0.0) lwork
			| info == 0 = (	s
						  , Dense vt k n 0 (k-1) 0 (n-1)
						  )
			| otherwise = abort ("svdR : info = " +++ toString info)
		| otherwise = abort ("svdR : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 3 // 'N'
		jobvt = 1 // 'S'

instance svdR Complex
where
	svdR (Dense a ma na sra lra sca lca)
		#! u = createArray 1 (0.0+:0.0)
		#! vt = createArray (k * n) (0.0+:0.0)
		#! s = createArray k 0.0
		#! work = createArray 1 (0.0+:0.0)
		#! rwork = createArray (5*k*(k+1)) 0.0
		#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1 rwork
		| info == 0 && fromReal (realPart work.[0]) >= 2 * k + max m n
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork (0.0+:0.0)) lwork rwork
			| info == 0 = (	s
						  , Dense vt k n 0 (k-1) 0 (n-1)
						  )
			| otherwise = abort ("svdR : info = " +++ toString info)
		| otherwise = abort ("svdR : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 3 // 'N'
		jobvt = 1 // 'S'

class singValues e :: !*(Matrix e) -> *{#Real}

instance singValues Real
where
	singValues (Dense a ma na sra lra sca lca)
		#! u = createArray 1 0.0
		#! vt = createArray 1 0.0
		#! s = createArray k 0.0
		#! work = createArray 1 0.0
		#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1
		| info == 0
			#! lwork = fromReal work.[0]
			#! info = my_dgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork 0.0) lwork
			| info == 0 = s
			| otherwise = abort ("singValues : info = " +++ toString info)
		| otherwise = abort ("singValues : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 3 // 'N'
		jobvt = 3 // 'N'

instance singValues Complex
where
	singValues (Dense a ma na sra lra sca lca)
		#! u = createArray 1 (0.0+:0.0)
		#! vt = createArray 1 (0.0+:0.0)
		#! s = createArray k 0.0
		#! work = createArray 1 (0.0+:0.0)
		#! rwork = createArray (5*k*(k+1)) 0.0
		#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k work -1 rwork
		| info == 0 && fromReal (realPart work.[0]) >= 2 * k + max m n
			#! lwork = fromReal (realPart work.[0])
			#! info = my_zgesvd jobu jobvt m n a dispa ma s u 0 m vt 0 k
								(createArray lwork (0.0+:0.0)) lwork rwork
			| info == 0 = s
			| otherwise = abort ("singValues : info = " +++ toString info)
		| otherwise = abort ("singValues : info = " +++ toString info)
	where
		m = lra - sra + 1
		n = lca - sca + 1
		k = min m n
		dispa = ma * sca + sra
		jobu = 3 // 'N'
		jobvt = 3 // 'N'

my_dgesvd :: !Int !Int !Int !Int !{#Real} !Int !Int !{#Real} !{#Real} !Int !Int
			!{#Real} !Int !Int !{#Real} !Int -> Int
my_dgesvd jobu jobvt m n a dispa lda u dispu ldu s vt dispvt ldvt work lwork =
	code {
		ccall my_dgesvd "IIIIAIIAAIIAIIAI:I"
	}

my_zgesvd :: !Int !Int !Int !Int !{#Complex} !Int !Int !{#Real} !{#Complex} !Int !Int
			!{#Complex} !Int !Int !{#Complex} !Int !{#Real} -> Int
my_zgesvd jobu jobvt m n a dispa lda u dispu ldu s vt dispvt ldvt work lwork rwork =
	code {
		ccall my_zgesvd "IIIIAIIAAIIAIIAIA:I"
	}


// Scaling routines, all in Clean

class (:.*:) e :: !u:{#Real} !*(Matrix e) -> *(Matrix e)

instance :.*: Real
where
	(:.*:) s (Dense a ma na sr lr sc lc) = Dense (loop a 0 0) ma na sr lr sc lc
	where
		m = lr - sr + 1
		n = lc - sc + 1
		loop :: !*{#Real} !Int !Int -> *{#Real}
		loop a i j
			| j >= n = a
			| i >= m = loop a 0 (j + 1)
			#! (aij, a) = a![ma * (sc + j) + sr + i]
			= loop {a & [ma * (sc + j) + sr + i] = aij * s.[i]} (i+1) j

instance :.*: Complex
where
	(:.*:) s (Dense a ma na sr lr sc lc) = Dense (loop a 0 0) ma na sr lr sc lc
	where
		m = lr - sr + 1
		n = lc - sc + 1
		scale :: !Real !Complex -> Complex
		scale s z = {re = z.re * s, im = z.im * s}
		loop :: !*{#Complex} !Int !Int -> *{#Complex}
		loop a i j
			| j >= n = a
			| i >= m = loop a 0 (j + 1)
			#! (aij, a) = a![ma * (sc + j) + sr + i]
			= loop {a & [ma * (sc + j) + sr + i] = scale s.[i] aij} (i+1) j

class (:*.:) e :: !*(Matrix e) !u:{#Real} -> *(Matrix e)

instance :*.: Real
where
	(:*.:) (Dense a ma na sr lr sc lc) s = Dense (loop a 0 0) ma na sr lr sc lc
	where
		m = lr - sr + 1
		n = lc - sc + 1
		loop :: !*{#Real} !Int !Int -> *{#Real}
		loop a i j
			| j >= n = a
			| i >= m = loop a 0 (j + 1)
			#! (aij, a) = a![ma * (sc + j) + sr + i]
			= loop {a & [ma * (sc + j) + sr + i] = aij * s.[j]} (i+1) j

instance :*.: Complex
where
	(:*.:) (Dense a ma na sr lr sc lc) s = Dense (loop a 0 0) ma na sr lr sc lc
	where
		m = lr - sr + 1
		n = lc - sc + 1
		scale :: !Real !Complex -> Complex
		scale s z = {re = z.re * s, im = z.im * s}
		loop :: !*{#Complex} !Int !Int -> *{#Complex}
		loop a i j
			| j >= n = a
			| i >= m = loop a 0 (j + 1)
			#! (aij, a) = a![ma * (sc + j) + sr + i]
			= loop {a & [ma * (sc + j) + sr + i] = scale s.[j] aij} (i+1) j

class scale e :: !e !*(Matrix e) -> *(Matrix e)
		
instance scale Real
where
	scale s (Dense a ma na sr lr sc lc) = Dense (loop a 0 0) ma na sr lr sc lc
	where
		m = lr - sr + 1
		n = lc - sc + 1
		loop :: !*{#Real} !Int !Int -> *{#Real}
		loop a i j
			| j >= n = a
			| i >= m = loop a 0 (j + 1)
			#! (aij, a) = a![ma * (sc + j) + sr + i]
			= loop {a & [ma * (sc + j) + sr + i] = aij * s} (i+1) j

instance scale Complex
where
	scale s (Dense a ma na sr lr sc lc) = Dense (loop a 0 0) ma na sr lr sc lc
	where
		m = lr - sr + 1
		n = lc - sc + 1
		loop :: !*{#Complex} !Int !Int -> *{#Complex}
		loop a i j
			| j >= n = a
			| i >= m = loop a 0 (j + 1)
			#! (aij, a) = a![ma * (sc + j) + sr + i]
			= loop {a & [ma * (sc + j) + sr + i] = s * aij} (i+1) j


// Non-dangerous version. Uses O(n) work-space and stack-space.
chol a
	#! (m, a) = noOfRows` a
	| m < 1 = a
	#! (a00, a) = a @! (0,0)
	#! g00 = sqrt a00
	#! (a0, a) = a $! (1 ... m-1, 0 ... 0)
	#! g0 = scale (1.0 / g00) a0
	= matrix2x2        1                 (m-1)
			1  ((0,0) <=. g00)             (scale 0.0)
		 (m-1) (iD g0)         (chol o (g0 :*~--: g0))   a

// Dangerous version. Uses O(1) work-space and stack-space.
cholesky a
	#! (m, a) = noOfRows` a
	#! (n, a) = noOfCols` a
	| m <> n = abort "cholesky: non-square matrix"
	= chole a 0
	  with
		chole a i
			| i >= m = a
			#! (aii, a) = a @! (i,i)
			#! gii = sqrt aii
			| i == m-1 = ((i,i) <=. gii) a
			#! a = matrix3x3  i            1                (m-i-1)
						   i  ooo         ooo                  ooo
						   1  ooo   ((0,0) <=. gii)     (scale 0.0)
					  (m-i-1) ooo (scale (1.0 / gii))          ooo   a
			#! (gi, a) = a $# (i+1 ... m-1, i ... i)
			#! a = matrix2x2 (i+1)        (m-i-1)
					  (i+1)   ooo            ooo
					(m-i-1)   ooo  (gi :*~--: gi)   a
			= chole a (i+1)


my_dswap :: !Int !{#Real} !Int !Int !{#Real} !Int !Int -> Int
my_dswap n x dispx incx y dispy incy =
	code {
		ccall my_cblas_dswap "IAIIAII:I"
	}

my_zswap :: !Int !{#Complex} !Int !Int !{#Complex} !Int !Int -> Int
my_zswap n x dispx incx y dispy incy =
	code {
		ccall my_cblas_zswap "IAIIAII:I"
	}

// Slightly different from _lapmt
permuteLr :: !{#Int} !*(Matrix Real) -> *(Matrix Real)
permuteLr ms (Dense a ma na sr lr sc lc) =
	Dense (oloop a (createArray m 'F') 0) ma na sr lr sc lc
where
	m = let m = lr - sr + 1 in if (m == size ms) m (abort "permuteLr:wrong size")
	n = lc - sc + 1
	oloop :: !*{#Real} !*{#Char} !Int -> *{#Real}
	oloop a oks i
		| i >= m = a
		#! (oki, oks) = oks![i]
		| oki == 'T' = oloop a oks (i+1)
		#! (a, oks) = iloop a oks i ms.[i]
		= oloop a oks (i+1)
	iloop :: !*{#Real} !*{#Char} !Int !Int -> (*{#Real}, *{#Char})
	iloop a oks i j
		#! (okj, oks) = oks![j]
		| okj == 'T' = (a, oks)
		#! s = my_dswap n a (ma * sc + sr + i) ma a (ma * sc + sr + j) ma
		| s == s = iloop a {oks & [i] = 'T'} j ms.[j]
		= undef

permuteLc :: !{#Int} !*(Matrix Complex) -> *(Matrix Complex)
permuteLc ms (Dense a ma na sr lr sc lc) =
	Dense (oloop a (createArray m 'F') 0) ma na sr lr sc lc
where
	m = let m = lr - sr + 1 in if (m == size ms) m (abort "permuteLc:wrong size")
	n = lc - sc + 1
	oloop :: !*{#Complex} !*{#Char} !Int -> *{#Complex}
	oloop a oks i
		| i >= m = a
		#! (oki, oks) = oks![i]
		| oki == 'T' = oloop a oks (i+1)
		#! (a, oks) = iloop a oks i ms.[i]
		= oloop a oks (i+1)
	iloop :: !*{#Complex} !*{#Char} !Int !Int -> (*{#Complex}, *{#Char})
	iloop a oks i j
		#! (okj, oks) = oks![j]
		| okj == 'T' = (a, oks)
		#! s = my_zswap n a (ma * sc + sr + i) ma a (ma * sc + sr + j) ma
		| s == s = iloop a {oks & [i] = 'T'} j ms.[j]
		= undef

invPermute :: !{#Int} -> {#Int}
invPermute ms = {createArray (size ms) 0 & [ms.[i]] = i \\ i <- [0 .. size ms - 1]}

permuteRr :: !*(Matrix Real) !{#Int} -> *(Matrix Real)
permuteRr (Dense a ma na sr lr sc lc) ns =
	Dense (oloop a (createArray n 'F') 0) ma na sr lr sc lc
where
	m = lr - sr + 1
	n = let n = lc - sc + 1 in if (n == size ns) n (abort "permuteRr:wrong size")
	oloop :: !*{#Real} !*{#Char} !Int -> *{#Real}
	oloop a oks i
		| i >= n = a
		#! (oki, oks) = oks![i]
		| oki == 'T' = oloop a oks (i+1)
		#! (a, oks) = iloop a oks i ns.[i]
		= oloop a oks (i+1)
	iloop :: !*{#Real} !*{#Char} !Int !Int -> (*{#Real}, *{#Char})
	iloop a oks i j
		#! (okj, oks) = oks![j]
		| okj == 'T' = (a, oks)
		#! s = my_dswap m a (ma * (sc + i) + sr) 1 a (ma * (sc + j) + sr) 1
		| s == s = iloop a {oks & [i] = 'T'} j ns.[j]
		= undef

permuteRc :: !*(Matrix Complex) !{#Int} -> *(Matrix Complex)
permuteRc (Dense a ma na sr lr sc lc) ns =
	Dense (oloop a (createArray n 'F') 0) ma na sr lr sc lc
where
	m = lr - sr + 1
	n = let n = lc - sc + 1 in if (n == size ns) n (abort "permuteRr:wrong size")
	oloop :: !*{#Complex} !*{#Char} !Int -> *{#Complex}
	oloop a oks i
		| i >= n = a
		#! (oki, oks) = oks![i]
		| oki == 'T' = oloop a oks (i+1)
		#! (a, oks) = iloop a oks i ns.[i]
		= oloop a oks (i+1)
	iloop :: !*{#Complex} !*{#Char} !Int !Int -> (*{#Complex}, *{#Char})
	iloop a oks i j
		#! (okj, oks) = oks![j]
		| okj == 'T' = (a, oks)
		#! s = my_zswap m a (ma * (sc + i) + sr) 1 a (ma * (sc + j) + sr) 1
		| s == s = iloop a {oks & [i] = 'T'} j ns.[j]
		= undef

my_dgetrf :: !Int !Int !{#Real} !Int !Int !{#Int} -> Int
my_dgetrf m n a dispa lda ipiv =
	code {
		ccall my_dgetrf "IIAIIA:I"
	}

my_zgetrf :: !Int !Int !{#Complex} !Int !Int !{#Int} -> Int
my_zgetrf m n a dispa lda ipiv =
	code {
		ccall my_zgetrf "IIAIIA:I"
	}

// Converts P from PLU to standard form.
toStdPermute :: !Int !{#Int} -> {#Int}
toStdPermute m ipiv = loop {createArray m 0 & [i] = i \\ i <- [0 .. m-1]} 0
where
	n = size ipiv
	loop perm i
		| i >= n = perm
		#! (pi, perm) = perm![i]
		#! (pj, perm) = perm![ipiv.[i]-1]
		= loop {perm & [i] = pj, [ipiv.[i]-1] = pi} (i+1)

class plu e :: !*(Matrix e) -> (Transform e, Factor e, Factor e)

instance plu Real
where
	plu (Dense a ma na sr lr sc lc)
		#! ipiv = createArray (min m n) -1
		#! info = my_dgetrf m n a (ma * sc + sr) ma ipiv
		| info < 0 = abort ("plu:info = " +++ toString info)
		= ( Permutation (invPermute (toStdPermute m ipiv))
		  , LofPLU a ma na sr lr sc lc
		  , UofPLU a ma na sr lr sc lc
		  )
	where
		m = lr - sr + 1
		n = lc - sc + 1

instance plu Complex
where
	plu (Dense a ma na sr lr sc lc)
		#! ipiv = createArray (min m n) -1
		#! info = my_zgetrf m n a (ma * sc + sr) ma ipiv
		| info < 0 = abort ("plu:info = " +++ toString info)
		= ( Permutation (invPermute (toStdPermute m ipiv))
		  , LofPLU a ma na sr lr sc lc
		  , UofPLU a ma na sr lr sc lc
		  )
	where
		m = lr - sr + 1
		n = lc - sc + 1

class pluSolve e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)

instance pluSolve Real
where
	pluSolve a b = pluSolveM a b

instance pluSolve Complex
where
	pluSolve a b = pluSolveM a b

pluSolveM a b :== pluSolve a b
where
	pluSolve a b
		#! (m, a) = noOfRows` a
		#! (n, a) = noOfCols` a
		#! (nrhs, b) = noOfCols` b
		#! (p, l, u) = plu a
		| m >= n
			#! b = p ~@* b
			#! b = l \* b
			#! x = matrix2x1 n  ((\*) u)
						(m - n) ooo      b
			= x $ (0 ... n-1, <.>)
		| otherwise = abort "pluSolve: fat matrix"

class pluSolveT e :: !*(Matrix e) !*(Matrix e) -> *(Matrix e)

instance pluSolveT Real
where
	pluSolveT a b = pluSolveTM a b

instance pluSolveT Complex
where
	pluSolveT a b = pluSolveTM a b

pluSolveTM a b :== pluSolveT a b
where
	pluSolveT a b
		#! (m, a) = noOfRows` a
		#! (n, a) = noOfCols` a
		#! (nrhs, b) = noOfCols` b
		#! (p, l, u) = plu a
		| m >= n
			#! x = matrix2x1 n  ((~\*) u)
						(m - n) (scale zero) b
			#! x = l ~\* x
			= p @* x
		| otherwise = abort "pluSolveT: fat matrix"

//class MatrixArithU e | :*:, :*++:, :*--: e

// Tests for module blop.

test1 = fool (zeros 2 2)
where
	fool :: !*(Matrix Real) -> Bool
	fool a
		#! a = set a 0 0 1.0
		#! a = set a 0 1 2.0
		#! a = set a 1 0 3.0
		#! a = set a 1 1 4.0
		= norm1 a == 6.0

test2 = fool (zeros 2 2)
where
	fool :: !*(Matrix Complex) -> Bool
	fool a
		#! a = set a 0 0 (1.0 +: 0.0)
		#! a = set a 0 1 (fromInt 2)
		#! a = set a 1 0 (fromInt 3)
		#! a = set a 1 1 (fromInt 4)
		= norm1 a == 6.0

test3 = fool (zeros 2 2)
where
	fool :: !*(Matrix Real) -> Bool
	fool a
		#! a = set a 0 0 1.0
		#! a = set a 0 1 2.0
		#! a = set a 1 0 3.0
		#! a = set a 1 1 4.0
		= normInf a == 7.0

test4 = fool (zeros 2 2)
where
	fool :: !*(Matrix Complex) -> Bool
	fool a
		#! a = set a 0 0 (1.0 +: 0.0)
		#! a = set a 0 1 (fromInt 2)
		#! a = set a 1 0 (fromInt 3)
		#! a = set a 1 1 (fromInt 4)
		= normInf a == 7.0

test5 =
	let
		c = (a :+: b) (zeros m n)
	in
		norm1 c == 12.0 && normInf c == 24.0
where
	m = 3
	n = 4
	a :: Matrix Real
	a = fromFun (\i j = 1.0 + fromInt i + fromInt j) m n
	b :: Matrix Real
	b = fromFun (\i j = 1.0 + fromInt i - fromInt j) m n

test6 =
	let m = 4
		n = 5
		a = fromFun (\i j = 1.0 + fromInt i + fromInt j) m n
		b = fromFun (\i j = 1.0 - fromInt i + fromInt j) m n
		c = (a :-: b) (zeros m n)
	in
		norm1 c == 12.0 && normInf c == 30.0
	
test7 =
	let m = 3
		n = 4
		a = fromFun (\i j = (1.0 +:  2.0) + fromInt i + fromInt j) m n
		b = fromFun (\i j = (1.0 +: -2.0) + fromInt i - fromInt j) m n
		c = (a :+: b) (zeros m n)
	in
		norm1 c == 12.0 && normInf c == 24.0
	
test8 =
	let m = 4
		n = 5
		a = fromFun (\i j = (1.0 + fromInt i) +: fromInt j) m n
		b = fromFun (\i j = (1.0 - fromInt i) +: fromInt j) m n
		c = (a :-: b) (zeros m n)
	in
		norm1 c == 12.0 && normInf c == 30.0

test9 =
	let m = 4
		n = 4
		k = 4
		a = fromFun (\i j = 1.0 + fromInt i + fromInt j) m k
		b = fromFun (\i j = 1.0 + fromInt i - fromInt j) k n
		c = (a :*: b) (zeros m n)
	in
		(norm1 c == 180.0 && normInf c == 120.0)

test10 =
	let m = 1
		n = 1
		k = 4
		a = fromFun (\_ _ = 1.0) m k
		b = fromFun (\_ _ = 1.0) k n
		c = (a :*: b) (zeros m n)
	in
		norm1 c == fromInt k && normInf c == fromInt k

test11 =
	let m = 4
		n = 5
		k = 4
		a = (fromFun (\i j = if (i == 1) 1.0 2.0) m k) $ (1 ... 1, <.>)
		b = (fromFun (\i j = if (j == 1) 1.0 2.0) k n) $ (<.>, 1 ... 1)
		c = (a :*: b) (zeros 1 1)
	in
		norm1 c == fromInt k && normInf c == fromInt k

test12 =
	let m = 1
		n = 1
		k = 4
		a = fromFun (\_ _ = 1.0 +: 0.0) m k
		b = fromFun (\_ _ = 1.0 +: 0.0) k n
		c = (a :*: b) (zeros m n)
	in
		norm1 c == fromInt k && normInf c == fromInt k

test13 =
	let m = 4
		n = 5
		k = 4
		a = (fromFun (\i j = if (i == 1) (1.0 +: 0.0) (1.0 +: 2.0)) m k) $
																	(1 ... 1, <.>)
		b = (fromFun (\i j = if (j == 1) (1.0 +: 0.0) (2.0 +: 2.0)) k n) $
																	(<.>, 1 ... 1)
		c = (a :*: b) (zeros 1 1)
	in
		norm1 c == fromInt k && normInf c == fromInt k

test14 =
	let m = 4
		n = 5
		k = 4
		a = (fromFun (\i j = if (j == 1) 1.0 2.0) k m) $ (<.>, 1 ... 1)
		b = (fromFun (\i j = if (j == 1) 1.0 2.0) k n) $ (<.>, 1 ... 1)
		c = (a :~*: b) (ones 1 1)
	in
		norm1 c == fromInt k && normInf c == fromInt k

test15 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +: 4.0) k m
		b = consts (3.0 +: 4.0) k n
		c = (a :~*: b) (ones m n)
	in
		norm1 c == 25.0 * fromInt (m * k) && normInf c == 25.0 * fromInt (n * k)

test16 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +: 4.0) m k
		b = consts (3.0 +: 4.0) n k
		c = (a :*~: b) (ones m n)
	in
		norm1 c == 25.0 * fromInt (m * k) && normInf c == 25.0 * fromInt (n * k)

test17 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 m k
		b = consts 3.0 n k
		c = (a :*~: b) (ones m n)
	in
		norm1 c == 9.0 * fromInt (m * k) && normInf c == 9.0 * fromInt (n * k)

test18 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +:  4.0) m k
		b = consts (3.0 +: -4.0) k n
		c = (a :*++: b) (ones m n)
	in
		norm1 c == fromInt m + 25.0 * fromInt (k * m) &&
		normInf c == fromInt n + 25.0 * fromInt (n * k)

test19 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 m k
		b = consts 3.0 k n
		c = (a :*++: b) (ones m n)
	in
		norm1 c == fromInt m + 9.0 * fromInt (m * k) &&
		normInf c == fromInt n + 9.0 * fromInt (n * k)

test20 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +:  4.0) m k
		b = consts (3.0 +: -4.0) k n
		c = (a :*--: b) (ones m n)
	in
		norm1 c == 25.0 * fromInt (m * k) - fromInt m &&
		normInf c == 25.0 * fromInt (n * k) - fromInt n

test21 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 m k
		b = consts 3.0 k n
		c = (a :*--: b) (ones m n)
	in
		norm1 c == 9.0 * fromInt (m * k) - fromInt m &&
		normInf c == 9.0 * fromInt (n * k) - fromInt n

test22 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +: 4.0) k m
		b = consts (3.0 +: 4.0) k n
		c = (a :~*++: b) (ones m n)
	in
		norm1 c == fromInt m + 25.0 * fromInt (k * m) &&
		normInf c == fromInt n + 25.0 * fromInt (n * k)

test23 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 k m
		b = consts 3.0 k n
		c = (a :~*++: b) (ones m n)
	in
		norm1 c == fromInt m + 9.0 * fromInt (m * k) &&
		normInf c == fromInt n + 9.0 * fromInt (n * k)

test24 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +: 4.0) k m
		b = consts (3.0 +: 4.0) k n
		c = (a :~*--: b) (ones m n)
	in
		norm1 c == 25.0 * fromInt (m * k) - fromInt m &&
		normInf c == 25.0 * fromInt (n * k) - fromInt n

test25 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 k m
		b = consts 3.0 k n
		c = (a :~*--: b) (ones m n)
	in
		norm1 c == 9.0 * fromInt (m * k) - fromInt m &&
		normInf c == 9.0 * fromInt (n * k) - fromInt n

test26 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +: 4.0) m k
		b = consts (3.0 +: 4.0) n k
		c = (a :*~++: b) (ones m n)
	in
		norm1 c == fromInt m + 25.0 * fromInt (k * m) &&
		normInf c == fromInt n + 25.0 * fromInt (n * k)

test27 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 m k
		b = consts 3.0 n k
		c = (a :*~++: b) (ones m n)
	in
		norm1 c == fromInt m + 9.0 * fromInt (m * k) &&
		normInf c == fromInt n + 9.0 * fromInt (n * k)

test28 =
	let m = 4
		n = 5
		k = 4
		a = consts (3.0 +: 4.0) m k
		b = consts (3.0 +: 4.0) n k
		c = (a :*~--: b) (ones m n)
	in
		norm1 c == 25.0 * fromInt (m * k) - fromInt m &&
		normInf c == 25.0 * fromInt (n * k) - fromInt n

test29 =
	let m = 4
		n = 5
		k = 4
		a = consts 3.0 m k
		b = consts 3.0 n k
		c = (a :*~--: b) (ones m n)
	in
		norm1 c == 9.0 * fromInt (m * k) - fromInt m &&
		normInf c == 9.0 * fromInt (n * k) - fromInt n

test30 =
	let m = 5
		n = 5
		a = consts 3.0 m n
		b = consts 3.0 m n
		(q, l) = ql a
		c = q ~@* b
	in
		norm1 (c $ (0 ... m-2, <.>)) <= 3.0 * machineEpsilon * fromInt m

test31 =
	let m = 5
		n = 5
		a = consts 3.0 m n
		b = consts 3.0 m n
		(l, q) = lq a
		c = b *@~ q
	in
		norm1 (c $ (<.>, 1 ... n-1)) <= 6.0 * machineEpsilon * fromInt m

test32 =
	let m = 5
		n = 5
		a = consts 3.0 m n
		b = consts 3.0 m n
		(l, q) = lq a
		c = q @* b
	in
		norm1 (c $ (1 ... m-1, <.>)) <= 6.0 * machineEpsilon * fromInt m

test33 =
	let m = 5
		n = 5
		a = consts 3.0 m n
		b = consts 3.0 m n
		(q, l) = ql a
		c = b *@ q
	in
		norm1 (c $ (<.>, 0 ... n-2)) <= 6.0 * machineEpsilon * fromInt m

test34 =
	let m = 5
		n = 5
		a = consts (3.0 +: 2.0) m n
		b = consts (3.0 +: 2.0) m n
		(q, l) = ql a
		c = q ~@* b
	in
		norm1 (c $ (0 ... m-2, <.>)) <= 6.0 * machineEpsilon * fromInt m

test35 =
	let m = 5
		n = 5
		a = consts (3.0 +: 2.0) m n
		b = consts (3.0 +: 2.0) m n
		(q, l) = ql a
		c = b *@ q
	in
		norm1 (c $ (<.>, 0 ... n-2)) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test36 =
	let m = 5
		n = 5
		a = consts (3.0 +: 2.0) m n
		b = consts (3.0 +: 2.0) m n
		(l, q) = lq a
		c = b *@~ q
	in
		norm1 (c $ (<.>, 1 ... n-1)) <= 12.0 * machineEpsilon * fromInt m * fromInt n

test37 =
	let m = 5
		n = 5
		a = consts (3.0 +: 2.0) m n
		b = consts (3.0 +: 2.0) m n
		(l, q) = lq a
		c = q @* b
	in
		norm1 (c $ (1 ... m-1, <.>)) <= 6.0 * machineEpsilon * fromInt m

test38 =
	let m = 5
		n = 5
		a = consts 4.0 m n
		b = copy a
	in
		norm1 (a - b) == 0.0

test39 =
	let m = 5
		n = 5
		a = consts (4.0 +: 2.0) m n
		b = copy a
	in
		norm1 (a - b) == 0.0

test40 =
	let m = 5
		n = 5
		a = consts (3.0 +: 2.0) m n
		b = consts (3.0 +: 2.0) m n
		(q, l) = ql a
		c = q @* copyL l
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test41 =
	let m = 5
		n = 5
		a = consts (3.0 +: 2.0) m n
		b = consts (3.0 +: 2.0) m n
		(l, q) = lq a
		c = copyL l *@ q
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test42 =
	let m = 10
		n = 5
		a = consts 3.0 m n
		b = consts 3.0 m n
		(l, q) = lq a
		c = copyL l *@ q
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test43 =
	let m = 5
		n = 10
		a = consts 3.0 m n
		b = consts 3.0 m n
		(l, q) = lq a
		c = copyL l *@ q
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test44 =
	let m = 10
		n = 5
		a = consts 3.0 m n
		b = consts 3.0 m n
		(q, l) = ql a
		c = q @* copyL l
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test45 =
	let m = 5
		n = 10
		a = consts 3.0 m n
		b = consts 3.0 m n
		(q, l) = ql a
		c = q ~@* b
	in
		norm1 (c $ (0 ... m-2, <.>)) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test46 =
	let m = 5
		n = 10
		a = consts 3.0 m n
		b = consts 3.0 m n
		(q, l) = ql a
		c = q @* copyL l
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test47 =
	let m = 5
		n = 10
		a = consts 3.0 m n
		b = consts 3.0 m m
		(q, l) = ql a
		c = (b *@ q) *@~ q
	in
		norm1 (c - consts 3.0 m m) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test48 =
	let m = 10
		n = 5
		a = consts (3.0 +: 5.0) m n
		b = consts (3.0 +: 5.0) m n
		(q, l) = ql a
		c = q @* copyL l
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test49 =
	let m = 5
		n = 10
		a = consts (3.0 +: 5.0) m n
		b = consts (3.0 +: 5.0) m n
		(q, l) = ql a
		c = q @* copyL l
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test50 =
	let m = 5
		n = 10
		a = consts (3.0 +: 5.0) m n
		b = consts (3.0 +: 5.0) m n
		(q, l) = ql a
		c = q ~@* b
	in
		norm1 (c $ (0 ... m-2, <.>)) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test51 =
	let m = 5
		n = 10
		a = consts (3.0 +: -7.0) m n
		b = consts (3.0 +: -7.0) m m
		(q, l) = ql a
		c = (b *@ q) *@~ q
	in
		norm1 (c - consts (3.0 +: -7.0) m m) <= 10.0 * machineEpsilon * fromInt m *
												fromInt n

test52 =
	let m = 5
		n = 10
		a = consts (3.0 +: 5.0) m n
		b = consts (3.0 +: 5.0) m n
		(l, q) = lq a
		c = copyL l *@ q
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test53 =
	let m = 10
		n = 5
		a = consts (3.0 +: 5.0) m n
		b = consts (3.0 +: 5.0) m n
		(l, q) = lq a
		c = copyL l *@ q
	in
		norm1 (c - b) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test54 =
	let m = 5
		n = 10
		a = consts (3.0 +: 5.0) m n
		b = consts (3.0 +: 5.0) m n
		(l, q) = lq a
		c = b *@~ q
	in
		norm1 (c - copyL l) <= 10.0 * machineEpsilon * fromInt m * fromInt n

test55 =
	let m = 5
		n = 10
		a = consts (3.0 +: -7.0) m n
		b = consts (3.0 +: -7.0) n n
		(l, q) = lq a
		c =  q ~@* (q @* b)
	in
		norm1 (c - consts (3.0 +: -7.0) n n) <= 10.0 * machineEpsilon * fromInt m *
												fromInt n

test56 =
	let m = 10
		n = 5
		a = consts (3.0 +: -7.0) m n
		b = consts (3.0 +: -7.0) n n
		(l, q) = lq a
		c =  q ~@* (q @* b)
	in
		norm1 (c - consts (3.0 +: -7.0) n n) <= 10.0 * machineEpsilon * fromInt m *
												fromInt n

test57 =
	let m = 5
		n = 5
		a = fun2mat (\i j = toReal (i - j)) (zeros m n)
		b = eye m n
		ac = fun2mat (\i j = toReal (i - j)) (zeros m n)
		c = a :\: eye m n
	in
		norm1 (ac * c - b) <= machineEpsilon * (norm1 ac * norm1 c + norm1 b)

test58
	#! ac = fun2mat (\i j = toReal (i - j)) (zeros m n)
	#! a = copy ac
	#! bc = (ac :*: eye n n) (zeros m n)
	#! b = copy bc
	#! x = a :\: b
	= norm1 (ac * x - bc) <= machineEpsilon * (norm1 ac * norm1 x + norm1 bc)
where
	m = 10
	n = 5

test59
	#! ac = fun2mat (\i j = toReal (i - j)) (zeros m n)
	#! a = copy ac
	#! bc = matrix2x1 m  (ac :*: eye n n)
				   (n-m) ooo              (zeros n n)
	#! b = copy bc
	#! x = a :\: b
	= norm1 (ac * x - (bc $ (0 ... m-1, <.>))) <= machineEpsilon *
											(norm1 ac * norm1 x + norm1 bc)
where
	m = 5
	n = 10

test60 =
	let m = 5
		n = 5
		a = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
		b = eye m n
		ac = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
		c = a :\: eye m n
	in
		norm1 (ac * c - b) <= machineEpsilon * (norm1 ac * norm1 c + norm1 b)

test61
	#! ac = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
	#! a = copy ac
	#! bc = (ac :*: eye n n) (zeros m n)
	#! b = copy bc
	#! x = a :\: b
	= norm1 (ac * x - bc) <= machineEpsilon * (norm1 ac * norm1 x + norm1 bc)
where
	m = 10
	n = 5

test62
	#! ac = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
	#! a = copy ac
	#! bc = matrix2x1 m  (ac :*: eye n n)
				   (n-m) ooo              (zeros n n)
	#! b = copy bc
	#! x = a :\: b
	= norm1 (ac * x - (bc $ (0 ... m-1, <.>))) <= machineEpsilon *
											(norm1 ac * norm1 x + norm1 bc)
where
	m = 5
	n = 10

test63 =
	let m = 5
		n = 5
		a = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
		b = eye m n
		ac = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
		c = a :~\: eye m n
	in
		norm1 (ac ~* c - b) <= machineEpsilon * (norm1 ac * norm1 c + norm1 b)

test64
	#! ac = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
	#! a = copy ac
	#! bc = (ac :~*: eye m m) (zeros n m)
	#! b = copy bc
	#! x = a :~\: b
	= norm1 (ac ~* x - bc) <= machineEpsilon * (norm1 ac * norm1 x + norm1 bc)
where
	m = 5
	n = 10

test65
	#! ac = fun2mat (\i j = toReal (i - j) +: toReal (i + j)) (zeros m n)
	#! a = copy ac
	#! bc = matrix2x1 n  (ac :~*: eye m n)
				   (m-n)  ooo              (zeros m n)
	#! b = copy bc
	#! x = a :~\: b
	= norm1 (ac ~* x - (bc $ (0 ... n-1, <.>))) <= machineEpsilon *
											(norm1 ac * norm1 x + norm1 bc)
where
	m = 10
	n = 5

test66 =
	let m = 5
		n = 5
		a = fun2mat (\i j = toReal (i - j)) (zeros m n)
		b = eye m n
		ac = fun2mat (\i j = toReal (i - j)) (zeros m n)
		c = a :~\: eye m n
	in
		norm1 (ac ~* c - b) <= machineEpsilon * (norm1 ac * norm1 c + norm1 b)

test67
	#! ac = fun2mat (\i j = toReal (i - j)) (zeros m n)
	#! a = copy ac
	#! bc = (ac :~*: eye m m) (zeros n m)
	#! b = copy bc
	#! x = a :~\: b
	= norm1 (ac ~* x - bc) <= machineEpsilon * (norm1 ac * norm1 x + norm1 bc)
where
	m = 5
	n = 10

test68
	#! ac = fun2mat (\i j = toReal (i - j)) (zeros m n)
	#! a = copy ac
	#! bc = matrix2x1 n  (ac :~*: eye m n)
				   (m-n)  ooo              (zeros m n)
	#! b = copy bc
	#! x = a :~\: b
	= norm1 (ac ~* x - (bc $ (0 ... n-1, <.>))) <= machineEpsilon *
											(norm1 ac * norm1 x + norm1 bc)
where
	m = 10
	n = 5

test69 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(q, l) = ql a
		b = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		xq = q ~@* b
		x = l \* xq
	in
		norm1 (ac * x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test70 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		x = q ~@* xl
		xl = l \* b
	in
		norm1 (ac * x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test71 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(q, l) = ql a
		b = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		xq = q ~@* b
		x = l \* xq
	in
		norm1 (ac * x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test72 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		x = q ~@* xl
		xl = l \* b
	in
		norm1 (ac * x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test73 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(q, l) = ql a
		b = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		xl = l ~\* b
		x = q @* xl
	in
		norm1 (ac ~* x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test74 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		xq = q @* b
		x = l ~\* xq
	in
		norm1 (ac ~* x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test75 =
	let m = 10
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(q, l) = ql a
		bc = ac * fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros n nrhs)
		b = copy bc
		xq = q ~@* b
		x = l \* (xq $ (m-n ... m-1, <.>))
	in
		norm1 (ac * x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test76 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		x = q ~@* xl
		xl = l \* b
	in
		norm1 (ac * x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test77 =
	let m = 10
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(q, l) = ql a
		bc = ac * fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros n nrhs)
		b = copy bc
		xq = q ~@* b
		x = l \* (xq $ (m-n ... m-1, <.>))
	in
		norm1 (ac * x - bc) <= 160.0 * machineEpsilon * toReal (m * n)

test78 =
	let m = 5
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		bc = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros (max m n) nrhs)
		x = q ~@* xl
		xl = l \* b
	in
		norm1 (ac * x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test79 =
	let m = 5
		n = 10
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(l, q) = lq a
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros m nrhs)
		b = matrix2x1 m  (fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)))
				   (n-m) ooo                                           (zeros n nrhs)
		xl = matrix2x1 m  ((\*) l)
					(n-m) ooo		b
		x = q ~@* xl
	in
		norm1 (ac * x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test80 =
	let m = 5
		n = 10
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(l, q) = lq a
		bc = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros m nrhs)
		b = matrix2x1 m  (fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)))
				   (n-m) ooo                                           (zeros n nrhs)
		xl = matrix2x1 m  ((\*) l)
					(n-m) ooo		b
		x = q ~@* xl
	in
		norm1 (ac * x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test81 =
	let m = 10
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(q, l) = ql a
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros n nrhs)
		b = matrix2x1 (m-n) ooo
						 n  (fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)))
						 (zeros m nrhs)
		xl = matrix2x1 (m-n)  ooo
						  n  ((~\*) l) b
		x = q @* xl
	in
		norm1 (ac ~* x - bc) <= 10.0 * machineEpsilon * toReal (m * n)

test82 =
	let m = 10
		n = 5
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(q, l) = ql a
		bc = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros n nrhs)
		b = matrix2x1 (m-n) ooo
						 n  (fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)))
						 (zeros m nrhs)
		xl = matrix2x1 (m-n)  ooo
						  n  ((~\*) l) b
		x = q @* xl
	in
		norm1 (ac ~* x - bc) <= 20.0 * machineEpsilon * toReal (m * n)

test83 =
	let m = 5
		n = 10
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j) +: toReal (i - j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros n nrhs)
		bc = fun2mat (\i j = toReal i +: 3.0 / toReal (j + 1)) (zeros n nrhs)
		x = l ~\* (xq $ (0 ... m-1, <.>))
		xq = q @* b
	in
		norm1 (ac ~* x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test84 =
	let m = 5
		n = 10
		nrhs = 2
		a = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		ac = fun2mat (\i j = toReal (i + 3 * j)) (zeros m n)
		(l, q) = lq a
		b = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros n nrhs)
		bc = fun2mat (\i j = toReal i - 3.0 / toReal (j + 1)) (zeros n nrhs)
		x = l ~\* (xq $ (0 ... m-1, <.>))
		xq = q @* b
	in
		norm1 (ac ~* x - bc) <= 20.0 * machineEpsilon * toReal (m * n * nrhs)

test85 =
	let m = 10
		n = 10
		nrhs = 3
		rs = genRandReal 1
		(a, rs1) = rands rs m n
		(b, _) = rands rs1 m nrhs
		x = a $\ b
	in
		norm1 (a * x - b) <= toReal (m * n * nrhs) * machineEpsilon *
								(norm1 a * norm1 x + norm1 b)

test86 =
	let m = 10
		n = 10
		nrhs = 3
		rs = genRandComplex 1 2
		(a, rs1) = rands rs m n
		(b, _) = rands rs1 m nrhs
		x = a $\ b
	in
		norm1 (a * x - b) <= toReal (m * n * nrhs) * machineEpsilon *
								(norm1 a * norm1 x + norm1 b)
								
test87 =
	let m = 10
		n = 10
		(u, s, vt) = svd (fun2mat (\_ _ = 1.0) (zeros m n))
	in
		abs s.[0] - fromInt (max m n) <= machineEpsilon

test88 =
	let m = 10
		n = 10
		(u, s, vt) = svd (fun2mat (\_ _ = 1.0 +: 0.0) (zeros m n))
	in
		abs s.[0] - fromInt (max m n) <= toReal (m * n) * machineEpsilon

test89 =
	let m = 10
		n = 10
		(a, _) = rands (genRandReal 1) m n
		ac = copy a
		(u, s, vt) = svd ac
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test90 =
	let m = 5
		n = 5
		(a, _) = rands (genRandComplex 1 2) m n
		ac = copy a
		(u, s, vt) = svd ac
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= 10.0 * toReal (m * n) * machineEpsilon * norm1 a

test91 =
	let m = 10
		n = 10
		(a, _) = rands (genRandReal 1) m n
		(u, s) = svdL (copy a)
		(_, _, vt) = svd (copy a)
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test92 =
	let m = 10
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		(u, s) = svdL (copy a)
		(_, _, vt) = svd (copy a)
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test93 =
	let m = 10
		n = 10
		(a, _) = rands (genRandReal 1) m n
		(s, vt) = svdR (copy a)
		(u, _) = svdL (copy a)
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test94 =
	let m = 10
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		(s, vt) = svdR (copy a)
		(u, _) = svdL (copy a)
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test95 =
	let m = 10
		n = 10
		(a, _) = rands (genRandReal 1) m n
		s = singValues (copy a)
		(u, _, vt) = svd (copy a)
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test96 =
	let m = 10
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		s = singValues (copy a)
		(u, _, vt) = svd (copy a)
		svt = s :.*: vt
	in
		norm1 (a - u * svt) <= toReal (m * n) * machineEpsilon * norm1 a

test97 =
	let m = 10
		n = 5
		(a, _) = rands (genRandComplex 1 2) m n
		(u, s, vt) = svd (copy a)
		us = u :*.: s
	in
		norm1 (a - us * vt) <= toReal (m * n) * machineEpsilon * norm1 a

test98 =
	let m = 10
		n = 5
		(a, _) = rands (genRandReal 1) m n
		(u, s, vt) = svd (copy a)
		us = u :*.: s
	in
		norm1 (a - us * vt) <= toReal (m * n) * machineEpsilon * norm1 a

test99
	#! g = chol (copy a)
	= norm1 (a - g *~ g) <= machineEpsilon * norm1 a * toReal m
where
	m = 10
	b = foo m
	a = b ~* b
	foo :: Int -> *(Matrix Real)
	foo m = let (a, _) = rands (genRandReal 1) m m in a

test100 =
	let a = permuteLr {2,1,0,4,3} (toMatrix (to2D {{1.0},{2.0},{3.0},{4.0},{5.0}}) 5 1)
		b = toMatrix (to2D {{3.0},{2.0},{1.0},{5.0},{4.0}}) 5 1
	in
		norm1 (a - b) <= machineEpsilon

test101 =
	let a = permuteLc {2,1,0,4,3}
				(toMatrix (to2D {{1.0+:0.0},{2.0+:0.1},{3.0+:0.2},{4.0+:0.3},{5.0+:0.4}}) 5 1)
		b = toMatrix (to2D {{3.0+:0.2},{2.0+:0.1},{1.0+:0.0},{5.0+:0.4},{4.0+:0.3}}) 5 1
	in
		norm1 (a - b) <= machineEpsilon

test102 =
	let aa = {{1.0},{2.0},{3.0},{4.0},{5.0}}
		a = toMatrix (to2D aa) 5 1
		p = {2,1,0,4,3}
	in
	norm1 (a - permuteLr (invPermute p) (permuteLr p (toMatrix (to2D aa) 5 1))) <=
		machineEpsilon
	
test103 =
	let aa = {{1.0+:0.0},{2.0+:0.1},{3.0+:0.2},{4.0+:0.3},{5.0+:0.4}}
		a = toMatrix (to2D aa) 5 1
		p = {2,1,0,4,3}
	in
	norm1 (a - permuteLc (invPermute p) (permuteLc p (toMatrix (to2D aa) 5 1))) <=
		machineEpsilon
	
test104 =
	let aa = {{1.0,2.0,3.0,4.0,5.0}}
		a = toMatrix (to2D aa) 1 5
		p = {2,1,0,4,3}
	in
	norm1 (a - permuteRr (permuteRr (toMatrix (to2D aa) 1 5) p) (invPermute p)) <=
		machineEpsilon
	
test105 =
	let aa = {{1.0+:0.0,2.0+:0.1,3.0+:0.2,4.0+:0.3,5.0+:0.4}}
		a = toMatrix (to2D aa) 1 5
		p = {2,1,0,4,3}
	in
	norm1 (a - permuteRc (permuteRc (toMatrix (to2D aa) 1 5) p) (invPermute p)) <=
		machineEpsilon
	
test106 =
	let aa = {{1.0},{2.0},{3.0},{4.0},{5.0}}
		a = toMatrix (to2D aa) 5 1
		pa = {2,1,0,4,3}
		p = Permutation pa
		ip = Permutation (invPermute pa)
	in
	norm1 (a - (ip @* (p @* (toMatrix (to2D aa) 5 1)))) <= machineEpsilon
	
test107 =
	let aa = {{1.0},{2.0},{3.0},{4.0},{5.0}}
		a = toMatrix (to2D aa) 5 1
		p = Permutation {2,1,0,4,3}
	in
	norm1 (a - (p ~@* (p @* (toMatrix (to2D aa) 5 1)))) <= machineEpsilon
	
test108 =
	let aa = {{1.0+:0.0},{2.0+:0.1},{3.0+:0.2},{4.0+:0.3},{5.0+:0.4}}
		a = toMatrix (to2D aa) 5 1
		p = Permutation {2,1,0,4,3}
	in
	norm1 (a - (p ~@* (p @* (toMatrix (to2D aa) 5 1)))) <= machineEpsilon
	
test109 =
	let aa = {{1.0,2.0,3.0,4.0,5.0}}
		a = toMatrix (to2D aa) 1 5
		p = Permutation {2,1,0,4,3}
	in
	norm1 (a - (((toMatrix (to2D aa) 1 5) *@ p) *@~ p)) <= machineEpsilon
	
test110 =
	let aa = {{1.0+:0.0,2.0+:0.1,3.0+:0.2,4.0+:0.3,5.0+:0.4}}
		a = toMatrix (to2D aa) 1 5
		p = Permutation {2,1,0,4,3}
	in
	norm1 (a - (((toMatrix (to2D aa) 1 5) *@ p) *@~ p)) <= machineEpsilon
	
test111 =
	let m = 500
		n = 500
		nrhs = 5
		(a, rn) = rands (genRandReal 1) m n
		(b, _) = rands rn m nrhs
		xh = pluSolve (copy a) (copy b)
	in
		norm1 (a * xh - b) / (norm1 a * norm1 xh + norm1 b) <= 2.0 * machineEpsilon

test112 =
	let m = 20
		n = 10
		nrhs = 5
		(a, rn) = rands (genRandReal 1) m n
		(x, _) = rands rn n nrhs
		b = a * x
		xh = pluSolve (copy a) (copy b)
	in
		norm1 (a * xh - b) <= (norm1 a * norm1 xh + norm1 b) * machineEpsilon

test113 =
	let m = 500
		n = 500
		nrhs = 5
		(a, rn) = rands (genRandComplex 1 2) m n
		(b, _) = rands rn m nrhs
		xh = pluSolve (copy a) (copy b)
	in
		norm1 (a * xh - b) / (norm1 a * norm1 xh + norm1 b) <= 4.0 * machineEpsilon

test114 =
	let m = 20
		n = 10
		nrhs = 5
		(a, rn) = rands (genRandComplex 1 2) m n
		(x, _) = rands rn n nrhs
		b = a * x
		xh = pluSolve (copy a) (copy b)
	in
		norm1 (a * xh - b) <= (norm1 a * norm1 xh + norm1 b) * 2.0 * machineEpsilon

test115 =
	let m = 500
		n = 500
		nrhs = 5
		(a, rn) = rands (genRandReal 1) m n
		(b, _) = rands rn m nrhs
		xh = pluSolveT (copy a) (copy b)
	in
		norm1 (a ~* xh - b) / (norm1 a * norm1 xh + norm1 b) <= 2.0 * machineEpsilon

test116 =
	let m = 20
		n = 10
		nrhs = 5
		(a, rn) = rands (genRandReal 1) m n
		(x, _) = rands rn m nrhs
		b = matrix2x1 n (a :~*: x)
				(m - n) ooo         (zeros m nrhs)
		xh = pluSolveT (copy a) (copy b)
	in
		norm1 (a ~* (xh - x)) <= (norm1 a * norm1 xh + norm1 b) * machineEpsilon

test117 =
	let m = 500
		n = 500
		nrhs = 5
		(a, rn) = rands (genRandComplex 1 2) m n
		(b, _) = rands rn m nrhs
		xh = pluSolveT (copy a) (copy b)
	in
		norm1 (a ~* xh - b) / (norm1 a * norm1 xh + norm1 b) <= 4.0 * machineEpsilon

test118 =
	let m = 20
		n = 10
		nrhs = 5
		(a, rn) = rands (genRandComplex 1 2) m n
		(x, _) = rands rn m nrhs
		b = matrix2x1 n (a :~*: x)
				(m - n) ooo         (zeros m nrhs)
		xh = pluSolveT (copy a) (copy b)
	in
		norm1 (a ~* (xh - x)) <= (norm1 a * norm1 xh + norm1 b) * machineEpsilon

test119 =
	let m = 10
		n = 10
		(a, _) = rands (genRandReal 1) m n
		(p, l, u) = plu (copy a)
		lc = copyL l
		uc = copyU u
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test120 =
	let m = 20
		n = 10
		(a, _) = rands (genRandReal 1) m n
		(p, l, u) = plu (copy a)
		lc = copyL l
		uc = copyU u
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test121 =
	let m = 10
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		(p, l, u) = plu (copy a)
		lc = copyL l
		uc = copyU u
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test122 =
	let m = 20
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		(p, l, u) = plu (copy a)
		lc = copyL l
		uc = copyU u
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test123 =
	let m = 10
		n = 10
		(a, _) = rands (genRandReal 1) m n
		(p, l, u) = plu (copy a)
		lc = idL l (zeros m n)
		uc = idU u (zeros n n)
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test124 =
	let m = 20
		n = 10
		(a, _) = rands (genRandReal 1) m n
		(p, l, u) = plu (copy a)
		lc = idL l (zeros m n)
		uc = idU u (zeros n n)
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test125 =
	let m = 10
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		(p, l, u) = plu (copy a)
		lc = idL l (zeros m n)
		uc = idU u (zeros n n)
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test126 =
	let m = 20
		n = 10
		(a, _) = rands (genRandComplex 1 2) m n
		(p, l, u) = plu (copy a)
		lc = idL l (zeros m n)
		uc = idU u (zeros n n)
	in
		norm1 (a - (p @* (copy lc)) * uc) / (norm1 a + norm1 lc * norm1 uc) <=
			2.0 * machineEpsilon

test127 =
	let m = 10
		n = 7
		m0 = 4
		n0 = 5
		(a, _) = rands (genRandReal 1) m0 n0
		b = ((1 ... 4, 2 ... 6) <=$ a) (zeros m n)
	in
		norm1 (b $ (1 ... 4, 2 ... 6) -$ a) <= norm1 a * machineEpsilon

test128 =
	let m = 10
		n = 7
		m0 = 4
		n0 = 5
		(a, _) = rands (genRandComplex 1 2) m0 n0
		b = ((1 ... 4, 2 ... 6) <=$ a) (zeros m n)
	in
		norm1 (b $ (1 ... 4, 2 ... 6) -$ a) <= norm1 a * machineEpsilon

test129
	#! g = cholesky (copy a)
	= norm1 (a - g *~ g) <= machineEpsilon * norm1 a * toReal m
where
	m = 10
	b = foo m
	a = b ~* b
	foo :: Int -> *(Matrix Real)
	foo m = let (a, _) = rands (genRandReal 1) m m in a


// Timers are written in pairs, with the difference giving the right timing,
// or, use the time profiling tool.
time1_1
	#! x = copy a
	#! y = copy b
	= x @ (0,0) + y @ (0,0)
where
	m = 1000
	a = foo m m
	b = foo m 1
	foo :: !Int !Int -> *(Matrix Real)
	foo m n = let (a, _) = rands (genRandReal 1) m n in a

time1_2
	#! x = (copy a) :\: (copy b)
	= x @ (0,0)
where
	m = 1000
	a = foo m m
	b = foo m 1
	foo :: !Int !Int -> *(Matrix Real)
	foo m n = let (a, _) = rands (genRandReal 1) m n in a

time2_1
	#! g = copy a
	= g @ (0, 0)
where
	m = 1000
	b = foo m
	a = b ~* b
	foo :: Int -> *(Matrix Real)
	foo m = let (a, _) = rands (genRandReal 1) m m in a

time2_2
	#! g = chol (copy a)
	= g @ (0, 0)
where
	m = 1000
	b = foo m
	a = b ~* b
	foo :: Int -> *(Matrix Real)
	foo m = let (a, _) = rands (genRandReal 1) m m in a

// 1.99 seconds
time3 =
	(pluSolve a b) @ (0,0)
where
	m = 1000
	(a, rn) = rands (genRandReal 1) m m
	(b, _) = rands rn m 1

// 5.1 seconds
time4 =
	(a :\: b) @ (0,0)
where
	m = 1000
	(a, rn) = rands (genRandReal 1) m m
	(b, _) = rands rn m 1

time5_1
	 = (chol (b ~* b)) @ (0, 0)
where
	m = 1000
	(b, _) = rands (genRandReal 1) m m

time5_2
	 = (b ~* b) @ (0, 0)
where
	m = 1000
	(b, _) = rands (genRandReal 1) m m


//runTests tests = ["Test " +++ toString i +++ ": " +++ toString test +++ "\n" \\
//					test <- tests & i <- [1 .. ]]
runTests tests = ["Failed test: " +++ toString i +++ "\n" \\
					i <- (snd o unzip) (filter (\(a,b) = not a) (zip2 tests [1 .. ]))]

tests = [ test1, test2, test3, test4, test5, test6, test7, test8, test9, test10
		, test11, test12, test13, test14, test15, test16, test17, test18, test19
		, test20, test21, test22, test23, test24, test25, test26, test27, test28
		, test29, test30, test31, test32, test33, test34, test35, test36, test37
		, test38, test39, test40, test41, test42, test43, test44, test45, test46
		, test47, test48, test49, test50, test51, test52, test53, test54, test55
		, test56, test57, test58, test59, test60, test61, test62, test63, test64
		, test65, test66, test67, test68, test69, test70, test71, test72, test73
		, test74, test75, test76, test77, test78, test79, test80, test81, test82
		, test83, test84, test85, test86, test87, test88, test89, test90, test91
		, test92, test93, test94, test95, test96, test97, test98, test99, test100
		, test101, test102, test103, test104, test105, test106, test107, test108
		, test109, test110, test111, test112, test113, test114, test115, test116
		, test117, test118, test119, test120, test121, test122, test123, test124
		, test125, test126, test127, test128
		]

//Start = runTests tests
Start = time5_1
//Start = test129